/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog;

import java.io.IOException;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

/**
 * Represents a handle to a spatial resource.
 * <p>
 * The resource is not guaranteed to exist, nor do we guarantee that we can connect
 * with the resource. Some/All potions of this handle may be loaded as required.
 * This resource handle may also be the result a metadata service query.
 * 
 * @author David Zwiers, Refractions Research
 * @since 0.6
 */
public abstract class IGeoResource implements IResolve{
    
    /**
     * Blocking operation to resolve into the adaptee, if available.
     * <p>
     * Required adaptions:
     * <ul>
     * <li>IGeoResourceInfo.class
     * <li>IService.class
     * </ul>
     * </p>
     * <p>
     * Example (no casting required):<pre><code>
     * IGeoResourceInfo info = resovle( IGeoResourceInfo.class ); 
     * </code></pre>
     * </p>
     * <p>
     * Recommendated adaptions:
     * <ul>
     * <li>ImageDescriptor.class (for icon provided by external service)
     * </ul>
     * </p>
     * @param adaptee
     * @param monitor
     * @return instance of adaptee, or null if unavailable (IGeoResourceInfo and IService must be supported)
     * 
     * @see IGeoResourceInfo
     * @see IService
     * @see IResolve#resolve(Class, IProgressMonitor)
     */
    public abstract <T> T resolve(Class<T> adaptee,IProgressMonitor monitor) throws IOException;
    
    /**
     * Blocking operation to describe this service.
     * <p>
     * As an example this method is used by LabelDecorators to aquire title, and icon.
     * </p> 
     * @return IGeoResourceInfo resolve(IGeoResourceInfo.class,IProgressMonitor monitor);
     * 
     * @see IGeoResource#resolve(Class, IProgressMonitor)
     */
    public IGeoResourceInfo getInfo(IProgressMonitor monitor) throws IOException{
        return resolve(IGeoResourceInfo.class,monitor);
    }
    
    /**
     * Note: Parent may be IService or IGeoResource depending.
     * <p>
     * As an example WMS Layers are allowed to have child layers.
     * </p>
     * @return Parent of this GeoResource
     * @see IGeoResource#resolve(Class, IProgressMonitor)
     */
    public IResolve parent(IProgressMonitor monitor) throws IOException{
        return resolve(IService.class,monitor);
    }
    
    /**
     * @return IService resolve(IService.class,IProgressMonitor monitor);
     * @see IGeoResource#resolve(Class, IProgressMonitor)
     */    
    public IService service( IProgressMonitor monitor ) throws IOException {
        return resolve(IService.class,monitor);        
    }
    
    /**
     * return null ... almost always a leaf
     * @see net.refractions.udig.catalog.IResolve#members(org.eclipse.core.runtime.IProgressMonitor)
     */
    public List<IResolve> members( IProgressMonitor monitor ){
        return null;
    }
    
    /**
     * This should represent the identifier
     * 
     * @see Object#equals(java.lang.Object)
     * 
     * @param arg0
     * @return
     */
    public boolean equals( Object arg0 ){
        if(arg0!=null && arg0 instanceof IGeoResource){
            IGeoResource resource = (IGeoResource)arg0;
            if(getIdentifier()!=null && resource.getIdentifier()!=null)
                return getIdentifier().equals(resource.getIdentifier());
        }
        return false;
    }
    /**
     * This should represent the identified
     * 
     * @see Object#hashCode()
     * 
     * @return
     */
    public int hashCode(){
        if(getIdentifier()!=null)
            return getIdentifier().hashCode();
        return super.hashCode();
    }
    
    /**
     * Non blocking label used by LabelProvider.
     *
    public static final String getGenericLabel(IGeoResource resource){
        assert resource.getIdentifier() != null;
        return resource==null || resource.getIdentifier()==null?"Resource":resource.getIdentifier().toString();
        
    }*/
    /**
     * Non blocking icon used by LabelProvider.
     *
    public static final ImageDescriptor getGenericIcon(IGeoResource resource){
        if(resource !=null){
            assert resource.getIdentifier() != null;
            if(resource.canResolve(FeatureSource.class)){
                // default feature
                return Images.getDescriptor(ISharedImages.FEATURE_OBJ);
            }
            if(resource.canResolve(GridCoverage.class)){
                // default raster
                return Images.getDescriptor(ISharedImages.GRID_OBJ);
            }
        }
        return Images.getDescriptor(ISharedImages.RESOURCE_OBJ);
    }*/
    /**
     * Indicate class and id.
     * 
     * @return string representing this IResolve
     */
    public String toString(){
        StringBuffer buf = new StringBuffer();
        String classname = getClass().getName();
        String name = classname.substring( classname.lastIndexOf('.')+1);
        buf.append( name );
        buf.append( "(" ); //$NON-NLS-1$
        buf.append( getIdentifier() );
        buf.append( ")" ); //$NON-NLS-1$
        return buf.toString();
    }
}
