/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog;

import java.net.URI;
import java.net.URL;

import org.eclipse.jface.resource.ImageDescriptor;

/**
 * 
 * Provides metadata information about a service.
 * <p>
 * A bean - style metadata which must already be loaded. Much of the 
 * names and motivation have been taken from Dublin Code and it's application profile 
 * for RDF. 
 * </p>
 * @author David Zwiers, Refractions Research
 * @since 0.6
 */
public class IServiceInfo {
    
    protected String title, description, _abstract;
    protected URI schema;
    protected URL source,publisher;
    protected String[] keywords;
    protected ImageDescriptor icon;
    
    protected IServiceInfo(){
        // to be used in an over-ride
    }
    
    public IServiceInfo(String title, String description, String _abstract, URL source, URL publisher, URI schema, String[] keywords, ImageDescriptor icon){
        this.title = title;this.description = description;this._abstract = _abstract;
        this.schema = schema; this.source = source;this.publisher = publisher;
        this.keywords = keywords;this.icon = icon;
    }
    
    /**
     * Returns the service title, may be empty or null if unsupported.
     * <p>
     * Note this is always metadata, and is in user terms.
     * </p>
     * @return title, may be empty, null if unsupported.
     */
    public String getTitle(){
        return title;
    }
    
    /**
     * Returns the service keywords. Maps to the Dublin Core Subject element.
     * 
     * @return
     */
    public String[] getKeywords(){ // aka Subject
        return keywords;
    }
    
    /**
     * Returns the service description
     * 
     * @return
     */
    public String getDescription(){
        return description;
    }
    
    /**
     * Return the service abstract
     * 
     * @return
     */
    public String getAbstract(){
        return _abstract;
    }
    
    /**
     * Return the service publisher
     * 
     * @return
     */
    public URL getPublisher(){
        return publisher;
    }
        
    /**
     * Returns the xml schema namespace for this service type. 
     * 
     * Maps to the Dublin Code Format element
     * 
     * @return
     */
    public URI getSchema(){ // aka format
        return schema;
    }
    
    /**
     * Returns the service source. 
     * 
     * Maps to the Dublin Core Server Element
     * 
     * @return
     */
    public URL getSource(){ // aka server
        return source;
    }
    
    /**
     * Base symbology (with out decorators) representing this IService.
     * <p>
     * The ImageDescriptor returned should conform the the Eclipse User Interface
     * Guidelines (16x16 image with a 16x15 glyph centered).
     * </p>
     * <p>
     * This plug-in provides default images based on service type:
     * <pre><code>
     * <b>return</b> ISharedImages.getImagesDescriptor( IService );
     * </code></pre>
     * <ul>
     * 
     * <p>
     * Any LabelProvider should use the default image, a label decorator should be
     * used to pick up these images in a separate thread. This allows services like
     * WFS make blocking request to pick up the image from their GetCapabilities. 
     * </p>
     * @return ImageDescriptor symbolizing this  IService.
     */
    public ImageDescriptor getIcon(){
        return icon;
    }
}
