/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog.internal;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import net.refractions.udig.catalog.CatalogPlugin;
import net.refractions.udig.catalog.ICatalog;
import net.refractions.udig.catalog.ICatalogInfo;
import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.catalog.IGeoResourceInfo;
import net.refractions.udig.catalog.IResolve;
import net.refractions.udig.catalog.IResolveChangeEvent;
import net.refractions.udig.catalog.IResolveChangeListener;
import net.refractions.udig.catalog.IResolveDelta;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.IServiceInfo;
import net.refractions.udig.catalog.util.AST;
import net.refractions.udig.catalog.util.ASTFactory;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.jface.util.ListenerList;

import com.vividsolutions.jts.geom.Envelope;

/**
 * Implementation of an in memory catalog.
 * 
 * @author David Zwiers, Refractions Research
 * @since 0.6
 */
public class CatalogImpl extends ICatalog {
    private HashSet<IService> services = new HashSet<IService>();
    private ICatalogInfo metadata;
    private ListenerList catalogListeners;

    public CatalogImpl() {
        CatalogInfoImpl metadata = new CatalogInfoImpl();
        metadata.setTitle(Policy.bind("CatalogImpl.localCatalog.title")); //$NON-NLS-1$
        try {
            metadata.setSource(new URL("localhost")); //$NON-NLS-1$
        } catch (MalformedURLException e) {
            // do nothing
        }

        this.metadata = metadata;
        catalogListeners = new ListenerList();
    }

    public CatalogImpl( ICatalogInfo metadata ) {
        this();
        this.metadata = metadata;
    }

    /**
     * @see net.refractions.udig.catalog.ICatalog#addCatalogListener(net.refractions.udig.catalog.ICatalog.ICatalogListener)
     * @param listener
     */
    public void addCatalogListener( IResolveChangeListener listener ) {
        catalogListeners.add(listener);
    }

    /**
     * @see net.refractions.udig.catalog.ICatalog#removeCatalogListener(net.refractions.udig.catalog.ICatalog.ICatalogListener)
     * @param listener
     */
    public void removeCatalogListener( IResolveChangeListener listener ) {
        catalogListeners.remove(listener);
    }

    /**
     * @see net.refractions.udig.catalog.ICatalog#add(net.refractions.udig.catalog.IService)
     * @param entry
     * @throws UnsupportedOperationException
     */
    public void add( IService entry ) throws UnsupportedOperationException {
        if (entry == null || entry.getIdentifier() == null)
            throw new NullPointerException("Cannot have a null id"); //$NON-NLS-1$
        services.add(entry);
        IResolveDelta deltaAdded = new ResolveDelta(entry, IResolveDelta.Kind.ADDED);
        IResolveDelta deltaChanged = new ResolveDelta(this, Collections.singletonList(deltaAdded));
        fire(new ResolveChangeEvent(CatalogImpl.this, IResolveChangeEvent.Type.POST_CHANGE,
                deltaChanged));
    }

    /**
     * @see net.refractions.udig.catalog.ICatalog#remove(net.refractions.udig.catalog.IService)
     * @param entry
     * @throws UnsupportedOperationException
     */
    public void remove( IService entry ) throws UnsupportedOperationException {
        if (entry == null || entry.getIdentifier() == null)
            throw new NullPointerException("Cannot have a null id"); //$NON-NLS-1$
        IResolveDelta deltaRemoved = new ResolveDelta(entry, IResolveDelta.Kind.REMOVED);
        IResolveDelta deltaChanged = new ResolveDelta(this, Collections.singletonList(deltaRemoved));
        fire(new ResolveChangeEvent(CatalogImpl.this, IResolveChangeEvent.Type.PRE_DELETE,
                deltaChanged));
        services.remove(entry);
        fire(new ResolveChangeEvent(CatalogImpl.this, IResolveChangeEvent.Type.POST_CHANGE,
                deltaRemoved));
    }

    public void replace( URL id, IService entry ) throws UnsupportedOperationException {
        if (entry == null || entry.getIdentifier() == null || id == null)
            throw new NullPointerException("Cannot have a null id"); //$NON-NLS-1$

        List<IService> victims = findService(id);

        List<IResolveDelta> changes = new ArrayList<IResolveDelta>();
        for( IService service : victims ) {
                List<IResolveDelta> childChanges = new ArrayList<IResolveDelta>();
                try {
                    List< ? extends IGeoResource> newChildren = entry.members(null);
                    for( IGeoResource oldChild : service.members(null) ) {
                        String oldName = oldChild.getIdentifier().toString();
                        for( IGeoResource child : newChildren ) {
                            String name = child.getIdentifier().toString();
                            if (oldName.equals(name)) {
                                childChanges.add(new ResolveDelta(child, oldChild,
                                        IResolveDelta.NO_CHILDREN));
                                break;
                            }
                        }
                    }
                    childChanges.add(new ResolveDelta(entry, service, childChanges));
                } catch (IOException ignore) {
                    // no children? Not a very good entry ..
                }
                changes.add(new ResolveDelta(service, entry, childChanges));
        }
        IResolveDelta deltas = new ResolveDelta(this, changes);
        IResolveChangeEvent event = new ResolveChangeEvent(this,
                IResolveChangeEvent.Type.PRE_DELETE, deltas);
        fire(event);

        services.removeAll(victims);
        services.add(entry);
    }

    /**
     * Quick search by url match.
     * @param query
     * 
     * @see net.refractions.udig.catalog.ICatalog#search(org.geotools.filter.Filter)
     * @return List<IResolve>
     * @throws IOException
     */
    public List<IResolve> find( URL query, IProgressMonitor monitor ) {
        List<IResolve> list = new ArrayList<IResolve>();        
        for( IService service : services ) {
            if (service.getIdentifier().getProtocol().equalsIgnoreCase(query.getProtocol())
                    && ((service.getIdentifier().getHost() == null || "".equals(service.getIdentifier().getHost())) || (service.getIdentifier().getHost() != null && service.getIdentifier().getHost().equalsIgnoreCase(query.getHost()))) && //$NON-NLS-1$
                    ((service.getIdentifier().getPath() == null || "".equals(service.getIdentifier().getPath())) || (service.getIdentifier().getPath() != null && service.getIdentifier().getPath().equalsIgnoreCase(query.getPath()))) && //$NON-NLS-1$
                    ((service.getIdentifier().getQuery() == null || "".equals(service.getIdentifier().getQuery())) || (service.getIdentifier().getQuery() != null && service.getIdentifier().getQuery().equalsIgnoreCase(query.getQuery()))) && //$NON-NLS-1$
                    ((service.getIdentifier().getAuthority() == null || "".equals(service.getIdentifier().getAuthority())) || (service.getIdentifier().getAuthority() != null && service.getIdentifier().getAuthority().equalsIgnoreCase(query.getAuthority())))) { //$NON-NLS-1$
                Iterator< ? extends IGeoResource> i;
                try {
                    if (query.getRef() == null) {
                        list.add(service);
                        continue;
                    }

                    /*
                     * Although the following is a 'blocking' call, we have deemed it safe based on
                     * the following reasons: This will only be called for Identifiers which are
                     * well known. The Services being checked have already been screened, and only a
                     * limited number of services (ussually 1) will be called. 1) The Id was aquired
                     * from the catalog ... and this is a look-up, in which case the uri exists. 2)
                     * The Id was persisted. In the future this will also be free, as we plan on
                     * caching the equivalent of a getCapabilities document between runs (will have
                     * to be updated too as the app has time).
                     */
                    List< ? extends IGeoResource> t = service.members(null);
                    i = t == null ? null : t.iterator();
                    if (query.getRef() != null) {
                        // it's a resource
                        while( i.hasNext() ) {
                            IGeoResource res = i.next();
                            if (query.getRef().equals(res.getIdentifier().getRef()))
                                list.add(res);
                        }
                    } else {
                        while( i != null && i.hasNext() ) {
                            list.add(i.next());
                        }
                    }
                } catch (IOException e) {
                    CatalogPlugin.trace(null, e);
                }

            }
        }
        return list;
    }

    /**
     * Quick search by url match.
     * 
     * @see net.refractions.udig.catalog.ICatalog#search(org.geotools.filter.Filter)
     * @param query
     * @return List<IResolve>
     * @throws IOException
     */
    public List<IService> findService( URL query ) {
        List<IService> list = new ArrayList<IService>();
        if (query.getRef() != null) 
            return list;
        for( IService service : services ) {
            if (service.getIdentifier().getProtocol().equalsIgnoreCase(query.getProtocol())
                    && ((service.getIdentifier().getHost() == null || "".equals(service.getIdentifier().getHost())) || (service.getIdentifier().getHost() != null && service.getIdentifier().getHost().equalsIgnoreCase(query.getHost()))) && //$NON-NLS-1$
                    ((service.getIdentifier().getPath() == null || "".equals(service.getIdentifier().getPath())) || (service.getIdentifier().getPath() != null && service.getIdentifier().getPath().equalsIgnoreCase(query.getPath()))) && //$NON-NLS-1$
                    ((service.getIdentifier().getQuery() == null || "".equals(service.getIdentifier().getQuery())) || (service.getIdentifier().getQuery() != null && service.getIdentifier().getQuery().equalsIgnoreCase(query.getQuery()))) && //$NON-NLS-1$
                    ((service.getIdentifier().getAuthority() == null || "".equals(service.getIdentifier().getAuthority())) || (service.getIdentifier().getAuthority() != null && service.getIdentifier().getAuthority().equalsIgnoreCase(query.getAuthority())))) { //$NON-NLS-1$
                        list.add(service);
            }
        }
        return list;
    }

    /**
     * Performs a search on this catalog based on the specified inputs. The pattern uses the
     * following conventions: use " " to surround a phase use + to represent 'AND' use - to
     * represent 'OR' use ! to represent 'NOT' use ( ) to designate scope The bbox provided shall be
     * in Lat - Long, or null if the search is not to be contained within a specified area.
     * 
     * @see net.refractions.udig.catalog.ICatalog#search(java.lang.String,
     *      com.vividsolutions.jts.geom.Envelope)
     * @param pattern
     * @param bbox used for an intersection test
     * @return
     */
    public synchronized List<IResolve> search( String pattern, Envelope bbox,
            IProgressMonitor monitor ) {
        if (monitor == null)
            monitor = new NullProgressMonitor();
        if ((bbox == null || bbox.isNull()) && pattern == null || "".equals(pattern.trim())) //$NON-NLS-1$
            return new LinkedList<IResolve>();

        AST ast = ASTFactory.parse(pattern);

        // TODO check cuncurrency issues here

        List<IResolve> result = new LinkedList<IResolve>();
        HashSet<IService> tmp = new HashSet<IService>();
        tmp.addAll(this.services);

        monitor.beginTask("Search: " + pattern, tmp.size());
        try {
            Iterator<IService> services = tmp.iterator();
            if (services != null) {
                while( services.hasNext() ) {
                    try {
                        IService service = services.next();
                        if (check(service, ast, monitor)) {
                            result.add(service);
                        }
                        Iterator< ? extends IGeoResource> resources;
                        try {
                            List< ? extends IGeoResource> t = service.members(null);
                            resources = t == null ? null : t.iterator();
                            while( resources != null && resources.hasNext() ) {
                                IGeoResource resource = resources.next();
                                if (check(resource, ast, bbox)) {
                                    result.add(resource);
                                }
                            }
                        } catch (IOException e) {
                            CatalogPlugin.log(null, e);
                        }
                    } finally {
                        monitor.worked(1);
                    }
                }
            }
        } finally {
            monitor.done();
        }
        return result;
    }

    /* check the fields we catre about */
    protected static boolean check( IService service, AST pattern, IProgressMonitor monitor ) {
        IServiceInfo info;
        try {
            info = service == null ? null : service.getInfo(monitor);
        } catch (IOException e) {
            info = null;
            CatalogPlugin.log(null, e);
        }
        boolean t = false;
        if (info != null) {
            if (info.getTitle() != null)
                t = pattern.accept(info.getTitle());
            if (!t && info.getKeywords() != null) {
                String[] keys = info.getKeywords();
                for( int i = 0; !t && i < keys.length; i++ )
                    if (keys[i] != null)
                        t = pattern.accept(keys[i]);
            }
            if (!t && info.getSchema() != null)
                t = pattern.accept(info.getSchema().toString());
            if (!t && info.getAbstract() != null)
                t = pattern.accept(info.getAbstract());
            if (!t && info.getDescription() != null)
                t = pattern.accept(info.getDescription());
        }
        return t;
    }

    /* check the fields we catre about */
    protected static boolean check( IGeoResource resource, AST pattern ) {
        IGeoResourceInfo info;
        try {
            info = (resource == null ? null : resource.getInfo(null));
        } catch (IOException e) {
            CatalogPlugin.log(null, e);
            info = null;
        }
        boolean t = false;
        if (info != null) {
            if (info.getTitle() != null)
                t = pattern.accept(info.getTitle());
            if (!t && info.getName() != null)
                t = pattern.accept(info.getName());
            if (!t && info.getKeywords() != null) {
                String[] keys = info.getKeywords();
                for( int i = 0; !t && i < keys.length; i++ )
                    if (keys[i] != null)
                        t = pattern.accept(keys[i]);
            }
            if (!t && info.getSchema() != null)
                t = pattern.accept(info.getSchema().toString());
            if (!t && info.getDescription() != null)
                t = pattern.accept(info.getDescription());
        }
        return t;
    }

    protected static boolean check( IGeoResource resource, AST pattern, Envelope bbox ) {
        if (!check(resource, pattern))
            return false;
        if (bbox == null || bbox.isNull())
            return true; // no checking here
        try {
            return bbox.intersects(resource.getInfo(null).getBounds());
        } catch (IOException e) {
            CatalogPlugin.log(null, e);
            return false;
        }
    }

    /**
     * Fire a resource changed event, these may be batched into one delta for performance.
     * 
     * @param resoruce IGeoResource undergoing change
     * @param mask of IDelta constants indicating change
     * @throws IOException protected void fireResourceEvent( IGeoResource resource,
     *         IResolveDelta.Kind kind ) throws IOException { Object[] listeners =
     *         catalogListeners.getListeners(); if( listeners.length == 0 ) return;
     *         GeoReferenceDelta rDelta = new GeoReferenceDelta( resource, kind ); ServiceDelta
     *         sDelta = new ServiceDelta( resource.getService(null), IDelta.Kind.NO_CHANGE,
     *         Collections.singletonList( rDelta ) ); CatalogDelta cDelta = new CatalogDelta(
     *         Collections.singletonList( (IDelta)sDelta ) ); fire( new CatalogChangeEvent(
     *         resource, ICatalogChangeEvent.Type.POST_CHANGE, cDelta ) ); }
     */

    public void fire( IResolveChangeEvent event ) {
        Object[] listeners = catalogListeners.getListeners();
        if (listeners.length == 0)
            return;

        for( int i = 0; i < listeners.length; ++i ) {
            try {
                ((IResolveChangeListener) listeners[i]).changed(event);
            } catch (Throwable die) {
                CatalogPlugin.log(null, new Exception(die));
            }
        }
    }
    /*
     * protected void fireServiceChanged( Object source, ServiceDelta sDelta ) { Object[] listeners =
     * catalogListeners.getListeners(); if( listeners.length == 0 ) return; CatalogDelta cDelta =
     * new CatalogDelta( Collections.singletonList( (IDelta)sDelta ) ); fire( new
     * CatalogChangeEvent( source , ICatalogChangeEvent.Type.POST_CHANGE, cDelta ) ); }
     */

    /**
     * @see net.refractions.udig.catalog.ICatalog#resolve(java.lang.Class,
     *      org.eclipse.core.runtime.IProgressMonitor)
     * @SuppressWarnings(value={"unchecked"})
     */
    public <T> T resolve( Class<T> adaptee, IProgressMonitor monitor ) {
        if (monitor == null)
            monitor = new NullProgressMonitor();
        monitor.beginTask("Resolve", 1);
        try {
            if (adaptee == null)
                return null;
            if (adaptee.isAssignableFrom(CatalogImpl.class))
                return adaptee.cast(this);
            if (adaptee.isAssignableFrom(CatalogInfoImpl.class))
                return adaptee.cast(metadata);
            if (adaptee.isAssignableFrom(services.getClass()))
                return adaptee.cast(services);
            if (adaptee.isAssignableFrom(List.class))
                return adaptee.cast(new LinkedList<IService>(services));
            if (adaptee.isAssignableFrom(catalogListeners.getClass()))
                return adaptee.cast(catalogListeners);
            return null;
        } finally {
            monitor.worked(1);
            monitor.done();
        }
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#canResolve(java.lang.Class)
     */
    public <T> boolean canResolve( Class<T> adaptee ) {
        Object value = resolve(adaptee, null);
        return value != null;
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#members(org.eclipse.core.runtime.IProgressMonitor)
     */
    public List<IResolve> members( IProgressMonitor monitor ) {
        if (monitor == null)
            monitor = new NullProgressMonitor();
        monitor.beginTask("members", 1);
        try {
            return new LinkedList<IResolve>(services);
        } finally {
            monitor.done();
        }
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#getStatus()
     */
    public Status getStatus() {
        return Status.CONNECTED;
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#getMessage()
     */
    public Throwable getMessage() {
        return null;
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#getIdentifier()
     */
    public URL getIdentifier() {
        return metadata.getSource();
    }
}