package net.refractions.udig.catalog.internal.postgis;

import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import org.geotools.data.DataSourceException;
import org.geotools.data.jdbc.JDBCUtils;
import org.geotools.data.jdbc.SQLBuilder;
import org.geotools.data.postgis.PostgisDataStore;
import org.geotools.data.postgis.PostgisFeatureLocking;
import org.geotools.feature.FeatureType;
import org.geotools.feature.GeometryAttributeType;
import org.geotools.filter.SQLEncoderException;
import org.geotools.geometry.JTS;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.io.ParseException;

/**
 * Subclass of PostgisFeatureLocking which optimizes some of the FeatureStore
 * operations.
 * 
 * @author Justin Deoliveira,Refractions Research Inc.,jdeolive@refractions.net
 *
 */
public class UDIGPostGISFeatureLocking extends PostgisFeatureLocking {
	
	public UDIGPostGISFeatureLocking(PostgisDataStore ds, FeatureType type) 
		throws IOException {
		
		super(ds,type);
	}
	
	@Override
	/**
	 * Optimized implementation of getBounds(). Avoids a full table scan by 
	 * using version > 1.0 postgis functionality. In the event of a < 0.1 
	 * database, an approximate bounds is generated.
	 */
	public Envelope getBounds() throws IOException {
        //return getBounds(Query.ALL);
    	FeatureType schema = getSchema();
    	GeometryAttributeType geom = schema.getDefaultGeometry();
    	
    	Envelope envelope = null;
    	if (geom != null) {
    		Connection conn = getConnection();
    		SQLBuilder sqlBuilder 
    			= getJDBCDataStore().getSqlBuilder(schema.getTypeName());
    		
    		try {
    			envelope = getEntireEnvelope(conn,geom.getName(),sqlBuilder);
    			if (envelope != null) {
    				CoordinateReferenceSystem crs = geom.getCoordinateSystem();
    				envelope = new JTS.ReferencedEnvelope(envelope,crs);
				}
    		} 
    		catch (SQLException sqle) {
    			JDBCUtils.close(conn, transaction, sqle);
    			throw new DataSourceException(Policy.bind("UDIGPostGISFeatureLocking.boundsError"), sqle); //$NON-NLS-1$
    		}
    		catch (Exception e) {
    			JDBCUtils.close(conn, transaction, null);
    			throw new DataSourceException(Policy.bind("UDIGPostGISFeatureLocking.boundsError"), e); //$NON-NLS-1$
    		}
    		
    		JDBCUtils.close(conn, transaction, null);
    	}
    	
    	return envelope;
    }
	
	/**
     * Optimized function to obtains the approximate bounds of the table. 
     */
    protected Envelope getEntireEnvelope(Connection conn, String geomName, SQLBuilder sqlBuilder)
		throws SQLException, SQLEncoderException, IOException, ParseException {
	
		//optimization, version >= 1.0 contains estimate_extent function 
		// to query the stats of the table to determine the bbox, however, 
		// it may return null
	    StringBuffer sql = new StringBuffer();
	    sql.append("SELECT postgis_version()"); //$NON-NLS-1$
	    Statement st = conn.createStatement();
	    ResultSet rs = st.executeQuery(sql.toString());
	    
	    //should always return a result
	    rs.next();
	    String version = rs.getString(1);
	    rs.close();
	    st.close();
	    
	    Envelope envelope = null;
	    if (version.trim().startsWith("1.0")) { //$NON-NLS-1$
	    	//try the estimated_extent function
	    	sql = new StringBuffer();
	    	PreparedStatement pst = conn.prepareStatement(
    			"SELECT  AsText(force_2d(envelope(estimated_extent(?,?))))" //$NON-NLS-1$
			);
	    	pst.setString(1, tableName);
	    	pst.setString(2, geomName);
	    	
	    	rs = pst.executeQuery();
	    	
	    	if (rs.next()) {
	    		//parse return value
	    		String wkt = rs.getString(1);
	    		if (wkt != null &&  !wkt.trim().equals("")) { //$NON-NLS-1$
	    			envelope = geometryReader.read(wkt).getEnvelopeInternal();
	    			
	    			// expand the bounds by 20% (10% in each direction)
	    			// Works whether or not the bounds are at the origin
	    			double minX = envelope.getMinX();
	    			double minY = envelope.getMinY();
	    			double maxX = envelope.getMaxX();
	    			double maxY = envelope.getMaxY();
	    			double deltaX = (maxX - minX)*0.1;
	    			double deltaY = (maxY - minY)*0.1;
	    			envelope.expandToInclude(minX - deltaX, minY - deltaY);
	    			envelope.expandToInclude(maxX + deltaX, maxY + deltaY);
	    		}
	    	}
	    	
	    	rs.close();
	    	pst.close();
	    }
	    
	    if (envelope == null) {
	    	
	    	//try to generate an approximation
	    	envelope = new Envelope();
	    	st = conn.createStatement();
	    	int offset = 0;
	    	for (int i = 0; i < 5; i++,offset+=10000) {
	    		String q = "SELECT AsText(force_2d(envelope(" + geomName +  //$NON-NLS-1$
	    			"))) FROM " + tableName + " LIMIT 1 OFFSET " + offset; //$NON-NLS-1$ //$NON-NLS-2$
	    		rs = st.executeQuery(q);
	    		if (rs.next()) {
	    			String wkt = rs.getString(1);
	    			if (wkt != null && !wkt.trim().equals("")) { //$NON-NLS-1$
	    				Envelope e = geometryReader.read(wkt)
	    					.getEnvelopeInternal();
	    				
	    				if (envelope.isNull()) 
	    					envelope.init(e);
	    				else 
	    					envelope.expandToInclude(e);
	    			}
	    			
	    		}
	    	}
	    	
	    	// expand generously since this is an approximation
	    	// Works whether or not the bounds are at the origin
	    	double minX = envelope.getMinX();
	    	double minY = envelope.getMinY();
	    	double maxX = envelope.getMaxX();
	    	double maxY = envelope.getMaxY();
	    	double deltaX = (maxX - minX)*3;
	    	double deltaY = (maxY - minY)*3;
	    	envelope.expandToInclude(minX - deltaX, minY - deltaY);
	    	envelope.expandToInclude(maxX + deltaX, maxY + deltaY);
	    }
	    
	    return envelope;
    }

}