package net.refractions.udig.catalog.internal.ui;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;

import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.catalog.IResolve;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.ui.CatalogUIPlugin;
import net.refractions.udig.catalog.ui.ConnectionErrorPage;
import net.refractions.udig.catalog.ui.IConnectionErrorHandler;
import net.refractions.udig.catalog.ui.IDataWizard;
import net.refractions.udig.catalog.ui.UDIGImportPage;
import net.refractions.udig.core.internal.ExtensionPointProcessor;
import net.refractions.udig.core.internal.ExtensionPointUtil;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.dialogs.IPageChangedListener;
import org.eclipse.jface.dialogs.PageChangedEvent;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.wizard.IWizardPage;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Text;

public class ConnectionPage extends WizardPage 
	implements UDIGImportPage, IPageChangedListener {

	/** underlying import page **/
	UDIGImportPage page;
	
	/** cache services **/
	List<IResolve> resources;
	
	public String getId() {
		return page.getId();
	}
	
	public boolean canProcess(Object object) {
		return page.canProcess(object);
	}
	
	public ConnectionPage(UDIGImportPage page) {
		super(page.getName());
		this.page = page;
	}

	public void createControl(Composite parent) {
		page.createControl(parent);
		setControl(page.getControl());
		
		if (getContainer() instanceof WizardDialog) {
			WizardDialog dialog = (WizardDialog)getContainer();
			dialog.addPageChangedListener(this);
		}
	}
	
	public List<IResolve> getResources(IProgressMonitor monitor) throws Exception {
		if (resources != null)
			return resources;
		
		return page.getResources(monitor);
	}
	
	@Override
	public boolean isPageComplete() {
		return page.isPageComplete();
	}
	
	@Override
	public boolean canFlipToNextPage() {
		//this one is tricky, if connection error occurs, then there is 
		// another secondary page, otherwise there is not unless there is 
		// another primary page, we assume that an error will not occur
		IDataWizard wizard = (IDataWizard)getWizard();
		return isPageComplete() && wizard.hasMorePrimaryPages();
	}
	
	@Override
	public String getName() {
		return page.getName();
	}
	
    @Override
    public Image getImage() {
        return page.getImage();
    }
    
	@Override
	public String getDescription() {
		return page.getDescription();
	}
	
	@Override
	public String getTitle() {
		return page.getTitle();
	}
	
	@Override
    /**
     * This method attempts a connection to the service. If the connection is 
     * made, control is given back to the wizard. If a connection error occurs
     * then the connection error page is returned.
     * 
     * @see ConnectionErrorPage
     */
	public IWizardPage getNextPage() {
	
		if (resources == null) {
			//attempt to make the connection
			final ArrayList<Throwable> throwable = new ArrayList<Throwable>();
			IRunnableWithProgress runnable = new IRunnableWithProgress() {
				public void run(IProgressMonitor monitor) 
					throws InvocationTargetException, InterruptedException {
					
					try {
						resources = getResources(monitor);
						
						//make sure connection occurred
						for (IResolve resource : resources) {
							if (resource instanceof IService) {
								IService service = (IService)resource;
								service.members(monitor);
							}
							else if (resource instanceof IGeoResource) {
								//ok, already have a handle, connection 
								// already took place
							}
						}
					}
					catch(Throwable t) {
						throwable.add(t); 
					}
				}
			};
			try {
				getContainer().run(false, true, runnable);
			} 
			catch (InvocationTargetException e) {
				//TODO: something
			} 
			catch (InterruptedException e) {
				//TODO: something
			}
			
			if (!throwable.isEmpty()) {
				Throwable t = throwable.get(0);
				
				//connection failed, look for a handler
				ConnectionHandlerProcessor p 
					= new ConnectionHandlerProcessor(throwable.get(0));
				ExtensionPointUtil.process(
					CatalogUIPlugin.getDefault(), IConnectionErrorHandler.XPID, p	
				);
				
				if (p.handlers.isEmpty()) {
					//no handlers available, create a default
					SimpleConnectionErrorHandler handler =
						new SimpleConnectionErrorHandler(t);
					p.handlers.add(handler);
				}
				
				ConnectionErrorPage page = new ConnectionErrorPage(p.handlers);
				IDataWizard wizard = (IDataWizard)getWizard();
				wizard.init(page);
				
				return page;
			}
		}
		
		return super.getNextPage(); //connection ok, no more pages in string
	}
	
	@Override
	public void dispose() {
		super.dispose();
		
		if (getContainer() instanceof WizardDialog) {
			WizardDialog dialog = (WizardDialog) getContainer();
			dialog.removePageChangedListener(this);
		}
	}
	/**
	 * Determines if hte current page is itself, if so, resets the resource 
	 * cache so that the user can make changes to connection parameters
	 */
	public void pageChanged(PageChangedEvent event) {
		if (this.equals(event.getSelectedPage())) {
			//reset the resource cache
			resources = null;
		}
	} 
	
	private class ConnectionHandlerProcessor implements ExtensionPointProcessor {

		Throwable t;
		List<IConnectionErrorHandler> handlers 
			= new ArrayList<IConnectionErrorHandler>();
		
		ConnectionHandlerProcessor(Throwable t) {
			this.t = t;
		}
		
		public void process(IExtension extension, IConfigurationElement element) 
			throws Exception {
			try {
				IConnectionErrorHandler handler = 
					(IConnectionErrorHandler)element.createExecutableExtension("class"); //$NON-NLS-1$
				
				if (handler.canHandle(t)) {
					handler.setName(element.getAttribute("name")); //$NON-NLS-1$
					handler.setThrowable(t);
					handlers.add(handler);
				}
			}
			catch(Throwable t) {
				String msg = t.getLocalizedMessage();
				CatalogUIPlugin.log(msg, t);
			}
			
		}
	}
	
	private class SimpleConnectionErrorHandler extends IConnectionErrorHandler {

		SimpleConnectionErrorHandler(Throwable t) {
			setName("Simple"); //$NON-NLS-1$
			setThrowable(t);
		}
		
		@Override
		public boolean canHandle(Throwable t) {
			return true;
		}

		@Override
		public boolean canRecover() {
			return false;
		}
		
		@Override
		protected Control create(Composite parent) {
			Throwable t = getThrowable();
			CatalogUIPlugin.log("Connection Error",t); //$NON-NLS-1$
			ByteArrayOutputStream out = new ByteArrayOutputStream();
			t.printStackTrace(new PrintStream(out));
			
			Text text = 
				new Text(parent, SWT.BORDER|SWT.MULTI|SWT.V_SCROLL|SWT.SCROLL_LINE|SWT.H_SCROLL);
			text.setText(new String(out.toByteArray()));
			text.setLayoutData(new GridData(GridData.FILL_BOTH));
			
			return text;
		}
		
	}

	
}

