package net.refractions.udig.catalog.internal.ui;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.window.Window;
import org.eclipse.jface.wizard.IWizard;
import org.eclipse.jface.wizard.IWizardPage;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbenchWizard;
import org.eclipse.ui.PlatformUI;

/**
 * 
 * A wizard container that runs headless (no ui) through wizard pages.
 * <p>
 * The headless wizard dialog becomes visible once a page cannot complete
 * itself. Headless wizard execution is dependent on the individual pages
 * using the workbench selection to "seed" themselves.
 * </p>
 * <p>
 * The following example illustrates use of the dialog. It is important to not
 * that open() must be called before runHeadless().
 * <pre><code>
 *  IWizard wizard = ...
 *  Shell shell = new Shell(Display.getDefault());
 *  HeadlessWizardDialog dialog = new HeadlessWizardDialog(shell,wizard);
 *                   
 *  dialog.setBlockOnOpen(true);
 *  dialog.open();
 *  dialog.runHeadless();
 *                   
 *  if (!shell.isDisposed())
 *      shell.dispose();
 *                   
 * </code></pre>
 * </p>
 * @author Justin Deoliveira,Refractions Research Inc.,jdeolive@refractions.net
 * @since 0.6.0
 */
public class HeadlessWizardDialog extends WizardDialog {

    /** override workbench selection **/
    IStructuredSelection selection;
    
    /**
     * Construct <code>HeadlessWizardDialog</code>.
     *
     * @param parentShell
     * @param newWizard
     */
	public HeadlessWizardDialog(Shell parentShell, IWizard newWizard) {
		super(new Shell(parentShell, SWT.SYSTEM_MODAL), newWizard);
        init();

	}

    /**
     * Initializes the wizard.
     *
     */
    protected void init() {
        // initialize the wizard
        IWizard wizard = getWizard();
        if (wizard instanceof IWorkbenchWizard) {
            IStructuredSelection s = (IStructuredSelection) PlatformUI
                .getWorkbench().getActiveWorkbenchWindow().getSelectionService()
                .getSelection();
            IWorkbenchWizard wwizard = (IWorkbenchWizard)wizard;
            
            if (selection == null) {
                //use the workbench selection
                if (s == null)
                    s = new StructuredSelection();
                    
                selection = s;
            }
            
            wwizard.init(PlatformUI.getWorkbench(), selection);
        }
    }
    /**
     * Sets the container selection. This method can be used in instances
     * where a workbench selection is not available.  
     * 
     * @param selection The selection.
     */
    public void setSelection( IStructuredSelection selection ) {
        this.selection = selection;
        
        //reinitialize the wizard with new selection
        init();
    }
    
    /**
     * Returns the selection. Empty if no selection set.
     *
     */
    public IStructuredSelection getSelection() {
        return selection != null ? selection : new StructuredSelection();
    }
    
    /**
     * Pseudo opens the dialog.
     * 
     *<p>
     *  Performs every step in the normal dialog open, except for setting
     *  the dialog shell visible.  
     *</p>
     */
    @Override
	public int open() {
		Shell shell = getShell();
		if (shell == null || shell.isDisposed()) {
            shell = null;
			// create the window
			create();
		}

		return Window.OK;
	}
    
    /**
     * Signals the container to process wizard pages starting at the 
     * current page.
     * 
     * <p>This method returns when a page return false from 
     * @see IWizardPage#isPageComplete(). It is important to note that 
     * this method should not be called before open().  
     * </p>
     */
    public void runHeadless() {
        IWizardPage page = getCurrentPage();
        IWizardPage curr = null;
        while (page != null && page.isPageComplete()) {
            nextPressed();

            curr = page;
            page = getCurrentPage();
            if (page == null)
                break;
            if (page == curr) {
                // we are done
                finishPressed();
                page = null;
            }
        }

        if (page != null) {
            // a page could not complete, open the dialog
            super.open();
        }
    }
    
	/**
	 * Resumes the dialog. This method is mostly used for testing purposes. The 
	 * dialog is normally signalled to continue when the user presses the next
	 * button.
	 */
	public void resume() {
		buttonPressed(IDialogConstants.NEXT_ID);
	}
	
    /**
     * Ensures that the finish button is always enabled. 
     * <p>
     * This provides the ability to resume headless execution after an 
     * intermediate page was not able to complete itself.
     * </p>
     */
    @Override
    public void updateButtons() {
        super.updateButtons();
        
        //make sure finish is always enabled
        getButton(IDialogConstants.FINISH_ID).setEnabled(true);
    }
    
    /**
     * Intercepts the "Finish" event. 
     * 
     * <p>
     * The Finish event resumes headless execution of the wizard.
     * </p>
     */
	@Override
	protected void buttonPressed(int buttonId) {
        if(buttonId == IDialogConstants.FINISH_ID) {
			getShell().setVisible(false);
			runHeadless();
            return;
		}
        
        super.buttonPressed(buttonId);
	}

}