/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.catalog.internal.wms;

import static net.refractions.udig.catalog.internal.wms.Trace.REQUEST;

import java.io.IOException;
import java.io.Serializable;
import java.net.URI;
import java.net.URL;
import java.text.MessageFormat;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import net.refractions.udig.catalog.CatalogPlugin;
import net.refractions.udig.catalog.IResolveChangeEvent;
import net.refractions.udig.catalog.IResolveDelta;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.IServiceInfo;
import net.refractions.udig.catalog.internal.CatalogImpl;
import net.refractions.udig.catalog.internal.ResolveChangeEvent;
import net.refractions.udig.catalog.internal.ResolveDelta;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.geotools.data.ows.Layer;
import org.geotools.data.ows.WMSCapabilities;
import org.geotools.data.wms.Specification;
import org.geotools.data.wms.WMS1_0_0;
import org.geotools.data.wms.WMS1_1_0;
import org.geotools.data.wms.WMS1_1_1;
import org.geotools.data.wms.WMSUtils;
import org.geotools.data.wms.WebMapServer;
import org.geotools.data.wms.request.GetCapabilitiesRequest;
import org.geotools.data.wms.request.GetFeatureInfoRequest;
import org.geotools.data.wms.request.GetMapRequest;
import org.geotools.data.wms.response.GetCapabilitiesResponse;
import org.geotools.data.wms.response.GetFeatureInfoResponse;
import org.geotools.data.wms.response.GetMapResponse;
import org.geotools.data.wms.xml.WMSSchema;
import org.geotools.ows.ServiceException;
import org.xml.sax.SAXException;

/**
 * Connect to a WMS.
 * 
 * @author David Zwiers, Refractions Research
 * @since 0.6
 */
public class WMSServiceImpl extends IService {

    /**
     * <code>WMS_URL_KEY</code> field
     * Magic param key for Catalog WMS persistence.
     */
    public static final String WMS_URL_KEY = "net.refractions.udig.catalog.internal.wms.WMSServiceImpl.WMS_URL_KEY"; //$NON-NLS-1$
    public static final String WMS_WMS_KEY = "net.refractions.udig.catalog.internal.wms.WMSServiceImpl.WMS_WMS_KEY"; //$NON-NLS-1$

    private Map<String, Serializable> params;

    private Throwable error;
    private URL url;

    /**
     * Construct <code>WMSServiceImpl</code>.
     *
     * @param url
     * @param params
     */
    public WMSServiceImpl(URL url, Map<String,Serializable> params) {
        this.params = params;
        this.url = url;
//System.out.println("WMS "+url);
        if (params.containsKey(WMS_WMS_KEY)) {
        	Object obj = params.get(WMS_WMS_KEY);
        	
        	if (obj instanceof WebMapServer) {
        		this.wms = (WebMapServer) obj;
        	}        	
        }
    }
    
    public Status getStatus() {
        return error != null? Status.BROKEN : wms == null? Status.NOTCONNECTED : Status.CONNECTED;
    }

    /**
     * Aquire the actual geotools WebMapServer instance.
     * <p>
     * Note this method is blocking and throws an IOException to indicate such.
     * </p>
     * @param theUserIsWatching 
     * @return WebMapServer instance
     * @throws IOException 
     */
    protected WebMapServer getWMS(IProgressMonitor theUserIsWatching) throws IOException {
        if (wms == null) {
            try {
                if( theUserIsWatching != null ) {
                	String message = MessageFormat.format(Policy.bind("WMSServiceImpl.connecting.to"), new Object[] { url }); //$NON-NLS-1$
                    theUserIsWatching.beginTask(message, 100 );
                }
                URL url1 = (URL) getConnectionParams().get(WMS_URL_KEY);
                if( theUserIsWatching != null )
                    theUserIsWatching.worked( 5 );                
                wms = new CustomWMS( url1 );
                if( theUserIsWatching != null )
                    theUserIsWatching.done();
            }
            catch( IOException persived){
                error = persived;
                throw persived;
            }
            catch( Throwable nak ){
            	
                IOException broken = new IOException( 
                        MessageFormat.format(Policy.bind("WMSServiceImpl.could.not.connect"), //$NON-NLS-1$
                        new Object[] { nak.getLocalizedMessage() }));
                broken.initCause( nak );
                error = broken;                
                throw broken;                
            }
            IResolveDelta delta = new ResolveDelta( this, IResolveDelta.Kind.CHANGED );
            ((CatalogImpl)CatalogPlugin.getDefault().getLocalCatalog()).fire( new ResolveChangeEvent( this, IResolveChangeEvent.Type.POST_CHANGE, delta )  );
        }
        return wms;
    }

    private WebMapServer wms = null;
    private WMSServiceInfo info;

    public IServiceInfo getInfo(IProgressMonitor monitor) throws IOException {
        if (info == null){
            synchronized(getWMS( monitor )){
                if(info == null){
                	info = new WMSServiceInfo( monitor );
                	IResolveDelta delta = new ResolveDelta( this, IResolveDelta.Kind.CHANGED );
            		((CatalogImpl)CatalogPlugin.getDefault().getLocalCatalog()).fire( new ResolveChangeEvent( this, IResolveChangeEvent.Type.POST_CHANGE, delta )  );
                }
            }
        }
        return info;
    }

    /*
     * @see net.refractions.udig.catalog.IService#resolve(java.lang.Class, org.eclipse.core.runtime.IProgressMonitor)
     */
    public <T> T resolve( Class<T> adaptee, IProgressMonitor monitor ) throws IOException {
        if (adaptee == null) {
            return null;
        }
        
        if (adaptee.isAssignableFrom(IServiceInfo.class)) {
            return adaptee.cast( getInfo(monitor));
        }
        
        if (adaptee.isAssignableFrom(List.class)) {
            return adaptee.cast( members(monitor));
        }
        
        if (adaptee.isAssignableFrom(WebMapServer.class)) {
            return adaptee.cast( getWMS( monitor ));
        }
        
        return null;
    }

    /**
     * @see net.refractions.udig.catalog.IService#getConnectionParams()
     */
    public Map<String,Serializable> getConnectionParams() {
        return params;
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#canResolve(java.lang.Class)
     */
    public <T> boolean canResolve( Class<T> adaptee ) {
        if (adaptee == null) {
            return false;
        }
        
        if (adaptee.isAssignableFrom(WebMapServer.class) ||
                adaptee.isAssignableFrom(IServiceInfo.class) ||
                adaptee.isAssignableFrom(List.class)) {
            return true;
        }
        
        return false;
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#members(org.eclipse.core.runtime.IProgressMonitor)
     */
    public List<WMSGeoResourceImpl> members( IProgressMonitor monitor ) throws IOException {

    	if(members == null){
            synchronized(getWMS( monitor )){
            	if(members == null){
            		getWMS( monitor ); // load ds
                    members = new LinkedList<WMSGeoResourceImpl>();
                    Layer[] layers = WMSUtils.getNamedLayers(getWMS( monitor ).getCapabilities());
                    /*
                     * Retrieved no layers from the WMS - something is wrong,
                     * either the WMS doesn't work, or it has no named layers.
                     */
                    if (layers != null && layers.length > 0) {
                        members.add(new WMSGeoResourceImpl(this, layers[0]));                        
                        
                        for( int i = 0; i < layers.length; i++ ) {
                            
                             // suppress layers that have children 
                             // TODO some people might not like this behavior
                             //      maybe we should make a preference for it.
                             // TODO should add hasChildren() to geotools                             
                            if (layers[i].getChildren().length == 0) {
                                Layer layer = layers[i];
                                members.add(new WMSGeoResourceImpl(this, layer));
                            }
                        }
                        
                    }
            	}
            }
    	}
        return members;
    }
    private List<WMSGeoResourceImpl> members = null;

    /*
     * @see net.refractions.udig.catalog.IResolve#getMessage()
     */
    public Throwable getMessage() {
        return error;
    }

    /*
     * @see net.refractions.udig.catalog.IResolve#getIdentifier()
     */
    public URL getIdentifier() {
        return url;
    }

    class WMSServiceInfo extends IServiceInfo {
        WMSServiceInfo( IProgressMonitor monitor) {
            try {
                caps = getWMS( monitor ).getCapabilities();
            } catch (Throwable t) {
                t.printStackTrace();
                caps = null;
            }
            

            keywords = caps == null ? null : caps.getService() == null ? null : caps
                    .getService().getKeywordList();
    
            String[] t;
            if (keywords == null) {
                t = new String[2];
            } else {
                t = new String[keywords.length + 2];
                System.arraycopy(keywords, 0, t, 2, keywords.length);
            }
            t[0] = "WMS"; //$NON-NLS-1$
            t[1] = getIdentifier().toString();
            keywords = t;
        }
        private WMSCapabilities caps = null;
           
        public String getAbstract() {
            return caps == null ? null : caps.getService() == null ? null : caps
                    .getService().get_abstract();
        }
        /*
         * @see net.refractions.udig.catalog.IServiceInfo#getIcon()
         */
        public ImageDescriptor getIcon() {
            //return CatalogUIPlugin.getDefault().getImages().getImageDescriptor( ISharedImages.WMS_OBJ );
            return AbstractUIPlugin.imageDescriptorFromPlugin( WmsPlugin.ID, "icons/obj16/wms_obj.gif" ); //$NON-NLS-1$
        }
        public String getDescription() {
            return getIdentifier().toString();
        }
    
        public URI getSchema() {
            return WMSSchema.NAMESPACE;
        }
    
        public URL getSource() {
            return getIdentifier();
        }
    
        public String getTitle() {
            return (caps == null || caps.getService() == null) ? (getIdentifier() == null
                    ? Policy.bind("WMSServiceImpl.broken") //$NON-NLS-1$
                    : getIdentifier().toString()) : caps.getService().getTitle();
        }
    }
    
	public class CustomWMS extends WebMapServer {

		/**
		 * @throws SAXException 
		 * @throws ServiceException 
		 * @param serverURL
		 * @throws IOException
		 */
		public CustomWMS(URL serverURL) throws IOException, ServiceException, SAXException {
			super(serverURL);
            if( WmsPlugin.isDebugging( REQUEST ) )
                    System.out.println("Connection to WMS located at: "+serverURL); //$NON-NLS-1$
            if (getCapabilities() == null) {
                throw new IOException("Unable to parse capabilities document."); //$NON-NLS-1$
            }
		}
		
		public GetCapabilitiesResponse issueRequest( GetCapabilitiesRequest arg0 ) throws IOException, ServiceException, SAXException {
            if( WmsPlugin.isDebugging( REQUEST ) )
                System.out.println("Issuing GetCapabilities request: "+arg0.getFinalURL()); //$NON-NLS-1$
            return super.issueRequest(arg0);
        }

        public GetFeatureInfoResponse issueRequest( GetFeatureInfoRequest arg0 ) throws IOException, ServiceException, SAXException {
            if( WmsPlugin.isDebugging( REQUEST ) )
                System.out.println("Issuing GetFeatureInfo request: "+arg0.getFinalURL()); //$NON-NLS-1$
            return super.issueRequest(arg0);
        }

        public GetMapResponse issueRequest( GetMapRequest arg0 ) throws IOException, ServiceException, SAXException {
            if( WmsPlugin.isDebugging( REQUEST ) )
                System.out.println("Issuing GetMap request: "+arg0.getFinalURL()); //$NON-NLS-1$
            return super.issueRequest(arg0);
        }

        protected void setupSpecifications() {
	        specs = new Specification[3];
	        specs[0] = new WMS1_0_0();
	        specs[1] = new WMS1_1_0();
	        specs[2] = new WMS1_1_1();
		}
	}
}
