
package net.refractions.udig.catalog.internal.wms.ui;
import java.io.IOException;
import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import net.refractions.udig.catalog.CatalogPlugin;
import net.refractions.udig.catalog.IResolve;
import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.internal.wms.Policy;
import net.refractions.udig.catalog.internal.wms.WMSGeoResourceImpl;
import net.refractions.udig.catalog.internal.wms.WMSServiceExtension;
import net.refractions.udig.catalog.internal.wms.WMSServiceImpl;
import net.refractions.udig.catalog.internal.wms.WmsPlugin;
import net.refractions.udig.catalog.ui.IDataWizard;
import net.refractions.udig.catalog.ui.UDIGImportPage;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.dialogs.IDialogSettings;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.IWizardPage;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.geotools.data.wms.WebMapServer;

/**
 * Data page responsible for aquiring WMS services.
 * <p>
 * Responsibilities:
 * <ul>
 * <li>defaults based on selection - for URL, WMSService, and generic IService (from search)
 * <li>remember history in dialog settings
 * <li>complete list here: <a
 * href="http://udig.refractions.net/confluence/display/DEV/UDIGImportPage+Checklist">Import Page
 * Checklist</a>
 * </ul>
 * </p>
 * <p>
 * This page is used in the Import and Add Layer wizards.
 * </p>
 * 
 * @author jgarnett
 * @since 1.0.0
 */
public class WMSWizardPage extends WizardPage implements ModifyListener, UDIGImportPage {

    final static String[] types = {"WMS", "Directory"}; //$NON-NLS-1$ //$NON-NLS-2$

    String url = ""; //$NON-NLS-1$

    private static final String WMS_WIZARD = "WMS_WIZARD"; //$NON-NLS-1$
    private static final String WMS_RECENT = "WMS_RECENT"; //$NON-NLS-1$
    private IDialogSettings settings;
    private static final int COMBO_HISTORY_LENGTH = 15;

    private Combo urlCombo;

    /**
     * Construct <code>WMSWizardPage</code>.
     * 
     * @param pageName
     */
    public WMSWizardPage() {
        super(Policy.bind("WMSWizardPage.title")); //$NON-NLS-1$

        settings = WmsPlugin.getDefault().getDialogSettings().getSection(WMS_WIZARD);
        if (settings == null) {
            settings = WmsPlugin.getDefault().getDialogSettings().addNewSection(WMS_WIZARD);
        }
    }

    public String getId() {
        return "net.refractions.udig.catalog.ui.WMS"; //$NON-NLS-1$
    }

    /**
     * Can process needs to return true *iff* we are sure this is a WMS.
     * <p>
     * This will return false for URLS like http://cnn.com/, even though that
     * is a valid url (and would show up in the url field of the wizard page.
     * This allows the wizard chooser to ask the user what "kind" of data they
     * are trying to load (choose between WMS and WFS), and then still proceed
     * with default.
     * </p>
     */
    public boolean canProcess( Object data ) {
        if( data instanceof IResolve ){
            IResolve resolve = (IResolve) data;
            return resolve.canResolve( WebMapServer.class );
        }
        return toCapabilitiesURL(data) != null;        
    }
    
    /** Can be called during createControl */
    protected Map<String,Serializable> defaultParams(){
        IStructuredSelection selection = ((IDataWizard) getWizard()).getSelection();
        return toParams( selection );
    }
    /** Retrieve "best" WMS guess of parameters based on provided context */
    protected Map<String,Serializable> toParams( IStructuredSelection context){
        if( context != null ) {
            for( Iterator itr = context.iterator(); itr.hasNext(); ) {
                Map<String,Serializable> params = toParams( itr.next() );
                if( !params.isEmpty() ) return params;
            }
        }
        return Collections.EMPTY_MAP;
    }
    /**
     * Get the connection parameters based on the provided context.
     * <p>
     * Context is often data from DnD, note this should accept anything that
     * even thinks it is a url. Compare and contrast with canProcess that must
     * only say true to urls it is sure of.
     * </p>
     * @param context Often from DnD
     * @return Map of connection parameters, or EMPTY_MAP if unknown
     */
    protected Map<String,Serializable> toParams( Object data ){
        if( data instanceof IResolve  ){
            Map params = createParams( (IResolve) data );
            if( !params.isEmpty() ) return params;            
        } 
        URL url = toCapabilitiesURL( data );
        if( url == null ){
            // so we are not sure it is a wms url
            // lets guess
            url = CatalogPlugin.locateURL(data);
        }
        if( url != null ) {
            // well we have a url - lets try it!            
            List<IResolve> list = CatalogPlugin.getDefault().getLocalCatalog().find( url, null );
            for( IResolve resolve : list ){
                Map params = createParams( resolve );
                if( !params.isEmpty() ) return params; // we got the goods!
            }
            return createParams( url );            
        }        
        return Collections.EMPTY_MAP;
    }
    static public Map<String,Serializable> createParams( IResolve handle ){
        if( handle instanceof WMSServiceImpl) {
            // got a hit!
            WMSServiceImpl wms = (WMSServiceImpl) handle;
            return wms.getConnectionParams();
        }
        else if (handle instanceof WMSGeoResourceImpl ){
            WMSGeoResourceImpl layer = (WMSGeoResourceImpl) handle;
            WMSServiceImpl wms;
            try {
                wms = layer.service( null );
                return wms.getConnectionParams();
            } catch (IOException e) {
                checkedURL( layer.getIdentifier() );
            }                    
        }
        else if( handle.canResolve( WebMapServer.class )){
            // must be some kind of handle from a search!
            return createParams( handle.getIdentifier() );
        }
        return Collections.EMPTY_MAP;
    }
    /** 'Create' params given the provided url, no magic occurs */
    static public Map<String,Serializable> createParams( URL url ){
        WMSServiceExtension factory = new WMSServiceExtension();
        Map params = factory.createParams( url );
        if( params != null) return params;
        
        Map<String,Serializable> params2 = new HashMap<String,Serializable>();
        params2.put(WMSServiceImpl.WMS_URL_KEY,url);
        return params2;
    }

    public void createControl( Composite parent ) {
        String[] recentWMSs = settings.getArray(WMS_RECENT);
        if (recentWMSs == null) {
            recentWMSs = new String[0];
        }

        GridData gridData;
        Composite composite = new Composite(parent, SWT.NULL);

        GridLayout gridLayout = new GridLayout();
        int columns = 1;
        gridLayout.numColumns = columns;
        composite.setLayout(gridLayout);

        gridData = new GridData();

        Label urlLabel = new Label(composite, SWT.NONE);
        urlLabel.setText(Policy.bind("WMSWizardPage.label.url.text")); //$NON-NLS-1$
        urlLabel.setLayoutData(gridData);

        gridData = new GridData(GridData.FILL_HORIZONTAL);
        gridData.widthHint = 400;

        // For Drag 'n Drop as well as for general selections
        // look for a url as part of the selction
        Map<String,Serializable> params = defaultParams(); // based on selection
        
        urlCombo = new Combo(composite, SWT.BORDER);
        urlCombo.setItems(recentWMSs);
        urlCombo.setVisibleItemCount(15);
        urlCombo.setLayoutData(gridData);
        
        URL selectedURL = getURL( params );
        if (selectedURL != null) {
            urlCombo.setText( selectedURL.toExternalForm() );
        } else if (url != null && url.length() != 0) {
            urlCombo.setText(url);
        } else {
            urlCombo.setText("http://"); //$NON-NLS-1$
        }
        urlCombo.addModifyListener(this);

        gridData = new GridData();

        gridData = new GridData(GridData.FILL_HORIZONTAL);

        setControl(composite);
        setPageComplete(true);
    }
    public URL getURL( Map<String,Serializable> params ){
        Object value = params.get( WMSServiceImpl.WMS_URL_KEY );
        if( value == null ) return null;
        if( value instanceof URL ) return (URL) value;
        if( value instanceof String) {
            try {
                URL url = new URL( (String) value );                
            }
            catch( MalformedURLException erp ){                
            }
        }
        return null;        
    }
    public boolean isPageComplete() {
        try {
            url = urlCombo.getText();
            new URL(url);
        } catch (MalformedURLException e) {
            return false;
        }
        return true;
    }

    public boolean canFlipToNextPage() {
        IWizardPage[] pages = getWizard().getPages();
        return isPageComplete() && !pages[pages.length - 1].equals(this);

    }

    /**
     * Double click in list, or return from url control.
     * 
     * @see org.eclipse.swt.events.SelectionListener#widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent)
     * @param e
     */
    public void widgetDefaultSelected( SelectionEvent e ) {
        e.getClass();// kill warning
        if (getWizard().canFinish()) {
            getWizard().performFinish();
        }
    }

    /**
     * This should be called using the Wizard .. job when next/finish is pressed.
     */
    public List<IResolve> getResources( IProgressMonitor monitor ) throws Exception {
        URL location = new URL(url);
        WMSServiceExtension creator = new WMSServiceExtension();

        Map<String, Serializable> params = creator.createParams(location);
        IService service = creator.createService(location, params);
        service.getInfo(monitor); // load it

        List<IResolve> servers = new ArrayList<IResolve>();
        servers.add(service);

        /*
         * Success! Store the URL in history.
         */
        saveWidgetValues();

        return servers;
    }

    public void modifyText( ModifyEvent e ) {
        try {
            url = ((Combo) e.getSource()).getText();
            new URL(url);
            setErrorMessage(null);
        } catch (MalformedURLException exception) {
            setErrorMessage(Policy.bind("WMSWizardPage.error.invalidURL")); //$NON-NLS-1$
        }
        getWizard().getContainer().updateButtons();
    }

    /**
     * Saves the widget values
     */
    private void saveWidgetValues() {
        // Update history
        if (settings != null) {
            String[] recentWMSs = settings.getArray(WMS_RECENT);
            if (recentWMSs == null) {
                recentWMSs = new String[0];
            }
            recentWMSs = addToHistory(recentWMSs, url);
            settings.put(WMS_RECENT, recentWMSs);
        }
    }

    /**
     * Adds an entry to a history, while taking care of duplicate history items and excessively long
     * histories. The assumption is made that all histories should be of length
     * <code>COMBO_HISTORY_LENGTH</code>.
     * 
     * @param history the current history
     * @param newEntry the entry to add to the history
     * @return the history with the new entry appended Stolen from
     *         org.eclipse.team.internal.ccvs.ui.wizards.ConfigurationWizardMainPage
     */
    private String[] addToHistory( String[] history, String newEntry ) {
        ArrayList<String> l = new ArrayList<String>(Arrays.asList(history));
        addToHistory(l, newEntry);
        String[] r = new String[l.size()];
        l.toArray(r);
        return r;
    }

    /**
     * Adds an entry to a history, while taking care of duplicate history items and excessively long
     * histories. The assumption is made that all histories should be of length
     * <code>COMBO_HISTORY_LENGTH</code>.
     * 
     * @param history the current history
     * @param newEntry the entry to add to the history Stolen from
     *        org.eclipse.team.internal.ccvs.ui.wizards.ConfigurationWizardMainPage
     */
    private void addToHistory( List<String> history, String newEntry ) {
        history.remove(newEntry);
        history.add(0, newEntry);

        // since only one new item was added, we can be over the limit
        // by at most one item
        if (history.size() > COMBO_HISTORY_LENGTH)
            history.remove(COMBO_HISTORY_LENGTH);
    }


    /**
     * Convert "data" to a wms capabilities url
     * <p>
     * Candidates for conversion are:
     * <ul>
     * <li>URL - from browser DnD
     * <li>URL#layer - from browser DnD
     * <li>WMSService - from catalog DnD
     * <li>WMSGeoResource - from catalog DnD
     * <li>IService - from search DnD
     * </ul>
     * </p>
     * <p>
     * No external processing should be required here, it is enough to guess and let
     * the ServiceFactory try a real connect.
     * </p>
     * @param data IService, URL, or something else
     * @return URL considered a possibility for a WMS Capabilities, or null
     */
    static URL toCapabilitiesURL( Object data ) {
        if( data instanceof IResolve ){
            return toCapabilitiesURL( (IResolve) data );
        }
        else if( data instanceof URL ){
            return toCapabilitiesURL( (URL) data );
        }
        else if( CatalogPlugin.locateURL(data) != null ){
            return toCapabilitiesURL( CatalogPlugin.locateURL(data) );
        }
        else {
            return null; // no idea what this should be
        }
    }

    static URL toCapabilitiesURL( IResolve resolve ){
        if( resolve instanceof IService ){
            return toCapabilitiesURL( (IService) resolve );
        }
        return toCapabilitiesURL( resolve.getIdentifier() );        
    }

    static URL toCapabilitiesURL( IService resolve ){
        if( resolve instanceof WMSServiceImpl ){
            return toCapabilitiesURL( (WMSServiceImpl) resolve );
        }
        return toCapabilitiesURL( resolve.getIdentifier() );        
    }

    /** No further QA checks needed - we know this one works */
    static URL toCapabilitiesURL( WMSServiceImpl wms ){
        return wms.getIdentifier();                
    }

    /** Quick sanity check to see if url is a WMS url */
    static URL toCapabilitiesURL( URL url ){
        if (url == null) return null;
    
        String path = url.getPath() == null ? null : url.getPath().toLowerCase();
        String query = url.getQuery() == null ? null : url.getQuery().toLowerCase();
        String protocol = url.getProtocol() == null ? null : url.getProtocol().toLowerCase();
    
        if (!"http".equals(protocol) //$NON-NLS-1$
                && !"https".equals(protocol)) { //$NON-NLS-1$ 
            return null;
        }
        if (query != null && query.indexOf("service=wms") != -1) { //$NON-NLS-1$
            return checkedURL( url );
        }
        if (path != null && path.indexOf("GEOSERVER/WMS") != -1) { //$NON-NLS-1$
            return checkedURL( url );
        }
        if (url.toExternalForm().indexOf("WMS") != -1) { //$NON-NLS-1$
            return checkedURL( url );
        }
        return null;
    }

    /** Check that any trailing #layer is removed from the url */
    static public URL checkedURL( URL url ){
        String check = url.toExternalForm();
        int hash = check.indexOf('#');
        if ( hash == -1 ){
            return url;            
        }
        try {
            return new URL( check.substring(0, hash ));
        } catch (MalformedURLException e) {
            return null;
        }
    }

    //
    // Land of dead code
    //
    /**
     * True if this is a *real* wms, or plays one on tv.
     * <p>
     * If this is real WMS, toWMS( Object data ) can be used to check for
     * explict parameter settings.
     * </p>
     * 
     * @param data
     * @return
     *
    private boolean isWMS( Object data ){
       if( data instanceof WMSServiceImpl ){
           return true;
       }
       URL url = toCapabilitiesURL( data );
       if( url == null ) return false;
       List<IResolve> list = CatalogPlugin.getDefault().getLocalCatalog().find( url );
       for( IResolve resolve : list ){
           if( resolve instanceof WMSServiceImpl) return true;
           if( resolve instanceof WMSGeoResourceImpl ){
               return true;
           }
       }
       return false;       
    }*/
}

