/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.catalog.internal.worldimage;

import java.io.File;
import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import net.refractions.udig.catalog.IService;
import net.refractions.udig.catalog.ServiceExtension;

import org.geotools.gce.image.WorldImageFormat;
import org.geotools.gce.image.WorldImageFormatFactory;

/**
 * Provides the interface to the catalog service extension point.
 * <p>
 * This class is responsible for ensuring that only those services that the 
 * WorldImage plugin is capable of processing are created.
 * </p>
 * @author mleslie
 * @since 0.6.0
 */
public class WorldImageServiceExtension implements ServiceExtension {
    /** <code>URL_PARAM</code> field */
    public static String URL_PARAM = "URL"; //$NON-NLS-1$

    private static WorldImageFormatFactory factory;

    /**
     * Construct <code>WorldImageServiceExtension</code>.
     *
     */
    public WorldImageServiceExtension() {
        super();
    }

    public IService createService(URL id, Map<String, Serializable> params ) {

        if (id == null) {
            if (params.containsKey(URL_PARAM)) {
                Object param = params.get(URL_PARAM);
                if (param instanceof String) {
                    try {
                        id = new URL((String) param);
                    } catch (MalformedURLException ex) {
                        return null;
                    }
                } else if (param instanceof URL) {
                    id = (URL) param;
                } else {
                    return null;
                }
            } else {
                return null;
            }
        }
        if (!canProcess(id)) {
            return null;
        }
        WorldImageServiceImpl service = 
                new WorldImageServiceImpl(id, getFactory());
        return service;
    }
    
    /**
     * Finds or creates a WorldImageFormatFactory.
     * 
     * @return Default instance of WorldImageFormatFactory
     */
    public static WorldImageFormatFactory getFactory() {
        if(factory == null) {
            factory = new WorldImageFormatFactory();
        }
        return factory;
    }

    private boolean canProcess( URL id ) {
        if(id == null) {
            return false;
        }
        String fileExt = id.getFile().substring(id.getFile().indexOf('.') + 1);
        if(fileExt.compareToIgnoreCase("PNG") != 0 &&  //$NON-NLS-1$
                fileExt.compareToIgnoreCase("GIF") != 0 && //$NON-NLS-1$
                fileExt.compareToIgnoreCase("JPG") != 0 && //$NON-NLS-1$
                fileExt.compareToIgnoreCase("JPEG") != 0 && //&& //$NON-NLS-1$
                fileExt.compareToIgnoreCase("TIF") != 0 && //$NON-NLS-1$
                fileExt.compareToIgnoreCase("TIFF") != 0) { //$NON-NLS-1$
           return false;
        }       
        String fileName=id.getFile();
        int index = fileName.indexOf('.');
        String fileRoot = fileName.substring(0, index);
        File worldFile = new File(fileRoot + ".wld"); //$NON-NLS-1$
        if(!worldFile.exists()) {
            String wldExt = WorldImageFormat.getWorldExtension(fileExt);
            File otherWorldFile = new File(fileRoot + wldExt);
            if(!otherWorldFile.exists()) {
                return false;
            }
        }
        File file = null;
        try {
            file = new File(id.getFile());
        } catch(IllegalArgumentException ex) {
            return false;
        }
        
        return (file == null) ? false : ((WorldImageFormat)getFactory().createFormat()).accepts(id);
    }


    public Map<String, Serializable> createParams( URL url ) {
    	if( !canProcess(url))
    		return null;

        
        if (url != null) {
        	Map<String, Serializable> params = new HashMap<String, Serializable>();	
            params.put(URL_PARAM, url);
            return params;
        }
        
        return null;
    }
}
