/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.catalog.rasterings;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.List;

import net.refractions.udig.catalog.IGeoResource;
import net.refractions.udig.catalog.IGeoResourceInfo;
import net.refractions.udig.catalog.IService;

import org.eclipse.core.runtime.IProgressMonitor;
import org.geotools.parameter.ParameterGroup;
import org.opengis.coverage.grid.Format;
import org.opengis.coverage.grid.GridCoverage;
import org.opengis.coverage.grid.GridCoverageReader;
import org.opengis.parameter.GeneralParameterValue;
import org.opengis.parameter.ParameterValueGroup;

/**
 * Provides a handle to a raster resource allowing the service to be lazily 
 * loaded.
 * <p>
 * This class provides functionality common to GridCoverage based resources.
 * @author mleslie
 * @since 0.6.0
 */
public abstract class AbstractRasterGeoResource extends IGeoResource {
    
        /** 
         * <code>info</code> field contains the object used to 
         * describe this resource
         */
        protected IGeoResourceInfo info;
        /** 
         * <code>service</code> field recalls the service that created 
         * this resource 
         */
        protected AbstractRasterService service;
        private GridCoverage coverage;
        private ParameterGroup readParams;
        private String name;

        /**
         * Construct <code>AbstractRasterGeoResource</code>.
         *
         * @param service The service creating this resource.
         * @param name Human readable name of this resource.
         */
        public AbstractRasterGeoResource(AbstractRasterService service, String name) {
            this.service = service;
            if(name == null) {
                name = service.getIdentifier().getFile();
                int slash = name.lastIndexOf('/');
                name = name.substring(
                        (slash == -1 && slash < name.length() - 1 
                                ? 0 : name.lastIndexOf('/')) + 1,
                        (name.lastIndexOf('.') == -1
                                ? name.length() : name . lastIndexOf('.')));
                         
            }
            this.name = name;
        }
     
        public Status getStatus() {
            return this.service.getStatus();
        }

        public Throwable getMessage() {
            return this.service.getMessage();
        }
        
        /**
         * Retrieves the parameters used to create the GridCoverageReader
         * for this resource.  This simply delegates the creation of these
         * parameters to a GridFormat.
         * 
         * @return ParameterGroup describing the GeoResource
         */
        public ParameterGroup getReadParameters() {
            if( this.readParams==null ){
                ParameterValueGroup desc =
                        ((Format)this.service.getFormat()).getReadParameters();
                if( desc == null )
                    return null;
                if(desc instanceof ParameterGroup) {
                    this.readParams = (ParameterGroup)desc;
                }
            }
            return this.readParams;
        }
        
        /**
         * Finds or creates the GridCoverage for this resource.
         * 
         * @return GridCoverage for this GeoResource
         * @throws IOException
         */
        public Object findResource() throws IOException {
            if( this.coverage==null ) {
                GridCoverageReader reader = this.service.getReader();
                ParameterGroup pvg = getReadParameters();
                List list = pvg.values();
                @SuppressWarnings("unchecked") GeneralParameterValue[] values =  //$NON-NLS-1$
                        (GeneralParameterValue[])list.toArray(new GeneralParameterValue[0]);
                this.coverage=reader.read(values);
            }
            return this.coverage;
        }
        
        public URL getIdentifier() {
            try {
                return new URL(this.service.getIdentifier().toString()+"#"+this.name); //$NON_NLS-1$ //$NON-NLS-1$
            } catch(MalformedURLException ex) {
                return this.service.getIdentifier();
            }
        }
        
        public <T>T resolve(Class<T> adaptee, IProgressMonitor monitor) throws IOException {
            if(monitor != null) 
                monitor.beginTask(Policy.bind("AbstractRasterGeoResource.resolve"), 3); //$NON-NLS-1$
            if(adaptee == null) {
                if(monitor != null)
                    monitor.done();
                return null;
            }
            if(adaptee.isAssignableFrom(GridCoverage.class)) {
                if(monitor != null) 
                    monitor.done();
                return adaptee.cast(findResource());
            }
            if(monitor != null) 
                monitor.worked(1);
            if(adaptee.isAssignableFrom(IService.class)) {
                if(monitor != null) 
                    monitor.done();
                return adaptee.cast(this.service);
            }
            if(monitor != null)
                monitor.worked(1);
            if(adaptee.isAssignableFrom(IGeoResourceInfo.class)) {
                if(monitor != null)
                    monitor.done();
                return adaptee.cast(getInfo(monitor));
            }
            if(monitor != null)
                monitor.done();
            return null;
        }
        
        public <T> boolean canResolve(Class<T> adaptee) {
            if(adaptee == null) return false;
            return(adaptee.isAssignableFrom(IGeoResourceInfo.class) ||
                    adaptee.isAssignableFrom(IService.class) || 
                    adaptee.isAssignableFrom(GridCoverage.class));
        }
        
        public abstract IGeoResourceInfo getInfo(IProgressMonitor monitor) throws IOException;
       
}
