/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.internal.ui;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import net.refractions.udig.core.internal.ExtensionPointList;
import net.refractions.udig.ui.Constants;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.ContributionItem;
import org.eclipse.jface.action.GroupMarker;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.action.ICoolBarManager;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.action.ToolBarManager;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.ToolBar;
import org.eclipse.swt.widgets.ToolItem;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.IWorkbenchWindowActionDelegate;
import org.eclipse.ui.actions.ActionFactory;
import org.eclipse.ui.actions.ContributionItemFactory;
import org.eclipse.ui.application.IActionBarConfigurer;
import org.eclipse.ui.application.IWorkbenchConfigurer;
import org.eclipse.ui.application.IWorkbenchWindowConfigurer;
import org.eclipse.ui.application.WorkbenchAdvisor;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.service.prefs.BackingStoreException;

/**
 * Public base class for configuring the UDIG workbench.
 * <p>
 * Note that the workbench advisor object is created in advance of creating the workbench. However,
 * by the time the workbench starts calling methods on this class,
 * <code>PlatformUI.getWorkbench</code> is guaranteed to have been properly initialized.
 * </p>
 * 
 * @author rgould
 * @version $Revision: 1.9 $
 */
public class UDIGWorkbenchAdvisor extends WorkbenchAdvisor {
    private static final String NEW_ACTION_ID = "net.refractions.udig.ui.newObjectAction"; //$NON-NLS-1$

    private class NewObjectActionComparator implements Comparator<IConfigurationElement> {
        /**
         * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
         */
        public int compare( IConfigurationElement arg0, IConfigurationElement arg1 ) {
            String id0 = arg0.getAttribute("id"); //$NON-NLS-1$
            if (id0.equals("net.refractions.udig.project.ui.newLayer")) //$NON-NLS-1$
                return -1;
            String id1 = arg1.getAttribute("id"); //$NON-NLS-1$
            if (id0.equals("net.refractions.udig.project.ui.newMap") && //$NON-NLS-1$
                    !id1.equals("net.refractions.udig.project.ui.newLayer")) //$NON-NLS-1$
                return -1;

            return 1;
        }

    }
    /**
     * An item that parses the IConfigurationElement data for the newObjectAction extension point.
     * 
     * @author jones
     * @since 0.6.0
     */
    private class NewItem {
        public final String id;
        public final String text;
        public final ImageDescriptor icon;
        public final IConfigurationElement element;
        private IWorkbenchWindowActionDelegate delegate;
        private IWorkbenchWindow window;
        /**
         * Construct <code>UDIGWorkbenchAdvisor.NewContribution.NewItem</code>.
         */
        public NewItem( IConfigurationElement element, IWorkbenchWindow window ) {
            this.element = element;
            text = element.getAttribute("label"); //$NON-NLS-1$
            id = element.getAttribute("id"); //$NON-NLS-1$
            String icon = element.getAttribute("icon"); //$NON-NLS-1$
            if (icon != null) {
                this.icon = AbstractUIPlugin
                        .imageDescriptorFromPlugin(element.getNamespace(), icon);
            } else
                this.icon = null;
            this.window = window;
        }
        public void runAction() {
            if (delegate == null) {
                try {
                    delegate = (IWorkbenchWindowActionDelegate) element
                            .createExecutableExtension("class"); //$NON-NLS-1$
                } catch (CoreException e) {
                    UiPlugin.log(null, e);
                }
            }
            if (delegate != null) {
                delegate.init(window);
                delegate.selectionChanged(null, new StructuredSelection());
                delegate.run(null);
            }
        }
    }

    /**
     * Contribution item that will add new actions to a drop down button
     * 
     * @author jeichar
     * @since 0.6.0
     */
    private class NewContribution extends ContributionItem {

        private ArrayList<NewItem> newItems;
        NewItem current;
        /**
         * Construct <code>UDIGWorkbenchAdvisor.NewContribution</code>.
         */
        public NewContribution( IWorkbenchWindow window ) {
            ExtensionPointList extensions = new ExtensionPointList(NEW_ACTION_ID);
            Collections.sort(extensions, new NewObjectActionComparator());
            newItems = new ArrayList<NewItem>();
            for( IConfigurationElement element : extensions ) {
                newItems.add(new NewItem(element, window));
            }
            if (newItems.size() > 0)
                current = newItems.get(0);
        }
        /**
         * @see org.eclipse.jface.action.ContributionItem#fill(org.eclipse.swt.widgets.ToolBar, int)
         */
        public void fill( final ToolBar parent, int index ) {
            ToolItem item = new ToolItem(parent, SWT.DROP_DOWN, index);
            item.setImage(Images.get(ImageConstants.NEW_WIZ));
            item.addSelectionListener(new SelectionListener(){

                protected ImageRegistry registry;
                public void widgetSelected( SelectionEvent e ) {
                    widgetDefaultSelected(e);
                }

                public void widgetDefaultSelected( SelectionEvent e ) {
                    if (e.detail == SWT.ARROW) {
                        Menu menu = new Menu(parent);
                        registry = UiPlugin.getDefault().getImageRegistry();
                        for( final NewItem newItem : newItems ) {
                            MenuItem menuItem = new MenuItem(menu, SWT.PUSH);
                            if (newItem.text != null)
                                menuItem.setText(newItem.text);
                            menuItem.setImage(getImage(newItem.id, newItem.icon));
                            menuItem.addSelectionListener(new SelectionListener(){

                                public void widgetSelected( SelectionEvent e ) {
                                    widgetDefaultSelected(e);
                                }

                                public void widgetDefaultSelected( SelectionEvent e ) {
                                    current = newItem;
                                    newItem.runAction();
                                }

                            });
                        }
                        menu.setVisible(true);
                    } else {
                        if (current != null)
                            current.runAction();
                    }
                }

                private Image getImage( String id, ImageDescriptor descriptor ) {
                    if (registry.get(id) == null)
                        registry.put(id, descriptor);
                    return registry.get(id);
                }

            });
        }

    }
    public void initialize( IWorkbenchConfigurer configurer ) {
        super.initialize(configurer);

        // make sure we always save and restore workspace state
        configurer.setSaveAndRestore(true);
    }

    /**
     * Returns the id of the perspective to use for the initial workbench window.
     * <p>
     * This method is called during startup when the workbench is creating the first new window.
     * Subclasses must implement.
     * </p>
     * <p>
     * If the {@link IWorkbenchPreferenceConstants#DEFAULT_PERSPECTIVE_ID} preference is specified,
     * it supercedes the perspective specified here.
     * </p>
     * 
     * @return MapPerspective.ID_PERSPECTIVE
     */
    public String getInitialWindowPerspectiveId() {
        return MapPerspective.ID_PERSPECTIVE;
    }

    /**
     * Performs arbitrary actions before the given workbench window is opened.
     * <p>
     * This method is called before the window's controls have been created. Clients must not call
     * this method directly (although super calls are okay). The default implementation does
     * nothing. Subclasses may override. Typical clients will use the configurer passed in to tweak
     * the workbench window in an application-specific way; however, filling the window's menu bar,
     * tool bar, and status line must be done in {@link #fillActionBars fillActionBars}, which is
     * called immediately after this method is called.
     * </p>
     * 
     * @param configurer an object for configuring the particular workbench window being opened
     */
    public void preWindowOpen( IWorkbenchWindowConfigurer configurer ) {
        super.preWindowOpen(configurer);
        configurer.setShowProgressIndicator(true);
        configurer.setInitialSize(new Point(3000, 3000));
        	
        // configurer.setShowPerspectiveBar( true );
        configurer.setShowCoolBar(true);
        configurer.setShowStatusLine(true);
        configurer.setShowFastViewBars(true);

        List<Transfer> transfers = UDIGDropTargetProcessor.getTransfers();
        for (Transfer transfer : transfers ) {
            configurer.addEditorAreaTransfer(transfer);
        }
        
        configurer.configureEditorAreaDropListener(new UDIGEditorDropListener());
    }

    @Override
    public void postWindowOpen( IWorkbenchWindowConfigurer configurer ) {
        super.postWindowOpen(configurer);
        if( Platform.getOS().equals(Platform.OS_MACOSX))
    		System.setProperty("java.awt.headless", "true");  //$NON-NLS-1$//$NON-NLS-2$;

        UiPlugin.getDefault().getOperationMenuFactory().contributeActions(configurer.getActionBarConfigurer().getMenuManager(), configurer.getWindow());
    }

    /**
     * Configures the action bars using the given action bar configurer.
     * 
     * @param window DOCUMENT ME!
     * @param configurer DOCUMENT ME!
     * @param flags DOCUMENT ME!
     */
    public void fillActionBars( IWorkbenchWindow window, IActionBarConfigurer configurer, int flags ) {
        super.fillActionBars(window, configurer, flags);
        if ((flags & FILL_MENU_BAR) != 0) {
            fillMenuBar(window, configurer);
        }
        if ((flags & FILL_COOL_BAR) != 0) {
            fillCoolBar(window, configurer);
        }
    }

    private void fillCoolBar( IWorkbenchWindow window, IActionBarConfigurer configurer ) {
        ICoolBarManager coolBar = configurer.getCoolBarManager();

        coolBar.add(createFileBar(window));

    }
    private ToolBarManager createFileBar( IWorkbenchWindow window ) {
        ToolBarManager toolbar = new ToolBarManager(SWT.FLAT);
        toolbar.add(new NewContribution(window));
        toolbar.add(new GroupMarker(IWorkbenchActionConstants.TOOLBAR_FILE));
        
        toolbar.add(ActionFactory.SAVE.create(window));
        toolbar.add(ActionFactory.SAVE_ALL.create(window));

        return toolbar;
    }
    private void fillMenuBar( IWorkbenchWindow window, IActionBarConfigurer configurer ) {
        IMenuManager menuBar = configurer.getMenuManager();
        menuBar.add(createFileMenu(window));
        menuBar.add(createEditMenu(window));
        menuBar.add(createNavigationMenu(window));
        menuBar.add(createLayerMenu());
        menuBar.add(createToolMenu());
        menuBar.add(UiPlugin.getDefault().getOperationMenuFactory().getMenu(window));
        menuBar.add(new GroupMarker(IWorkbenchActionConstants.MB_ADDITIONS));
        menuBar.add(createWindowMenu(window));
        menuBar.add(createHelpMenu(window));
    }

    private MenuManager createFileMenu( IWorkbenchWindow window ) {
        MenuManager menu = new MenuManager(Policy.bind("UDIGWorkbenchAdvisor.file"), //$NON-NLS-1$
                IWorkbenchActionConstants.M_FILE);
        menu.add(new GroupMarker(IWorkbenchActionConstants.FILE_START));

        MenuManager newMenu = new MenuManager(Policy.bind("UDIGWorkbenchAdvisor.new"), //$NON-NLS-1$
                ActionFactory.NEW.getId());

        menu.add(newMenu);
        newMenu.add(new GroupMarker("new.start")); //$NON-NLS-1$
        ExtensionPointList list = new ExtensionPointList(NEW_ACTION_ID);
        Collections.sort(list, new NewObjectActionComparator());
        for( IConfigurationElement element : list ) {
            final NewItem item = new NewItem(element, window);
            Action newAction = new Action(){
                public void runWithEvent( org.eclipse.swt.widgets.Event event ) {
                    item.runAction();
                }
            };
            newAction.setText(item.text);
            newAction.setImageDescriptor(item.icon);
            newMenu.appendToGroup("new.start", newAction); //$NON-NLS-1$
        }
        newMenu.add(ContributionItemFactory.NEW_WIZARD_SHORTLIST.create(window));

        menu.add(new GroupMarker(IWorkbenchActionConstants.OPEN_EXT));
        menu.add(new Separator());

        menu.add(new Separator());
        menu.add(new GroupMarker("close.ext")); //$NON-NLS-1$
        
        IAction close = ActionFactory.CLOSE.create(window);
        close.setText(Policy.bind("UDIGWorkbenchAdvisor.close.text")); //$NON-NLS-1$
        menu.add(close);
        
        IAction closeAll = ActionFactory.CLOSE_ALL.create(window);
        closeAll.setText(Policy.bind("UDIGWorkbenchAdvisor.closeAll.text")); //$NON-NLS-1$
        menu.add(closeAll);

        menu.add(new Separator());
        
        IAction save = ActionFactory.SAVE.create(window);
        save.setText(Policy.bind("UDIGWorkbenchAdvisor.save.text")); //$NON-NLS-1$
        menu.add(save);
        
        IAction saveAll = ActionFactory.SAVE_ALL.create(window);
        saveAll.setText(Policy.bind("UDIGWorkbenchAdvisor.saveAll.text")); //$NON-NLS-1$
        menu.add(saveAll);
        
        menu.add(new GroupMarker("commit.ext")); //$NON-NLS-1$
        menu.add(new GroupMarker("revert.ext")); //$NON-NLS-1$
        menu.add(new Separator());
        menu.add(new GroupMarker("rename.ext")); //$NON-NLS-1$
        menu.add(new Separator());
        menu.add(new GroupMarker(IWorkbenchActionConstants.MB_ADDITIONS));

        menu.add(new Separator());
        IAction _import = ActionFactory.IMPORT.create(window);
        _import.setText(Policy.bind("UDIGWorkbenchAdvisor.import.text")); //$NON-NLS-1$
        menu.add(_import);

        menu.add(new Separator());
        menu.add(new GroupMarker(Constants.CONFIG));
//        menu.add(ActionFactory.PROPERTIES.create(window));

        menu.add(new Separator());
        menu.add(new GroupMarker(IWorkbenchActionConstants.FILE_END));
        IAction exit = ActionFactory.QUIT.create(window);
        exit.setText(Policy.bind("UDIGWorkbenchAdvisor.exit.text")); //$NON-NLS-1$
        menu.add(exit);
        return menu;
    }

    private MenuManager createEditMenu( IWorkbenchWindow window ) {
        MenuManager menu = new MenuManager(Policy.bind("UDIGWorkbenchAdvisor.edit"), //$NON-NLS-1$
                IWorkbenchActionConstants.M_EDIT);
        menu.add(new GroupMarker(IWorkbenchActionConstants.EDIT_START));

//        menu.add(ActionFactory.UNDO.create(window));
//        menu.add(ActionFactory.REDO.create(window));
        menu.add(new GroupMarker(IWorkbenchActionConstants.UNDO_EXT));
        menu.add(new Separator());

//        menu.add(ActionFactory.CUT.create(window));
//        menu.add(ActionFactory.COPY.create(window));
//        menu.add(ActionFactory.PASTE.create(window));
        menu.add(new GroupMarker(IWorkbenchActionConstants.CUT_EXT));
        menu.add(new Separator());

//        menu.add(ActionFactory.DELETE.create(window));
//        menu.add(ActionFactory.SELECT_ALL.create(window));
        menu.add(new Separator());

        menu.add(new GroupMarker(IWorkbenchActionConstants.ADD_EXT));

        menu.add(new GroupMarker(IWorkbenchActionConstants.EDIT_END));
        menu.add(new Separator());
        menu.add(new GroupMarker(IWorkbenchActionConstants.MB_ADDITIONS));

        return menu;
    }

    private MenuManager createLayerMenu() {
        MenuManager menu = new MenuManager(Policy.bind("UDIGWorkbenchAdvisor.layerMenu"), "layer"); //$NON-NLS-1$ //$NON-NLS-2$
        menu.add(new GroupMarker("add.ext")); //$NON-NLS-1$
        menu.add(new Separator());
        menu.add(new GroupMarker("edit.ext")); //$NON-NLS-1$

        menu.add(new Separator());
        menu.add(new GroupMarker(IWorkbenchActionConstants.MB_ADDITIONS));

        return menu;
    }

    private MenuManager createNavigationMenu( IWorkbenchWindow window ) {
        MenuManager menu = new MenuManager(
                Policy.bind("UDIGWorkbenchAdvisor.navigationMenu"), "nav"); //$NON-NLS-1$ //$NON-NLS-2$
//        menu.add(ActionFactory.BACKWARD_HISTORY.create(window));
//        menu.add(ActionFactory.FORWARD_HISTORY.create(window));
        menu.add( new GroupMarker(IWorkbenchActionConstants.NAV_START));
        menu.add(new Separator());
        menu.add(new GroupMarker("zoom.ext")); //$NON-NLS-1$

        menu.add(new Separator());
        menu.add(new GroupMarker(IWorkbenchActionConstants.MB_ADDITIONS));

        menu.add(new Separator());
        menu.add(new GroupMarker("bottom")); //$NON-NLS-1$

        return menu;
    }

    private MenuManager createToolMenu() {
        MenuManager menu = new MenuManager(Policy.bind("UDIGWorkbenchAdvisor.tools"), "tools"); //$NON-NLS-1$ //$NON-NLS-2$
        menu.add(new GroupMarker("action.ext")); //$NON-NLS-1$
        menu.add(new Separator());
        menu.add(new GroupMarker("modal.ext")); //$NON-NLS-1$
        menu.add(new Separator());
        menu.add(new GroupMarker(IWorkbenchActionConstants.MB_ADDITIONS));

        return menu;
    }

    private MenuManager createWindowMenu( IWorkbenchWindow window ) {
        MenuManager menu = new MenuManager(Policy.bind("UDIGWorkbenchAdvisor.window"), //$NON-NLS-1$
                IWorkbenchActionConstants.M_WINDOW);

        IAction openNewWindow = ActionFactory.OPEN_NEW_WINDOW.create(window);
        openNewWindow.setText(Policy.bind("UDIGWorkbenchAdvisor.newWindow.text")); //$NON-NLS-1$
        menu.add(openNewWindow);

        menu.add(new Separator());

        MenuManager perspectiveMenu = new MenuManager(Policy
                .bind("UDIGWorkbenchAdvisor.open.perspective"), //$NON-NLS-1$
                "openPerspective"); //$NON-NLS-1$
        IContributionItem perspectiveList = ContributionItemFactory.PERSPECTIVES_SHORTLIST
                .create(window);
        perspectiveMenu.add(perspectiveList);
        menu.add(perspectiveMenu);

        MenuManager viewMenu = new MenuManager(Policy.bind("UDIGWorkbenchAdvisor.show.view")); //$NON-NLS-1$
        IContributionItem viewList = ContributionItemFactory.VIEWS_SHORTLIST.create(window);
        viewMenu.add(viewList);
        menu.add(viewMenu);

        menu.add(new Separator());
        
        IAction resetPerspective = ActionFactory.RESET_PERSPECTIVE.create(window);
        resetPerspective.setText(Policy.bind("UDIGWorkbenchAdvisor.resetPerspective.text")); //$NON-NLS-1$
        menu.add(resetPerspective);
        
        IAction closePerspective = ActionFactory.CLOSE_PERSPECTIVE.create(window);
        closePerspective.setText(Policy.bind("UDIGWorkbenchAdvisor.closePerspective.text")); //$NON-NLS-1$
        menu.add(closePerspective);
        
        IAction closeAllPerspectives = ActionFactory.CLOSE_ALL_PERSPECTIVES.create(window);
        closeAllPerspectives.setText(Policy.bind("UDIGWorkbenchAdvisor.closeAllPerspectives.text")); //$NON-NLS-1$
        menu.add(closeAllPerspectives);
        menu.add(new Separator());
        menu.add(new Separator());
        menu.add(new GroupMarker(IWorkbenchActionConstants.MB_ADDITIONS));

        IAction preferences = ActionFactory.PREFERENCES.create(window);
        preferences.setText(Policy.bind("UDIGWorkbenchAdvisor.preferences.text")); //$NON-NLS-1$
        menu.add(preferences);

        menu.add(ContributionItemFactory.OPEN_WINDOWS.create(window));

        return menu;
    }

    private MenuManager createHelpMenu( IWorkbenchWindow window ) {
        MenuManager menu = new MenuManager(Policy.bind("UDIGWorkbenchAdvisor.help"), //$NON-NLS-1$
                IWorkbenchActionConstants.M_HELP);

        IAction welcome = ActionFactory.INTRO.create(window);
        welcome.setText(Policy.bind("UDIGWorkbenchAdvisor.welcome.text")); //$NON-NLS-1$
        menu.add(welcome);
        
        IAction helpContents = ActionFactory.HELP_CONTENTS.create(window);
        helpContents.setText(Policy.bind("UDIGWorkbenchAdvisor.helpContents.text")); //$NON-NLS-1$
        menu.add(helpContents);

        // Tips and tricks page would go here
        menu.add(new GroupMarker(IWorkbenchActionConstants.HELP_START));
        menu.add(new GroupMarker(IWorkbenchActionConstants.HELP_END));
        menu.add(new Separator());
        menu.add(new GroupMarker(IWorkbenchActionConstants.MB_ADDITIONS));

        // About should always be at the bottom
        // To use the real RCP About dialog uncomment these lines
        menu.add(new Separator());
        IAction about = ActionFactory.ABOUT.create(window);
        about.setText(Policy.bind("UDIGWorkbenchAdvisor.aboutUDig.text")); //$NON-NLS-1$
        menu.add(about);
        return menu;
    }
}
