/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.project.internal.render.impl;

import java.util.ArrayList;
import java.util.Collection;

import net.refractions.udig.project.internal.Policy;
import net.refractions.udig.project.internal.ProjectPlugin;
import net.refractions.udig.project.internal.render.ExecutorVisitor;
import net.refractions.udig.project.internal.render.RenderExecutor;
import net.refractions.udig.project.internal.render.Renderer;
import net.refractions.udig.project.render.RenderException;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.emf.common.notify.Notification;

import com.vividsolutions.jts.geom.Envelope;

/**
 * An Executor specifically for executing CompositeRenderers
 * 
 * @author jeichar
 * @since 0.6.0
 */
public class CompositeRendererExecutorImpl extends RendererExecutorMultiLayer{

	int timeout = 0;

	/**
	 * Listens to the render events
	 * 
	 * @author jeichar
	 * @since 0.6.0
	 */
	protected static class CompositeRendererListener extends
			MultiLayerRendererListener {

		/**
		 * Construct <code>CompositeRendererListener</code>.
		 * 
		 * @param executor
		 */
		CompositeRendererListener(CompositeRendererExecutorImpl executor) {
			super(executor);
		}

		CompositeRendererExecutorImpl getExecutor() {
			return (CompositeRendererExecutorImpl) executor;
		}

		/**
		 * @see net.refractions.udig.project.internal.render.impl.RendererExecutorMultiLayer.MultiLayerRendererListener#stateChanged(org.eclipse.emf.common.notify.Notification)
		 */
		protected void stateChanged(Notification msg) {
			if (msg.getNewIntValue() == Renderer.RENDERING)
				getExecutor().resetTimeout();
            executor.setState(msg.getNewIntValue());
            if( msg.getNewIntValue() == Renderer.RENDER_REQUEST)
                ((CompositeRendererExecutorImpl)executor).refresh();
		}
	}

	protected static class CompositeRendererJob extends RenderJob {

		private static final int TIMEOUT = 20000;
		
		/**
		 * Construct <code>CompositeRendererJob</code>.
		 * 
		 * @param executor
		 */
		public CompositeRendererJob(CompositeRendererExecutorImpl executor) {
			super(executor);
		}

		/**
		 * @see net.refractions.udig.project.internal.render.impl.RenderExecutorImpl.RenderJob#getExecutor()
		 */
		public CompositeRendererExecutorImpl getExecutor() {
			return (CompositeRendererExecutorImpl) executor;
		}
		

		/**
		 * @see net.refractions.udig.project.internal.render.impl.RenderExecutorImpl.RenderJob#startRendering()
		 */
		protected void startRendering(Envelope bounds,IProgressMonitor monitor) throws Throwable {
			super.startRendering(bounds,monitor);
			incrementalUpdate();
		}

        /**
         * Incrementally updates the display until rendering is complete.
         * 
         * @throws InterruptedException
         * @throws RenderException
         */
        public void incrementalUpdate() throws InterruptedException, RenderException {
            getExecutor().resetTimeout();
            getExecutor().getRenderer().refreshImage();
            synchronized (getExecutor().getRenderer()) {
				while ( (getMonitor()==null || !getMonitor().isCanceled() ) && (getExecutor().getRenderer().getState() == RENDERING ||
                        getExecutor().getRenderer().getState() == RENDER_REQUEST )
						&& getExecutor().timeout < TIMEOUT) {
					getExecutor().getRenderer().wait(500);
                    getExecutor().timeout+=500;
					if (getExecutor().getState() == RENDERING) {
						getExecutor().getRenderer().refreshImage();
						getExecutor().setState(RENDERING);
					}
				}
			}

            getExecutor().getRenderer().refreshImage();
        }

		/**
		 * @see net.refractions.udig.project.internal.render.impl.RenderExecutorImpl.RenderJob#postRendering()
		 */
		protected void postRendering() {
			if (getExecutor().timeout >= TIMEOUT)
				getExecutor().stopRendering();

			super.postRendering();
		}
	}

	/**
	 * Construct <code>CompositeRendererExecutorImpl</code>.
	 */
	public CompositeRendererExecutorImpl() {
		renderJob = new CompositeRendererJob(this);	
	}

	/**
	 * @see net.refractions.udig.project.internal.render.impl.RendererExecutorMultiLayer#registerFeatureListener()
	 */
	protected void registerFeatureListener() {
		// do nothing
	}

	/**
	 * Signals that the timeout for rendering should be reset.
	 * <p>
	 * The timeout is used to determine whether a renderer has stalled
	 * </p>
	 */
	protected synchronized void resetTimeout() {
		timeout = 0;
	}

	/**
	 * @see net.refractions.udig.project.internal.render.impl.RenderExecutorImpl#getRenderer()
	 */
	public CompositeRendererImpl getRenderer() {
		return (CompositeRendererImpl) renderer;
	}

	/**
	 * @see net.refractions.udig.project.internal.render.impl.RenderExecutorImpl#setRenderer(net.refractions.udig.project.render.Renderer)
	 */
	public void setRenderer(Renderer newRenderer) {
		renderJob.setSystem(false);
		super.setRendererInternal(newRenderer);
	}

    /**
     * @see net.refractions.udig.project.internal.render.impl.RenderExecutorImpl#stopRendering()
     */
    public void stopRendering() {
        super.stopRendering();
        Collection<RenderExecutor> executors=new ArrayList<RenderExecutor>(getRenderer().getRenderExecutors());
        for( RenderExecutor executor : executors){
            executor.stopRendering();
        }
    }
    
	protected void resyncState(Renderer renderer) {

		// FIXME sync sub renderers, witness composite renderer
		//
		// for( Renderer child : parent.children() ) resyncState( child );
	}

    @Override
	protected RendererListener getRendererListener() {
		return new CompositeRendererListener(this);
	}

    @Override
    protected LayerListener getLayerListener() {
        return new LayerListener(this){
        	@Override
        	public void notifyChanged(Notification msg) {
        		// do nothingsuper.notifyChanged(msg);
        	}
        };
    }
	/**
	 * @see net.refractions.udig.project.internal.render.impl.RenderExecutorImpl#visit(net.refractions.udig.project.render.ExecutorVisitor)
	 */
	public void visit(ExecutorVisitor visitor) {
		visitor.visit(this);
	}

	Job refreshJob;
    /**
     * Continually refresh the display until all renderers are done.  
     */
    public void refresh(){
    	if( refreshJob==null ){
    		refreshJob=new Job(Policy.bind("CompositeRendererExecutorImpl.0")){ //$NON-NLS-1$
    	        @Override
    	        protected IStatus run( IProgressMonitor monitor ) {
    	            try {
    	                ((CompositeRendererJob) renderJob).incrementalUpdate();
    	            } catch (Exception e) {
    	                // log error
    	                ProjectPlugin.log( null, e);
    	            }
    	            return Status.OK_STATUS;
    	        }

    	        };
    	        refreshJob.setSystem(true);
    	        refreshJob.setPriority(Job.LONG);
    	}
        refreshJob.schedule();
    }
    
}
