/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.project.render;

import net.refractions.udig.project.ILayer;
import net.refractions.udig.project.internal.render.Renderer;

/**
 * A RenderMetrics object is used to calculate estimated metrics for a renderer while rendering a
 * particular layer. RenderMetrics objects are used to decide which renderer to instantiate to
 * render a layer. As the state of udig changes different renderers may be more suited so the
 * RenderMetrics object are also used to judge whether a new renderer should be used for rendering a
 * layer.
 * <p>
 * The changed() method is used when some udig state changes, for example the zoom level changes or
 * the bounding box changes. Etc...
 * </p>
 * <p>
 * Note: These are active objects.
 * </p>
 * 
 * @author Jesse Eichar
 * @version $Revision: 1.9 $
 */
public interface IRenderMetrics {
    /**
     * Creates a new Renderer that is represented by this Metrics object
     * 
     * @return a new Renderer that is represented by this Metrics object
     * 
     */
    public Renderer createRenderer();

    /**
     * Measure of the accuracy of the rendering job in meters.
     * </p>
     * <p>
     * Examples:
     * <ul>
     * <li>0.05 would location would be within 5 cm of intended
     * <li>5.0 would indicate 5 m
     * </ul>
     * </p>
     * <p>
     * When playing with pixels lets assume we are talking about the sample
     * at the center of the pixel. Reprojecting the raster would
     * obviously compound this error, almost stuck with 1/2 the diagonal of
     * a pixel as an accuracy measure. Note: To indicate that the renderer 
     * can not perform the job at all, return a value less then zero.
     * </p>
     *
     * @return an estimate of the accuracy of the rendering job.
     */ 
    //public float accuracy();

    /**
     * Returns an estimate of how long the wait is from when the request is made to when the request
     * <b>begins</b> to be fullfilled
     * 
     * @return an estimate of how long the wait is from when the request is made to when the request
     *         <b>begins</b> to be fullfilled
     */
    //public float latency();

    /**
     * Returns an estimate of how much data can be served per second. For example number of features
     * per second.
     * 
     * @return an estimate of how much data can be served per second
     */
    //public float throughput();

    /**
     * Returns an estimate of the time in seconds it takes for an response from request until the
     * request if satisfied
     * 
     * @return an estimate of the time it take for an response from request until the request if
     *         satisfied (in seconds)
     */
    //public float responseTime();

    /** Appearance considered 
    static final float PERFECT = 1.0f;
    static final float MUD = 0.0f;
    /**
     * Percentile score for rendering appearance, 100 would be perfect.
     * <p>
     * Some quick guidelines:
     * <ul>
     * <li>Don't exceed 30% when using a "defaults" not provided by the user
     * <li>Perfectionists: Use a 100% to force the system (like with a Scalebar)
     * <li>Generalists: limit yourself to 90% when everything (in say a SLD) matches your
     * abilities. (leave room for a perfectionist to step in)
     * <li>Generalists: stay under 60% when you find content exists you don't understand. Imagine
     * <i>FastLineStringRenderer</i> that checks for LineSymbolizers, a drop to below would be in
     * order 60% if TextSymbolizers that it cannot handle are around. (But don't give up, maybe
     * nobody handles TextSymbolizers yet).
     * <li>0% indicates utter and complete incompetence, basically canStyle()==false
     * </ul>
     * </p>
     * 
     * @return score in percent for style black board conformance.
     */
    //public float appearance();

    /**
     * 
     * Used by renderers to update performance metrics. 
     * 
     * <p>
     * Picture this:<pre><code>
     * 
     * | request   | response                             | done
     * +--wait-----+----------------duration--------------+
     * | latency   | 
     * |                  response time                   |
     * 
     * and throughtput = amount / response time
     * </code></pre>
     * 
     * @param wait How long it has taken to start the rendering job. Measured from 
     * when the rendering request is made.
     * @param duration How long it has taken to render the data to present. Measured 
     * from when the first bit of data has been rendered.
     * @param amount How much data has been rendered. For instance a number of 
     * features or pixels.
     */
    //public void update( long wait, long duration, int amount );

    /**
     * Processes notifications of a change in the render environment.
     * <p>
     * For example, if the zoom level of the viewport has changed, this method will be called with a
     * ViewportEvent which contains the old and new zoom values.
     * </p>
     * 
     * @param event The notification of an event. Subclass determines what action should be
     *        performed.
     * @return true if the metrics have changed as a result of processing the event
     */
    //public boolean changed( DeltaEvent event );

    /**
     * @return the RenderContext that this RenderMetrics is valid for.
     * @see IRenderContext
     */
    public IRenderContext getRenderContext();
    /**
     * Returns the IRenderMetricsFactory that created this object.
     * 
     * @return the IRenderMetricsFactory that created this object.
     * @see IRenderMetricsFactory
     */
    public IRenderMetricsFactory getRenderMetricsFactory();
    
    /**
     * True if renderer use the provided style information.
     * <p>
     * The style information should be considred with respect to
     * the current getRenderContext().
     * </p>
     * @param SyleID
     * @param value
     * @return true if renderer can use the provided style information
     */
    public boolean canStyle( String styleID, Object value );    
    /**
     * Check to see if this layers can be added to rendered.
     * <p>
     * This is only called for renderers that are MultiLayer renderers.
     * </p>
     */
    public boolean canAddLayer(ILayer layer);
}
