/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.project.render;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.geom.AffineTransform;

import net.refractions.udig.project.IMap;
import net.refractions.udig.project.internal.render.impl.ViewportModelImpl;

import org.opengis.referencing.crs.CoordinateReferenceSystem;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;

/**
 * Models the Viewport on the map.
 * 
 * @author Jesse
 * @since 0.5
 */
public interface IViewportModel {

	/**
	 * A Default setting to use for the viewport CRS.
	 */
	public static final CoordinateReferenceSystem DEFAULT_CRS = ViewportModelImpl.DEFAULT_CRS;

	/**
	 * Returns the local coordinate system. The local coordinate system is the CRS that all the
	 * layer data will be transformed into. Once the layer data is transformed into the local CRS
	 * then it is transformed for display onto the screen
	 * 
	 * @return the local coordinate system
	 * @see CoordinateReferenceSystem
	 */
	public CoordinateReferenceSystem getCRS();

	/**
	 * Returns the bounding box of the Viewport in world coordinates.
	 * <p>
	 * Note: Since Envelope is not a UDIG element changes to the bounds envelope object will not
	 * raise events. Therefore the bounds should only be modified via the ViewportModel interface
	 * </p>
	 * 
	 * The bounds are in the same CRS as returned by {@linkplain #getCRS()}
	 * 
	 * @return the bounding box of the Viewport in world coordinates.
	 * @see Envelope
	 */
	public Envelope getBounds(); // TODO: THIS MUST BE A REFERENCED ENVELOPE!

	/**
	 * Returns the center of the viewport in world coordinates.
	 * 
     * The bounds are in the same CRS as returned by {@linkplain #getCRS()}
	 * 
	 * @return the center of the viewport in world coordinates
	 * @see Coordinate
	 */
	public Coordinate getCenter();

	/**
	 * Returns the Viewport's height in world coordinates.
     * The bounds are in the same CRS as returned by {@linkplain #getCRS()}
	 * 
	 * @return the Viewport's height in world coordinates.
	 */
	public double getHeight();

	/**
	 * Returns the Viewport's width in world coordinates.
     * The bounds are in the same CRS as returned by {@linkplain #getCRS()}
	 * 
	 * @return the Viewport's width in world coordinates.
	 */
	public double getWidth();

	/**
	 * Returns the aspect ratio of the viewport.
	 * 
	 * @return The aspect ratio of the viewport.
	 */
	public double getAspectRatio();

	/**
	 * Gets the Map that contains the current ViewportModel
	 * 
	 * @return the Map that contains the current ViewportModel
	 */
	public IMap getMap();

	/**
	 * Gets up the affine transform that will transform from the world to screen. A convenience
	 * method.  
	 * 
	 * @return a transform that maps from real world coordinates to the screen
	 * @see AffineTransform
	 */
	public AffineTransform worldToScreenTransform();

	/**
	 * Returns the pixel on the screen for a given coordinate in world space.
	 * 
	 * @param coord A coordinate in world space.
	 * @return The pixel on the screen that the world coordinate is drawn on.
	 * @see Point
	 * @see Coordinate
	 */
	public Point worldToPixel(Coordinate coord);
    /**
     * Gets up the affine transform that will transform from the world to the display of size destination. 
     * A convenience method. This method is independent of the CRS.
     * 
     * @return a transform that maps from real world coordinates to the screen
     */
    public AffineTransform worldToScreenTransform(Envelope mapExtent,
            Dimension destination);
	/**
	 * Converts a coordinate expressed on the device space back to real world coordinates
	 * 
	 * @param x horizontal coordinate on device space
	 * @param y vertical coordinate on device space
	 * @return The correspondent real world coordinate
	 * @see Coordinate
	 */
	public Coordinate pixelToWorld(int x, int y);

	/**
	 * Returns the size of a pixel in world units.
	 * 
	 * @return the size of a pixel in world units.
	 * @see Coordinate
	 */
	public Coordinate getPixelSize();

	 /**
	  * Find the scale denominator of the map.
	  *   Method:
	  *    1. find the diagonal distance (meters)
	  *    2. find the diagonal distance (pixels)
	  *    3. find the diagonal distance (meters) -- use DPI
	  *    4. calculate scale (#1/#2)
	  * 
	  *   NOTE: return the scale denominator not the actual scale (1/scale = denominator)
	  * 
	  * TODO:  (SLD spec page 28):
	  * Since it is common to integrate the output of multiple servers into a single displayed result in the 
	  * web-mapping environment, it is important that different map servers have consistent behaviour with respect to 
	  * processing scales, so that all of the independent servers will select or deselect rules at the same scales.
	  * To insure consistent behaviour, scales relative to coordinate spaces must be handled consistently between map 
	  * servers. For geographic coordinate systems, which use angular units, the angular coverage of a map should be 
	  * converted to linear units for computation of scale by using the circumference of the Earth at the equator and 
	  * by assuming perfectly square linear units. For linear coordinate systems, the size of the coordinate space 
	  * should be used directly without compensating for distortions in it with respect to the shape of the real Earth.
	  * 
	  * NOTE: we are actually doing a a much more exact calculation, and accounting for non-square pixels (which are allowed in WMS)
	  * 
	  * @return the scale denominator of the map on the current display.  Return -1 if something is wrong.  For example the display has not
	  * yet been created.
	  */
	public double getScaleDenominator();
}
