/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.project.ui.internal.properties;

import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import net.refractions.udig.project.command.EditCommand;
import net.refractions.udig.project.command.EditCommandFactory;
import net.refractions.udig.project.internal.Layer;
import net.refractions.udig.project.ui.internal.MapEditor;
import net.refractions.udig.project.ui.internal.Policy;

import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertySource2;
import org.eclipse.ui.views.properties.PropertyDescriptor;
import org.geotools.data.FeatureStore;
import org.geotools.feature.AttributeType;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureType;
import org.geotools.filter.Filter;
import org.geotools.filter.FilterFactory;
import org.opengis.util.CodeList;

import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;

/**
 * An adapter that allows features to act as a property source for a property
 * sheet. The sheet allows the victim's attributes to be editted and viewed.
 * 
 * @author jeichar
 * @since 0.3
 */
public class FeaturePropertySource implements IPropertySource2 {
    private final String ID = "ID"; //$NON-NLS-1$
    private final String DEFAULT_GEOM = "DEFAULT_GEOM"; //$NON-NLS-1$
    private static final String BOUNDING_BOX = "BOUNDING_BOX"; //$NON-NLS-1$
    private static final String FEATURE = "FEATURE"; //$NON-NLS-1$
    private Feature feature = null;
    private Feature old = null;
    private Map geomProperties = new HashMap();
    private Map attrProperties = new HashMap();
    private AttributeType[] attrs;
    private IPropertyDescriptor[] descriptors;
    private boolean attribute;

    /**
     * Creates a new instance of FeaturePropertySource
     * 
     * @param feature The feature that this property source refers to.
     */
    public FeaturePropertySource(Feature feature) {
        this(feature, false);
    }
    /**
     * Creates a new instance of FeaturePropertySource
     * 
     * @param feature2
     * @param attribute
     */
    public FeaturePropertySource(Feature feature2, boolean attribute) {
        this.feature = feature2;
        try {
            if(feature2!=null)
                old = feature.getFeatureType().duplicate(feature);
        } catch (Exception e) {
            e.printStackTrace();
        }
        this.attribute = attribute;
    }
    /**
     * @see org.eclipse.ui.views.properties.IPropertySource#getEditableValue()
     */
    public Object getEditableValue() {
        return ""; //$NON-NLS-1$
    }
    /**
     * @see org.eclipse.ui.views.properties.IPropertySource#getPropertyDescriptors()
     */
    public IPropertyDescriptor[] getPropertyDescriptors() {
        if (descriptors == null) {
            boolean hasAttrs=false;
            List descrps = new LinkedList();
            PropertyDescriptor d = new PropertyDescriptor(ID, "ID"); //$NON-NLS-1$
            d.setCategory(Policy.bind("FeaturePropertySource.feature")); //$NON-NLS-1$
            descrps.add(d);
            d = new GeometryPropertyDescriptor(DEFAULT_GEOM, Policy.bind("FeaturePropertySource.defaultGeometry")); //$NON-NLS-1$
            d.setCategory(Policy.bind("FeaturePropertySource.geometries")); //$NON-NLS-1$
            descrps.add(d);
            d = new PropertyDescriptor(BOUNDING_BOX, Policy.bind("FeaturePropertySource.bounds")); //$NON-NLS-1$
            d.setCategory(Policy.bind("FeaturePropertySource.feature")); //$NON-NLS-1$
            d.setLabelProvider(new LabelProvider() {
                /**
                 * @see org.eclipse.jface.viewers.LabelProvider#getText(java.lang.Object)
                 */
                public String getText(Object element) {
                    Envelope bbox = (Envelope) element;
                    String minx = String.valueOf(bbox.getMinX());
                    minx = minx.substring(0, Math.min(10, minx.length()));
                    String maxx = String.valueOf(bbox.getMaxX());
                    maxx = maxx.substring(0, Math.min(10, maxx.length()));
                    String miny = String.valueOf(bbox.getMinY());
                    miny = miny.substring(0, Math.min(10, miny.length()));
                    String maxy = String.valueOf(bbox.getMaxY());
                    maxy = maxy.substring(0, Math.min(10, maxy.length()));
                    return "(" + minx + "," + miny + "), (" + maxx + "," + maxy + ")"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
                }
            });
            descrps.add(d);
            if (!attribute) {
                FeatureType ft = feature.getFeatureType();
                attrs = ft.getAttributeTypes();
                for (int i = 0; i < attrs.length; i++) {
                    AttributeType at = attrs[i];
                    String name = at.getName().toLowerCase();
                    name = name.substring(0, 1).toUpperCase() + name.substring(1);
                    if (at.isGeometry()) {
                        if (feature.getAttribute(i) != feature.getDefaultGeometry()) {
                            d = new GeometryPropertyDescriptor(new Integer(i), name+Policy.bind("FeaturePropertySource.geometry")); //$NON-NLS-1$
                            d.setCategory(Policy.bind("FeaturePropertySource.geometries")); //$NON-NLS-1$
                            descrps.add(d);
                        }
                    } else {
                        if (Feature.class.isAssignableFrom(at.getType())) {
                            d = new PropertyDescriptor(FEATURE + new Integer(i), name);
                        } else if (Collection.class.isAssignableFrom(at.getType()))
                            d = new PropertyDescriptor(new Integer(i), name);
                        else {
                            d=new AttributePropertyDescriptor(new Integer(i), name, at);
//                            if (String.class.isAssignableFrom(at.getType()))
//                                d = new TextPropertyDescriptor(new Integer(i), name);
//                            if (Integer.class.isAssignableFrom(at.getType()))
//                                d = new TextPropertyDescriptor(new Integer(i), name);
//                            if (Double.class.isAssignableFrom(at.getType()))
//                                d = new TextPropertyDescriptor(new Integer(i), name);
//                            if (Boolean.class.isAssignableFrom(at.getType()))
//                                d = new ComboBoxPropertyDescriptor(new Integer(i), name,
//                                                new String[]{"true","false"});
//                            if (Float.class.isAssignableFrom(at.getType()))
//                                d = new TextPropertyDescriptor(new Integer(i), name);
                        }
//                        d.setValidator(new AttributeValidator(at));
                        
                        if( name.equalsIgnoreCase("name") ){ //$NON-NLS-1$
                            d.setCategory(Policy.bind("FeaturePropertySource.feature")); //$NON-NLS-1$
                            descrps.add(0,d);
                        }
                        else{
                            hasAttrs=true;
                            d.setCategory(Policy.bind("FeaturePropertySource.featureAttributes")); //$NON-NLS-1$
                            descrps.add(d);
                        }
                    }
                }
            }
            if( !hasAttrs ){
                d=new PropertyDescriptor("", Policy.bind("FeaturePropertySource.noOtherAttributes")); //$NON-NLS-1$ //$NON-NLS-2$
                d.setCategory(Policy.bind("FeaturePropertySource.featureAttributes")); //$NON-NLS-1$
                descrps.add(d);
            }
            descriptors = new IPropertyDescriptor[descrps.size()];
            descrps.toArray(descriptors);
        }
        return descriptors;
    }
    /**
     * @see org.eclipse.ui.views.properties.IPropertySource#getPropertyValue(java.lang.Object)
     */
    public Object getPropertyValue(Object id) {
        if (id instanceof String) {
            String sid = (String) id;
            if (sid.equals(ID))
                return feature.getID();
            if (sid.equals(DEFAULT_GEOM))
                return getGeomProperty(feature.getDefaultGeometry());
            if (sid.equals(BOUNDING_BOX))
                return feature.getBounds();
            if (sid.startsWith(FEATURE)) {
                int i = Integer.parseInt(sid.substring(FEATURE.length()));
                return new FeaturePropertySource((Feature) feature.getAttribute(i), true);
            }
        }
        if (id instanceof Integer) {
            Integer i = (Integer) id;
            AttributeType attrType = attrs[i.intValue()];
            if (attrType.isGeometry())
                return getGeomProperty((Geometry) feature.getAttribute(i.intValue()));
            if (Collection.class.isAssignableFrom(attrType.getType()))
                return getAttrProperty(attrType, feature.getAttribute(i.intValue()));
//            return feature.getAttribute(i.intValue()).toString();
            Object attr=feature.getAttribute(i.intValue());
            
            if( !(attr instanceof Boolean) && !(attr instanceof CodeList) ){
                if( attr!=null )
                    return attr.toString();
                return ""; //$NON-NLS-1$
            }
            
            if( attr instanceof Boolean ){
                if( ((Boolean)attr).booleanValue() )
                    return new Integer(1);
                return new Integer(0);
            }
            if( attr instanceof CodeList ){
                CodeList list=(CodeList) attr;
                return new Integer(list.ordinal()).toString();
            }
        }
        return null;
    }
    /**
     * @see org.eclipse.ui.views.properties.IPropertySource#isPropertySet(java.lang.Object)
     */
    public boolean isPropertySet(Object id) {
        if (id.equals(ID))
            return feature.getID() == old.getID();
        if (id.equals(DEFAULT_GEOM))
            return feature.getDefaultGeometry() == old.getDefaultGeometry();
        if (id.equals(BOUNDING_BOX))
            return feature.getBounds() == old.getBounds();
        if (id instanceof Integer) {
            int i = ((Integer) id).intValue();
            Object attr = feature.getAttribute(i);
            if (attr instanceof String || attr instanceof Integer || attr instanceof Double
                            || attr instanceof Float || attr instanceof Boolean)
                return attr.equals(old.getAttribute(i));
        }
        return false;
    }
    /**
     * @see org.eclipse.ui.views.properties.IPropertySource#resetPropertyValue(java.lang.Object)
     */
    public void resetPropertyValue(Object id) {
        try {
            if (id instanceof Integer) {
                int i = ((Integer) id).intValue();
                Object attr = feature.getAttribute(i);
                if (attr instanceof String || attr instanceof Integer || attr instanceof Double
                                || attr instanceof Float || attr instanceof Boolean)
                    feature.setAttribute(i, old.getAttribute(i));
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
    /**
     * @see org.eclipse.ui.views.properties.IPropertySource#setPropertyValue(java.lang.Object,
     *      java.lang.Object)
     */
    public void setPropertyValue(Object id, Object value) {
        try {
            if (id instanceof Integer) {
                int i = ((Integer) id).intValue();
                Object attr = feature.getAttribute(i);
                EditCommand command=EditCommandFactory.getInstance().createSetAttributeCommand(attrs[i].getName(),value);
                getMap().sendCommandASync(command);
                if (attr instanceof String){
                    feature.setAttribute(i, value);
                }
                else if (attr instanceof Integer){
                    feature.setAttribute(i, new Integer((String) value));
                }
                else if (attr instanceof Double){
                    feature.setAttribute(i, new Double((String) value));
                }
                else if (attr instanceof Float){
                    feature.setAttribute(i, new Float((String) value));
                }
                else if (attr instanceof Boolean){
                    feature.setAttribute(i, new Boolean(((Integer) value).intValue()==0?true:false));
                }
            }
            if( value instanceof Geometry ){
                if( id.equals(DEFAULT_GEOM) )
                    feature.setDefaultGeometry((Geometry) value);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * TODO summary sentence for getMap ...
     * 
     */
    private net.refractions.udig.project.internal.Map getMap() {
        IEditorPart editor=PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActiveEditor();
        if( editor==null ){
            //TODO
            return null;
        }
        if( editor instanceof MapEditor ){
            return ((MapEditor)editor).getMap();
        }
        return null;
    }
    
    private FeatureStore getFeatureStore() throws IOException {
        IEditorPart editor=PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActiveEditor();
        if( editor==null ){
            //TODO
            return null;
        }
        if( editor instanceof MapEditor ){
            Filter filter=FilterFactory.createFilterFactory().createFidFilter(feature.getID());
            List layers=((MapEditor)editor).getMap().getContextModel().getLayers();
            for( Iterator iter = layers.iterator(); iter.hasNext(); ) {
                Layer layer = (Layer) iter.next();
                FeatureStore store=layer.getResource(FeatureStore.class, null);
                if( store!=null ){
                    if ( store.getFeatures(filter).getCount() > 0 ){
                        return store;
                    }
                }
            }
        }
        return null;
    }
    
    private Object getGeomProperty(Geometry id) {
        Object geom = geomProperties.get(id);
        if (geom == null) {
            geom = new GeomPropertySource(id);
            geomProperties.put(id, geom);
        }
        return geom;
    }
    private Object getAttrProperty(AttributeType id, Object value) {
        Object attr = geomProperties.get(id);
        if (attr == null) {
            attr = new AttributePropertySource(id, value);
            attrProperties.put(id, attr);
        }
        return attr;
    
    }
    
    /**
     * @see org.eclipse.ui.views.properties.IPropertySource2#isPropertyResettable(java.lang.Object)
     */
    public boolean isPropertyResettable(Object id) {
        return true;
    }
}