package net.refractions.udig.style.sld;

import java.util.ArrayList;
import java.util.List;

import net.refractions.udig.project.internal.Layer;
import net.refractions.udig.style.sld.simple.FillViewer;
import net.refractions.udig.style.sld.simple.LabelViewer;
import net.refractions.udig.style.sld.simple.Mode;
import net.refractions.udig.style.sld.simple.GraphicViewer;
import net.refractions.udig.style.sld.simple.StrokeViewer;
import net.refractions.udig.ui.graphics.SLDs;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.widgets.Composite;
import org.geotools.data.FeatureSource;
import org.geotools.feature.FeatureType;
import org.geotools.styling.FeatureTypeStyle;
import org.geotools.styling.Fill;
import org.geotools.styling.Graphic;
import org.geotools.styling.LineSymbolizer;
import org.geotools.styling.PointSymbolizer;
import org.geotools.styling.PolygonSymbolizer;
import org.geotools.styling.Rule;
import org.geotools.styling.Stroke;
import org.geotools.styling.Style;
import org.geotools.styling.Symbolizer;
import org.geotools.styling.TextSymbolizer;

/**
 * Defines a "simple" StyleConfigurator for working with SLD documents.
 * <p>
 * This style configurator is defined as follows:
 * 
 * <pre><code>
 *          +-+ +-------+ +------+ +------+
 *    Line: |x| | color | |size\/| |100%\/|
 *          +-+ +-------+ +------+ +------+
 *          +-+ +-------+ +------+             
 *    Fill: |x| | color | | 90%\/| 
 *          +-+ +-------+ +------+
 *          +-+ +----------------+ +------+
 *   Label: |x| |         title\/| | Font |
 *          +-+ +----------------+ +------+
 *          +-+ +-------+ +------+
 *   Point: |x| | star\/| |size\/|
 *          +-+ +-------+ +------+
 * </code></pre>
 * </p>
 * Where:
 * <ul>
 * <li>Line is used for:
 *   <br>LineString: line color, width, opacity
 *   <br>Polygon: border color, width, opacity
 *   <br>Point: border color, width, opacity
 * <li>Fill is used for Polygon or Point fill color, opacity
 * <li>Label is used to choose attribute and set font (the only dialog)
 * <li>Point is used to set the marker type and size
 * </ul>
 * </p>
 * <p>
 * Notes:
 * <ul>
 * <li>RasterSymbolizer is handled by its own thing, as is WMS etc...
 * <li>Layout as per the SLDEditorPart examples - so we can take advantage
 *     of more or less room.
 * <li>Presets is a good idea, just not here
 * <li>Apply/Revert buttons to be green/red
 * <li>Advanced (edit the SLD) can be in the view menu
 * <li>If possible replace color button with a drop down list (may not be possible)
 * </ul>
 * </P>
 * <p>
 * We will do our best to make this thing reusable on an
 * Array of Symbolizers.
 * </p>
 * @author Jody Garnett
 * @since 1.0.0
 */
public class SimpleStyleConfigurator extends AbstractSimpleConfigurator {
	StrokeViewer line = new StrokeViewer();
    FillViewer fill = new FillViewer();
    GraphicViewer point = new GraphicViewer();
    LabelViewer label = new LabelViewer();
    
    Mode mode;
            
    SelectionListener synchronize = new SelectionListener(){

        public void widgetSelected( SelectionEvent e ) {
            synchronize();
        }

        public void widgetDefaultSelected( SelectionEvent e ) {
            synchronize();
            // apply(); // and apply?
        }
        
    };
	/**
	 * Construct <code>SimpleStyleConfigurator</code>.
	 *
	 */
	public SimpleStyleConfigurator() {
		super();
        this.line.addListener( this.synchronize );
        this.fill.addListener( this.synchronize );
        this.label.addListener( this.synchronize );
        this.point.addListener( this.synchronize );
	}

	@Override
	public boolean canStyle(Layer aLayer) {
        if (aLayer.isType(FeatureSource.class))
            return true;        
        return false;
	}
	
    @Override
	protected void refresh() {
	    Style style = getSLD(); // grab an SLD or bust
        
        FeatureType schema = getLayer().getSchema();
        Stroke stroke = null;
        Fill filll = null;
        Graphic graphic = null;
        TextSymbolizer text = null;
        
        if( schema == null ){
            this.mode = Mode.NONE;
        }
        else if( SLD.isLine( schema ) ){
            this.mode = Mode.LINE;
            LineSymbolizer sym = SLDs.lineSymbolizer(style);
            stroke = SLDs.stroke(sym);
        }
        else if( SLD.isPoint( schema )){
        	this.mode = Mode.POINT;
            PointSymbolizer sym = SLDs.pointSymbolizer(style);
            stroke = SLDs.stroke(sym);
            filll = SLDs.fill(sym);
            graphic = SLDs.graphic(sym);
        }
        else if( SLD.isPolygon( schema )){
        	this.mode = Mode.POLYGON;
            PolygonSymbolizer sym = SLDs.polySymbolizer(style);
            stroke = SLDs.stroke(sym);
            filll = SLDs.fill(sym);
        }
        else {
            this.mode = Mode.ALL;
        }
        if(this.mode != Mode.NONE) {
            text = SLDs.textSymbolizer(style);
        }
        
        this.line.setStroke( stroke, this.mode );
        this.fill.setFill(filll, this.mode );
        this.point.setGraphic(graphic, this.mode );
        this.label.set(schema, text, this.mode );
	}
    
    /** Synchronize the SLD with the array of symbolizers */
    public void synchronize(){
        List<Symbolizer> aquire = new ArrayList<Symbolizer>();
        TextSymbolizer textSym = this.label.get( this.build );
        switch (this.mode) {
        case LINE:
            aquire.add(this.build.createLineSymbolizer(this.line.getStroke( this.build )));
            if(textSym != null) {
                aquire.add(textSym);
            }
            break;
        case POLYGON:
            aquire.add(this.build.createPolygonSymbolizer(
                    this.line.getStroke( this.build ), this.fill.getFill( this.build )));
            if(textSym != null) {
                aquire.add(textSym);
            }
            break;
        case POINT:
            aquire.add(this.build.createPointSymbolizer(
                    this.point.getGraphic(
                            this.fill.getFill( this.build ),
                            this.line.getStroke( this.build ),
                            this.build)));
            if(textSym != null) {
                aquire.add(textSym);
            }
            break;
        case ALL:
            aquire.add(this.build.createLineSymbolizer(this.line.getStroke(this.build)));
            aquire.add(this.build.createPolygonSymbolizer(
                    this.line.getStroke(this.build), this.fill.getFill(this.build)));
            aquire.add(this.build.createPointSymbolizer(
                    this.point.getGraphic(
                            this.fill.getFill(this.build),
                            this.line.getStroke(this.build),
                            this.build)));
            if(textSym != null) {
                aquire.add(textSym);
            }
            break;
        case NONE:
        }
           
        Symbolizer[] array = aquire.toArray( new Symbolizer[ aquire.size() ] );
        Rule rule = this.build.createRule( array );
        FeatureTypeStyle featureTypeStyle = this.build.createFeatureTypeStyle( "Feature", rule ); //$NON-NLS-1$
        Style style = this.build.createStyle();
        style.setDefault( true );
        style.setName( "simple" ); //$NON-NLS-1$
        style.setFeatureTypeStyles( new FeatureTypeStyle[]{featureTypeStyle,} );
        
        getStyleBlackboard().put(SLDContent.ID, style);        
    }
    
	@Override
	public void createControl(Composite parent) {
	    setLayout(parent);
        KeyAdapter adp = new KeyAdapter() {
            @SuppressWarnings("syntheticAccess") //$NON-NLS-1$
            public void keyReleased(KeyEvent e) {
                /*
                 * I don't like having different ways of checking for keypad enter
                 * and the normal one.  Using the keyCode would be better, but
                 * I couldn't readily find the value for CR.
                 */
                if(e.keyCode == SWT.KEYPAD_CR 
                        || e.character == SWT.CR) {
                    makeActionDoStuff();
                }
            }
        };
        this.line.createControl(parent, adp);
        this.fill.createControl(parent, adp);
        this.point.createControl(parent, adp, this.build );
        this.label.createControl(parent, adp);
	}

}
