/* uDig - User Friendly Desktop Internet GIS client
 * http://udig.refractions.net
 * (C) 2004, Refractions Research Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */
package net.refractions.udig.style.sld.simple;

import java.awt.Color;
import java.util.ArrayList;
import java.util.List;

import net.refractions.udig.style.sld.AbstractSimpleConfigurator;
import net.refractions.udig.style.sld.Policy;
import net.refractions.udig.style.sld.internal.FontEditor;
import net.refractions.udig.ui.graphics.SLDs;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.geotools.feature.AttributeType;
import org.geotools.feature.FeatureType;
import org.geotools.filter.Expression;
import org.geotools.filter.FilterFactory;
import org.geotools.filter.IllegalFilterException;
import org.geotools.styling.Fill;
import org.geotools.styling.Font;
import org.geotools.styling.StyleBuilder;
import org.geotools.styling.TextSymbolizer;

/**
 * Allows editing/viewing of a Style Layer Descriptor "Stroke".
 * <p>
 * Here is the pretty picture: <pre><code>
 *          +-+ +----------------+ +------+
 *   Label: |x| |         title\/| | Font |
 *          +-+ +----------------+ +------+
 * </code></pre>
 * </p>
 * <p>
 * Workflow:
 * <ol>
 * <li>createControl( parent ) - set up controls
 * <li>set( stroke, mode ) - provide content from SimpleStyleConfigurator
 *    <ol>
 *    <li> Symbolizer values copied into fields based on mode
 *    <li> fields copied into controls
 *    <li> controls enabled based on mode & fields
 *    </ol>
 * <li>Listener.widgetSelected/modifyText - User performs an "edit"
 * <li>Listener.sync( SelectionEvent ) - update fields with values of controls
 * <li>fire( SelectionSevent ) - notify SimpleStyleConfigurator of change
 * <li>get( StyleBuilder ) - construct based on fields
 * </ul>
 * </p>  
 * @author Jody Garnett
 * @since 1.0.0
 */
public class LabelViewer {
    boolean enabled;
    FeatureType schema;
    String labelType;
    FontData[] font;
    Color colour;
    
    Button on;
    Combo field;
    FontEditor fonter;
    
    private class Listener implements SelectionListener,ModifyListener {
        public void widgetSelected( SelectionEvent e ) {
            sync(e);
        };
        public void widgetDefaultSelected( SelectionEvent e ) {
            sync(e);
        };
        public void modifyText( final ModifyEvent e ) {
            sync(AbstractSimpleConfigurator.selectionEvent(e));
        };         
        private void sync( SelectionEvent selectionEvent ) {
            try {
                LabelViewer.this.enabled = LabelViewer.this.on.getSelection();
                LabelViewer.this.colour = LabelViewer.this.fonter.getAWTColor();
                LabelViewer.this.font = LabelViewer.this.fonter.getFontList();
                LabelViewer.this.labelType = LabelViewer.this.field.getText();
            }
            catch( Throwable t ) {
                // meh
            }
            finally {                    
                LabelViewer.this.field.setEnabled(LabelViewer.this.enabled);
                LabelViewer.this.fonter.setEnabled(LabelViewer.this.enabled);
            }
            fire( selectionEvent );
        }            
    }; 
    Listener sync = new Listener();
    
    private SelectionListener listener; 
    
    /**
     * Accepts a listener that will be notified when content changes.
     * @param listener1 
     */
    public void addListener( SelectionListener listener1 ) {
        this.listener = listener1;
    }

    /**
     * Remove listener.
     * @param listener1 
     */
    public void removeListener( SelectionListener listener1 ) {
        if (this.listener == listener1)
            this.listener = null;
    }

    /**
     * TODO summary sentence for fire ...
     * 
     * @param event
     */
    protected void fire( SelectionEvent event ) {
        if (this.listener == null)
            return;
        this.listener.widgetSelected(event);
    }
    /**
     * TODO summary sentence for get ...
     * @param build 
     * 
     * @return TextSymbolizer defined by this model
     */
    public TextSymbolizer get( StyleBuilder build ) {
        if(!this.enabled) return null;
        if(this.font == null || this.font.length == 0) return null;
        if(this.labelType == null || "".equals(this.labelType)) return null; //$NON-NLS-1$

        String fontName = this.font[0].getName();
        boolean fontBold = (this.font[0].getStyle() == SWT.BOLD);
        boolean fontItalic = (this.font[0].getStyle() == SWT.ITALIC);
        double fontSize = this.font[0].getHeight();
        Font gtFont = build.createFont(
                fontName, fontItalic, fontBold, fontSize);
        Fill filll = build.createFill(this.colour);
        
        try {
            Expression exp = FilterFactory.createFilterFactory().
                    createAttributeExpression(this.schema, this.labelType);
            return build.createTextSymbolizer(filll, new Font[] {gtFont},
                    null, exp, null, null);
        } catch (IllegalFilterException ex) {
            // TODO: some kind of reporting
            return null;
        }
        
    }
    
    /**
     * Start editing the provided symbolizer.
     * 
     * @param schema
     * @param sym
     * @param mode 
     */
    public void set(FeatureType schema, TextSymbolizer sym, Mode mode ) {
        listen(false);
        try {
            this.enabled = ( mode != Mode.NONE 
                    && sym != null);
            
            this.font = SLDs.textFont(sym);
            if(this.font == null || this.font.length == 0) {
                this.font = new FontData[] {new FontData("Arial", 12, SWT.NORMAL)}; //$NON-NLS-1$
            }
            this.labelType = SLDs.textLabelString(sym);
            this.colour = SLDs.textFontFill(sym);
            if(this.colour == null) {
                this.colour = Color.BLACK;
            }
            
            this.on.setEnabled( mode != Mode.NONE);
            this.fonter.setColorValue(this.colour);
            this.fonter.setFontList(this.font);
            
            if(schema != null) {
                AttributeType[] attrs = schema.getAttributeTypes();
                List<String> list = new ArrayList<String>();
                for(int i = 0; i < attrs.length; i++) {
                    Class cls = attrs[i].getType();
                    if(String.class.isAssignableFrom(cls)) {
                        list.add(attrs[i].getName());
                    } else if(Number.class.isAssignableFrom(cls)) {
                        list.add(attrs[i].getName());
                    }
                }
                this.field.removeAll();
                this.field.setItems(list.toArray(new String[0]));
                if(this.labelType != null) {
                    this.field.setText(this.labelType.toString());
                }
            }
            
            this.on.setSelection(this.enabled);
            this.field.setEnabled(this.enabled);
            this.fonter.setEnabled(this.enabled);
        } finally {
            listen(true);
        }
        
    }
    
    /**
     * TODO summary sentence for listen ...
     * 
     * @param listen
     */
    public void listen(boolean listen) {
        if(listen) {
            this.on.addSelectionListener(this.sync);
            this.field.addSelectionListener(this.sync);
            this.field.addModifyListener(this.sync);
            this.fonter.setListener(this.sync);
        } else {
            this.on.removeSelectionListener(this.sync);
            this.field.removeSelectionListener(this.sync);
            this.field.removeModifyListener(this.sync);
            this.fonter.clearListener();
        }
    }
    
    /**
     * TODO summary sentence for createControl ...
     * 
     * @param parent
     * @param klisten 
     * @return Generated composite
     */
    public Composite createControl(Composite parent, KeyListener klisten) {
        Composite part = AbstractSimpleConfigurator.subpart( parent, "label" ); //$NON-NLS-1$
        
        this.on = new Button( part, SWT.CHECK );
        
        this.field = new Combo(part, SWT.DROP_DOWN | SWT.READ_ONLY);
        this.field.addKeyListener(klisten);
        if(this.schema != null) {
            AttributeType[] types = this.schema.getAttributeTypes();
            String[] typeStrings = new String[types.length];
            for(int i = 0; i < types.length; i++) {
                typeStrings[i] = types[i].getName();
            }
            this.field.setItems(typeStrings);
        }
        this.field.setToolTipText(Policy.bind("LabelViewer.field.tooltip")); //$NON-NLS-1$
        
        this.fonter = new FontEditor(part);
        
        listen(true);
        
        return part;    
    }

}