package net.refractions.udig.tool.info.internal;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.refractions.udig.project.ILayer;
import net.refractions.udig.project.render.ICompositeRenderContext;
import net.refractions.udig.project.render.IRenderContext;
import net.refractions.udig.tool.info.LayerPointInfo;

import org.eclipse.core.runtime.IProgressMonitor;
import org.geotools.data.ows.Layer;
import org.geotools.data.wms.WebMapServer;
import org.geotools.data.wms.request.GetFeatureInfoRequest;
import org.geotools.data.wms.request.GetMapRequest;
import org.geotools.data.wms.response.GetFeatureInfoResponse;
import org.geotools.geometry.JTS;
import org.geotools.geometry.JTS.ReferencedEnvelope;
import org.geotools.referencing.CRS;
import org.opengis.metadata.Identifier;
import org.opengis.referencing.NoSuchAuthorityCodeException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.xml.sax.SAXException;

public class WMSDescribeLayer {
    
    /** Figures out the mapping from wms layers to udig layers */
    private Map<Layer, ILayer> getLayerMap( ICompositeRenderContext composite, IProgressMonitor monitor ) throws IOException {
        Map<Layer, ILayer> mapping=new HashMap<Layer, ILayer>();
        for( IRenderContext context: composite.getContexts() ) {
            ILayer layer = context.getLayer();
            if( context.getLayer().isVisible()
                //&& layer.isApplicable("information" )
                ) {                
                Layer wmslayer = layer.getResource( Layer.class, monitor );
                mapping.put( wmslayer, layer );
            }
        }
        return mapping;
    }
    
    public WebMapServer getWMS( ICompositeRenderContext context, IProgressMonitor monitor ) throws IOException {
        return context.getLayer().getResource(WebMapServer.class, monitor );
    }
    
    /** Get list of applicable wms layers from composite */
    private List<Layer> getLayerList(ICompositeRenderContext composite) throws IOException {
        List<Layer> layers = new ArrayList<Layer>();
        for( IRenderContext context: composite.getContexts()) {
            ILayer layer = context.getLayer();
            if( layer.isVisible()
                // && layer.isApplicable("information")
                    ) {
                Layer wmslayer = layer.getResource( org.geotools.data.ows.Layer.class, null );                
                layers.add( wmslayer );
            }            
        }    
        return layers;
    }
    
    /**
     * Implementation is forgiving codes must be in uppercase etc ... 
     *
     * @param crs CoordinateReferenceSystem
     * @param codes Set of valid vodes
     * 
     * @return Code common to both or null
     */
    private static String commonCode( CoordinateReferenceSystem crs, Set<String> codes ) {        
        // First pass based on string identity
        //
        Set<String> crsCodes = new HashSet<String>();        
        for( Identifier id : crs.getIdentifiers() ) {
            String code = id.toString();
            if( codes.contains( code ) ) return code;
            crsCodes.add( code );
        }
        // Second pass based on CoordinateReferenceSystem equality
        for( String code : codes ) {
            try {               
                CoordinateReferenceSystem check = CRS.decode( code );
                if( crs.equals( check )) {
                    // note we are trusting the code of the matched crs
                    // (because if the id provided by crs worked we would
                    // not get this far
                    //
                    return check.getIdentifiers().iterator().next().getCode();
                }                
            } catch (NoSuchAuthorityCodeException e) {
                // could not understand code
            }
        }
        // last pass do the power set lookup based on id
        //
        for( String code : codes ) {
            try {               
                CoordinateReferenceSystem check = CRS.decode( code );
                for( Identifier checkId : crs.getIdentifiers() ) {
                    String checkCode = checkId.toString();
                    if( crsCodes.contains( checkCode ) ) {
                        return checkCode;
                    }                    
                }                
            } catch (NoSuchAuthorityCodeException e) {
                // could not understand code
            }
        }        
        return null;
    }
    
    @SuppressWarnings("unchecked")
    public static List<LayerPointInfo> info( ILayer layer, ReferencedEnvelope bbox ) throws IOException {
        LayerPointInfo info = info2( layer, bbox );
        if( info == null ) return Collections.EMPTY_LIST;
        
        return Collections.singletonList( info );        
    }
    
    /**
     * Aquire info for the provided bbox
     * 
     * @param layer Must be associated with a wms layer 
     * @param bbox
     * @return LayerPointInfo object or null if information is unavailable.
     * @throws IOException 
     * @throws IOException if Response could not be understood
     */
    public static LayerPointInfo info2( ILayer layer, ReferencedEnvelope bbox ) throws IOException {        
        Layer wmslayer;
        wmslayer = layer.getResource( Layer.class, null );
        
        if( wmslayer == null ) {
            throw new IllegalArgumentException("Provided layer cannot resolve to a wms layer" );
        }
        // TODO: Fix wmslayer so we can ask who its "source" is.
        final WebMapServer wms = layer.getResource( WebMapServer.class, null );        
        if( wmslayer == null ) {
            throw new IllegalArgumentException("Provided layer cannot resolve to a wms" );
        }
        if( !wmslayer.isQueryable() ) return null;
                            
        String desiredFormat = desiredInfoFormat( wms );                
        if( desiredFormat == null ) return null;
                        
        GetMapRequest getmap = wms.createGetMapRequest();
@SuppressWarnings("unchecked")
            String code = commonCode( bbox.getCRS(), wmslayer.getSrs() );

        if( code != null ) {
            // we have a common crs! Use the bbox as is            
        }
        else {
            // lets reproject our bounding box then
            for( Iterator i= wmslayer.getSrs().iterator(); code == null && i.hasNext(); ) {
                try {
                    String tryCode = (String) i.next();
                    CoordinateReferenceSystem crs = CRS.decode( code );
                    bbox = JTS.transform( bbox, crs, true );
                    code = tryCode;
                    break;
                }
                catch (Throwable t ) {
                    // could not use tryCode
                }
            }
            if( code == null ) {
                return null; // we really can't get there from here
            }            
            try {
                CoordinateReferenceSystem crs = CRS.decode(code);                
                bbox = JTS.transform(bbox, crs, true );
            } catch (Exception e) {
                return null; // we cannot transform this
            }
        }
        
        // okay we got the bbox
        // lets fake a request to query against
        // 3x3 pixels the dimension of the bbox                
        getmap.setBBox(bbox.getMinX()+","+bbox.getMinY()+","+bbox.getMaxX()+","+bbox.getMaxY());
        getmap.setProperty( GetMapRequest.LAYERS, wmslayer.getName() );
        getmap.setDimensions( "3", "3" );
        getmap.setSRS(code);
        
        List formats = Arrays.asList(wms.getCapabilities().getRequest().getGetMap().getFormatStrings());
        if (formats.contains("image/png")) { //$NON-NLS-1$
            getmap.setProperty(GetMapRequest.FORMAT, "image/png"); //$NON-NLS-1$
        } else if (formats.contains("image/gif")) { //$NON-NLS-1$
            getmap.setProperty(GetMapRequest.FORMAT, "image/gif"); //$NON-NLS-1$
        } else if (formats.contains("image/jpeg")) { //$NON-NLS-1$
            getmap.setProperty(GetMapRequest.FORMAT, "image/jpeg"); //$NON-NLS-1$
        } else if (formats.contains("image/bmp")) { //$NON-NLS-1$
            getmap.setProperty(GetMapRequest.FORMAT, "image/bmp"); //$NON-NLS-1$
        }
        
        final GetFeatureInfoRequest request = wms.createGetFeatureInfoRequest( getmap );                       
        request.setInfoFormat( desiredFormat );
        request.setQueryPoint( 1, 1 );                
        request.setQueryLayers( Collections.singleton( wmslayer ) );
                
        LayerPointInfo info = new LayerPointInfo( layer ){
                    
            private GetFeatureInfoResponse response;
            /** Lazy request */
                    
            protected GetFeatureInfoResponse getResponse() throws IOException {
                if (this.response == null) {
                    try {
                        System.out.println(request.getFinalURL());
                        this.response = wms.issueRequest( request );
                    } catch (SAXException e) {
                        throw new IOException(
                                "Unable to parse the returned response from the server. Reason unknown."); //$NON-NLS-1$
                    }
                }
                return this.response;
            }
            public URL getRequestURL() {
                return request.getFinalURL();
            }
            /** Only aquire the value if needed */
            public String getMimeType() {
                try {
                    return getResponse().getContentType();
                } catch (IOException e) {
                    return null; // unavailable
                }
            }
            /** Only aquire the value if needed */
            public Object aquireValue() throws IOException {
                // final String CONTENT_TYPE = response.getContentType();
                // if ("text/plain".equals( CONTENT_TYPE ) ||
                // "text/html".equals( CONTENT_TYPE )) {
                String result = ""; //$NON-NLS-1$
                InputStream input = getResponse().getInputStream();
                BufferedReader reader = new BufferedReader(new InputStreamReader(input));
                String line = null;
                while( (line = reader.readLine()) != null ) {
                    result = result + line;
                }
                return result;
                // } else if ( LayerPointInfo.GML.equals( CONTENT_TYPE ) ) {
                // // parse w/ GML parser from WFS !
                // return Collections.singletonList(new LayerPointInfo( getContext().getLayer() ) {
                // protected Object aquireValue() {
                // return null; // should perform lazy parsing here
                // }
                // public String getMimeType() {
                // return LayerPointInfo.GML;
                // }
                // });
                // }
                // else {
                // return Collections.EMPTY_LIST;
                // }
            }
        };
        return info;
    }

    /**
     * This assumes the same formats are available for all layers? Hope that is a good assumption.
     *
     * @param wms
     * @return MIME type of prefered content, or null if we can't get our info on
     */
    private static String desiredInfoFormat( WebMapServer wms ) {
        String array[] =
            wms.getCapabilities().getRequest().getGetFeatureInfo().getFormatStrings();
        List formats = Arrays.asList( array );
      
        String desiredFormat;
      
        if (formats.contains("text/html")) { //$NON-NLS-1$
            desiredFormat = "text/html"; //$NON-NLS-1$
        } else if (formats.contains("text/plain")) { //$NON-NLS-1$
            desiredFormat = "text/plain"; //$NON-NLS-1$
        } else if (formats.contains(LayerPointInfo.GML)) {
            desiredFormat = LayerPointInfo.GML;
        } else {
            desiredFormat = null;
        }
        return desiredFormat;
    }
}
