/*
 *    uDig - User Friendly Desktop Internet GIS client
 *    http://udig.refractions.net
 *    (C) 2004, Refractions Research Inc.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 */
package net.refractions.udig.tool.select;

import java.util.ArrayList;
import java.util.List;

import net.refractions.udig.project.internal.ContextModel;
import net.refractions.udig.project.internal.ContextModelListenerAdapter;
import net.refractions.udig.project.internal.Layer;
import net.refractions.udig.project.internal.Map;
import net.refractions.udig.project.internal.ProjectPackage;
import net.refractions.udig.project.ui.internal.ISharedImages;
import net.refractions.udig.project.ui.internal.Images;
import net.refractions.udig.ui.ImageCache;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.ILabelProvider;
import org.eclipse.jface.viewers.ILabelProviderListener;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;

/**
 * Shows layers with selected content.
 * 
 * @author jgarnett
 * @since 0.6.0
 */
public class SelectTreeViewer extends TreeViewer implements ITreeContentProvider, ILabelProvider {
    
    private ImageCache imageCache;
    Map map;
    
    private ContextModelListenerAdapter contextListener = new ContextModelListenerAdapter() {        
        /**
         * If the notification indicates a change in the list of layers then update the structure,
         * but not labels, of the tree.
         * If the notification is not a context model change then it must be a layer change, if this is the case then
         * update the labels of the layer changed</p> 
         * @param msg The notification
         */
        public void notifyChanged( Notification msg ) {
            switch( msg.getFeatureID(Map.class) ) {
            case ProjectPackage.CONTEXT_MODEL__LAYERS:
                refresh();
                return;
            case Notification.NO_FEATURE_ID:
                refresh();
                //refresh(msg.getNotifier(), true);
                return;
            }
        }
    };
    /**
     * Construct <code>SelectTreeViewer</code>.
     *
     * @param parent
     */
    public SelectTreeViewer( Composite parent ) {
        super(parent);
        setContentProvider( this );
        setLabelProvider( this );
        imageCache = new ImageCache(); 
    }
    
    /**
     * Limited to selectable layer!
     * 
     * @see org.eclipse.jface.viewers.ITreeContentProvider#getChildren(java.lang.Object)
     */
    public Object[] getChildren( Object parent ) {
        if( parent instanceof Map ){
            Map m = (Map) parent;
            List<Layer> list = new ArrayList<Layer>();
            for( Layer layer : m.getContextModel().layers() ){
                if( layer.isSelectable() ) list.add( layer );
            }
            return list.toArray();
        }
        return null;
    }
    /*
     * @see org.eclipse.jface.viewers.ITreeContentProvider#getParent(java.lang.Object)
     */
    public Object getParent( Object element ) {
        return null;
    }
    /*
     * @see org.eclipse.jface.viewers.ITreeContentProvider#hasChildren(java.lang.Object)
     */
    public boolean hasChildren( Object element ) {
        return element == map;
    }
    /*
     * @see org.eclipse.jface.viewers.IStructuredContentProvider#getElements(java.lang.Object)
     */
    public Object[] getElements( Object inputElement ) {
        return getChildren( inputElement );
    }
    /*
     * @see org.eclipse.jface.viewers.IContentProvider#dispose()
     */
    public void dispose() {
        if( map != null){            
            map = null;            
        }
    }
    /*
     * @see org.eclipse.jface.viewers.IContentProvider#inputChanged(org.eclipse.jface.viewers.Viewer, java.lang.Object, java.lang.Object)
     */
    public void inputChanged( Viewer viewer, Object oldInput, Object newInput ) {
        // We do not change our viewer to be different from ourself.
        if( oldInput == newInput ) return;
        
        if( oldInput != null && oldInput instanceof Map){
            Map m = (Map) oldInput;
            ContextModel model = m.getContextModel();
            model.getDeepAdapters().remove( contextListener );
            map = null;
        }
        if( newInput != null && newInput instanceof ContextModel){
            Map m = (Map) newInput;
            ContextModel model = m.getContextModel();
            model.getDeepAdapters().add( contextListener );
            map = m;
        }
    }
    /*
     * @see org.eclipse.jface.viewers.ILabelProvider#getImage(java.lang.Object)
     */
    public Image getImage( Object element ) {
        if( element instanceof Layer ){
            Layer layer = (Layer) element;
            ImageDescriptor icon = layer.getGlyph();
            if( icon == null ){
                icon = Images.getDescriptor( ISharedImages.LAYER_OBJ );        
            }
            return imageCache.getImage( icon );
            
        }
        return imageCache.getMissingImage();
    }
    /*
     * @see org.eclipse.jface.viewers.ILabelProvider#getText(java.lang.Object)
     */
    public String getText( Object element ) {
        if( element instanceof Layer ){
            Layer layer = (Layer) element;
            return layer.getName();            
        }
        return "";         //$NON-NLS-1$
    }
    
    /*
     * @see org.eclipse.jface.viewers.IBaseLabelProvider#addListener(org.eclipse.jface.viewers.ILabelProviderListener)
     */
    public void addListener( ILabelProviderListener listener ) {
        // ignore
    }
    /*
     * @see org.eclipse.jface.viewers.IBaseLabelProvider#isLabelProperty(java.lang.Object, java.lang.String)
     */
    public boolean isLabelProperty( Object element, String property ) {
        return true;
    }
    /*
     * @see org.eclipse.jface.viewers.IBaseLabelProvider#removeListener(org.eclipse.jface.viewers.ILabelProviderListener)
     */
    public void removeListener( ILabelProviderListener listener ) {
        //
    }
}
