/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */

package net.refractions.udig.tools.edit;

import java.util.Iterator;

import org.geotools.data.FeatureStore;
import org.geotools.feature.Feature;
import org.geotools.geometry.JTS;
import org.opengis.referencing.operation.MathTransform;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.MultiPoint;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Polygon;

import net.refractions.udig.project.ILayer;
import net.refractions.udig.project.IMap;
import net.refractions.udig.project.internal.impl.EditManagerImpl;
import net.refractions.udig.project.ui.tool.AbstractContextMenuTool;

/**
 * A common super class that has useful functionality for all editing tools
 * @author jones
 */
public abstract class AbstractEditTool extends AbstractContextMenuTool{

    int TOP = 0x8, BOTTOM = 0x4, RIGHT = 0x2, LEFT = 0x1;
    protected static final int OVER_EDGE = -1;
    protected static final int NO_INTERSECTION = -2;
    protected boolean editing;
    Geometry lastGeom;
    Coordinate[] coords;
    
	public AbstractEditTool() {
		super(MOUSE|MOTION);
	}

	
    /**
     * Returns the geometry of the edit feature
     * 
     * @return Geometry
     */
    protected Geometry getGeometry() {
        if( getEditFeature()==null )
            return null;
        return getEditFeature().getDefaultGeometry();
    }
    
    protected Coordinate[] getCoords() {
        if( lastGeom==getGeometry()){
            return coords;
        }
        lastGeom=getGeometry();
        if(lastGeom==null )
            return null;
        coords=lastGeom.getCoordinates();
        return coords;
    }
    
    /**
     * Gets the current edit feature
     * 
     * @return Feature
     */
    protected Feature getEditFeature() {
        return getContext().getEditManager().getEditFeature();
    }
    
    protected Coordinate getTransformedClick( Coordinate click ) {

        try {
            MathTransform transform = getContext().getEditManager().getEditLayer()
                    .mapToLayerTransform();
            if (transform == null || transform.isIdentity())
                return click;
            return JTS.transform(click, new Coordinate(), transform);
        } catch (Exception e1) {
            // CorePlugin.log(ToolsPlugin.getDefault(), e1);
            return click;
        }
    }
    
    protected int overVertext( Envelope env ) {
        if( getCoords()==null )
            return NO_INTERSECTION;
        for( int i = 0; i < getCoords().length; i++ ) {
            Coordinate coord = getCoords()[i];
            if (env.contains(coord))
                return i;
        }
        return OVER_EDGE;
    }
    
    protected boolean overFeature( double xmin, double ymin, double xmax, double ymax ) {
        Coordinate coord0 = getCoords()[getCoords().length - 1];
        for( int i = 0; i < getCoords().length; i++ ) {
            Coordinate coord1 = getCoords()[i];
            if (overLine(coord0.x, coord0.y, coord1.x, coord1.y, xmin, ymin, xmax, ymax))
                return true;
            coord0 = coord1;
        }
        return false;
    }
    
    private boolean overLine( double x0, double y0, double x1, double y1, double xmin, double ymin,
            double xmax, double ymax ) {
        boolean accept = false, done = false;
        int outcode0 = compOutCode(x0, y0, xmin, xmax, ymin, ymax);
        int outcode1 = compOutCode(x1, y1, xmin, xmax, ymin, ymax);
        do {
            if (outcode0 == 0 || outcode1 == 0)
                accept = done = true;
            else if ((outcode0 & outcode1) != 0)
                done = true;
            else {
                double x, y;
                int outcodeOut = outcode0 > 0 ? outcode0 : outcode1;
                if ((outcodeOut & TOP) > 0) {
                    x = x0 + (x1 - x0) * (ymax - y0) / (y1 - y0);
                    y = ymax;
                } else if ((outcodeOut & BOTTOM) > 0) {
                    x = x0 + (x1 - x0) * (ymin - y0) / (y1 - y0);
                    y = ymin;
                } else if ((outcodeOut & RIGHT) > 0) {
                    y = y0 + (y1 - y0) * (xmax - x0) / (x1 - x0);
                    x = xmax;
                } else {
                    y = y0 + (y1 - y0) * (xmin - x0) / (x1 - x0);
                    x = xmin;
                }
                if (outcodeOut == outcode0) {
                    x0 = x;
                    y0 = y;
                    outcode0 = compOutCode(x0, y0, xmin, xmax, ymin, ymax);
                } else {
                    x1 = x;
                    y1 = y;
                    outcode1 = compOutCode(x1, y1, xmin, xmax, ymin, ymax);
                }
            }
        } while( done == false );
        return accept;
    }

    private int compOutCode( double x, double y, double xmin, double xmax, double ymin, double ymax ) {
        int outcode = 0;
        if (y > ymax)
            outcode |= TOP;
        if (y < ymin)
            outcode |= BOTTOM;
        if (x > xmax)
            outcode |= RIGHT;
        if (x < xmin)
            outcode |= LEFT;
        return outcode;
    }
    protected Coordinate closestPoint( Coordinate end1, Coordinate end2, Coordinate p ) {
        Coordinate v = new Coordinate();
        v.x = end2.x - end1.x;
        v.y = end2.y - end1.y;
        double t = ((p.x - end1.x) * v.x + (p.y - end1.y) * v.y) / (v.x * v.x + v.y * v.y);
        if (t > 1 || t < 0 || Double.isInfinite(t) || Double.isNaN(t))
            return null;
        Coordinate result = new Coordinate();
        result.x = end1.x + t * v.x;
        result.y = end1.y + t * v.y;
        return result;
    }

    int getClosest( Coordinate click, Coordinate[] coordinates ) {
        int prev = 0;
        double mindist = Double.MAX_VALUE;
        Coordinate coord1 = coordinates[coordinates.length - 1];
        for( int i = 0; i < coordinates.length; i++ ) {
            Coordinate coord2 = coordinates[i];
            Coordinate point = closestPoint(coord1, coord2, click);
            if (point == null) {
                coord1 = coord2;
                continue;
            }
            double x = click.x - point.x;
            double y = click.y - point.y;
            double dist = Math.sqrt(x * x + y * y);
            if (dist < mindist) {
                mindist = dist;
                prev = i;
            }
            coord1 = coord2;
        }
        if (getGeometry() instanceof LinearRing) {
            return (prev == 0 || prev == coordinates.length) ? coordinates.length - 2 : prev;
        }
        return prev != 0 ? prev - 1 : coordinates.length - 2;
    }

    protected ILayer findEditLayer() {
        ILayer layer = null;
        IMap map = getContext().getMap();
        //The selected layer will become the edit layer! If it's a featureStore and editable of course.
        ILayer selectedLayer = ((EditManagerImpl)map.getEditManager()).getSelectedLayer();
        if ( selectedLayer != null && selectedLayer.isType(FeatureStore.class) 
                && selectedLayer.isApplicable("editing") ) { //$NON-NLS-1$
            return selectedLayer;
        }
        //Otherwise we'll fall back to the current editLayer
        if (map.getEditManager().getEditLayer() != null)
            return map.getEditManager().getEditLayer();
        //If all else fails we'll iterate through the layer and find the first layer eligable for editing.
        //We should really never get here though.
        for( Iterator<ILayer> iter = map.getMapLayers().iterator(); iter.hasNext(); ) {
            layer = iter.next();
            if (layer.isType(FeatureStore.class) && layer.isApplicable("editing") //$NON-NLS-1$
                    && layer.isVisible())
                break;
        }
        return layer;
    }

    protected Geometry createGeometry( Geometry oldGeom, Coordinate newCoord, Coordinate prev ) {
        assert newCoord != null && prev != null && oldGeom != null;
        GeometryFactory factory = oldGeom.getFactory();
        if (oldGeom.getClass() == MultiPoint.class) {
            return factory.createMultiPoint(getCoords());
        }
        if (oldGeom.getClass() == MultiPolygon.class) {
            MultiPolygon collection = (MultiPolygon) oldGeom;
            return createMultiPolygon(newCoord, prev, factory, collection);
        }
        if (oldGeom.getClass() == MultiLineString.class) {
            MultiLineString collection = (MultiLineString) oldGeom;
            return createMultiLineString(newCoord, prev, factory, collection);
        }
        if (oldGeom.getClass() == Polygon.class) {
            Polygon polygon = (Polygon) oldGeom;
            return createPolygon(newCoord, prev, factory, polygon);
        }
        if (oldGeom.getClass() == LinearRing.class) {
            LinearRing ring = (LinearRing) oldGeom;
            return createLinearRing(newCoord, prev, factory, ring);
        }
        if (oldGeom.getClass() == LineString.class) {
            LineString line = (LineString) oldGeom;
            return createLineString(newCoord, prev, factory, line);
        }
        return null;
    }

    protected MultiLineString createMultiLineString( Coordinate newCoord, Coordinate prev,
            GeometryFactory factory, MultiLineString collection ) {
        LineString[] subgeoms = new LineString[collection.getNumGeometries()];
        for( int i = 0; i < subgeoms.length; i++ ) {
            subgeoms[i] = (LineString) createGeometry(collection.getGeometryN(i), newCoord, prev);
        }
        return factory.createMultiLineString(subgeoms);
    }

    protected LineString createLineString( Coordinate newCoord, Coordinate prev,
            GeometryFactory factory, LineString line ) {
        Coordinate[] coords1 = line.getCoordinates();
        Coordinate[] newcoords = null;
        newcoords = createNewCoordinate(newCoord, prev, coords1, newcoords);
        if (newcoords == null)
            return line;
        return factory.createLineString(newcoords);
    }

    protected Coordinate[] createNewCoordinate( Coordinate newCoord, Coordinate prev,
            Coordinate[] coords1, Coordinate[] newcoords ) {
        for( int i = 0; i < coords1.length; i++ ) {
            if (coords1[i] == prev) {
                i++;
                newcoords = new Coordinate[coords1.length + 1];
                System.arraycopy(coords1, 0, newcoords, 0, i);
                newcoords[i] = newCoord;
                if (i < coords1.length)
                    System.arraycopy(coords1, i, newcoords, i + 1, coords1.length - i);
                break;
            }
        }
        return newcoords;
    }

    protected LinearRing createLinearRing( Coordinate newCoord, Coordinate prev,
            GeometryFactory factory, LinearRing ring ) {
        Coordinate[] newcoords = null;
        newcoords = createNewCoordinate(newCoord, prev, ring.getCoordinates(), newcoords);
        if (newcoords == null)
            return ring;
        return factory.createLinearRing(newcoords);
    }

    protected Geometry createPolygon( Coordinate newCoord, Coordinate prev, GeometryFactory factory,
            Polygon polygon ) {
        LinearRing outer = (LinearRing) polygon.getExteriorRing();
        outer = (LinearRing) createGeometry(outer, newCoord, prev);
        LinearRing[] inner = new LinearRing[polygon.getNumInteriorRing()];
        for( int i = 0; i < inner.length; i++ ) {
            inner[i] = (LinearRing) createGeometry(polygon.getInteriorRingN(i), newCoord, prev);
        }
        return factory.createPolygon(outer, inner);
    }

    protected Geometry createMultiPolygon( Coordinate newCoord, Coordinate prev,
            GeometryFactory factory, MultiPolygon collection ) {
        Polygon[] subgeoms = new Polygon[collection.getNumGeometries()];
        for( int i = 0; i < subgeoms.length; i++ ) {
            subgeoms[i] = (Polygon) createGeometry(collection.getGeometryN(i), newCoord, prev);
        }
        return factory.createMultiPolygon(subgeoms);
    }
	
}
