/*
 * uDig - User Friendly Desktop Internet GIS client http://udig.refractions.net (C) 2004,
 * Refractions Research Inc. This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; version 2.1 of the License. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 */
package net.refractions.udig.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.refractions.udig.ui.graphics.AWTGraphics;
import net.refractions.udig.ui.graphics.SLDs;
import net.refractions.udig.ui.graphics.SWTGraphics;
import net.refractions.udig.ui.graphics.ViewportGraphics;

import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.geotools.data.DataUtilities;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureType;
import org.geotools.feature.IllegalAttributeException;
import org.geotools.feature.SchemaException;
import org.geotools.filter.FilterFactory;
import org.geotools.geometry.JTS;
import org.geotools.renderer.lite.LiteShape;
import org.geotools.styling.FeatureTypeStyle;
import org.geotools.styling.LineSymbolizer;
import org.geotools.styling.PointSymbolizer;
import org.geotools.styling.PolygonSymbolizer;
import org.geotools.styling.RasterSymbolizer;
import org.geotools.styling.Rule;
import org.geotools.styling.Style;
import org.geotools.styling.StyleBuilder;
import org.geotools.styling.Symbolizer;
import org.geotools.styling.TextSymbolizer;
import org.opengis.referencing.operation.MathTransform;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.MultiPoint;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;

/**
 * Drawing utitlity package - make your own previes and glyphs!
 * 
 * @author jones
 * @since 0.6.0
 */
public final class Drawing {
    private GeometryFactory gf = new GeometryFactory();
    
    private Drawing(){
        // prevent subclassing
    }
    
    /**
     * Retrieve the default Drawing implementation.
     * 
     * @return Drawing ready for use
     */
    public static Drawing create(){
        return new Drawing();
    }
    
    /**
     * Creates a ViewportGraphics object based backed by SWT. 
     * 
     * <p><b>REMEMBER to dispose of graphics.</b>
     * 
     * @param gc A SWT GC object that the ViewportGraphics object will draw on.
     * @param display The display object that will be used to create new 
     * @param displaySize
     * @return Wrapper around a normal SWT Image
     */
    public static ViewportGraphics createGraphics( GC gc, Display display, Dimension displaySize ){
        return new SWTGraphics(gc,display, displaySize );
    }
    /**
     * Creates a ViewportGraphics object based backed by SWT. 
     * 
     * @param graphics
     * @return Wrapper allowing system to draw onto j2d images
     */
    public static ViewportGraphics createGraphics( Graphics2D graphics ){
        return new AWTGraphics(graphics);
    }
   
    
    /**
     * Used to draw a freature directly onto the provided image.
     * <p>
     * Feature coordintes are in the same coordinates as the image.
     * </p>
     * <p>
     * You may call this method multiple times to draw several features onto the same
     * Image (say for glyph creation).
     * </p>
     *  
     * @param image Image to render on to 
     * @param display Needed to create Colors for image
     * @param feature Feature to be rendered
     * @param style Style to render feature with
     */
    public void drawDirect( Image image, Display display, Feature feature, Style style ){
        SWTGraphics graphics = new SWTGraphics(image, display );
		drawFeature( graphics, feature, style, new AffineTransform() );
		graphics.dispose();
    }
    
    public void drawDirect( Image image, Display display, Feature feature, 
            Rule rule ){
        AffineTransform worldToScreenTransform = new AffineTransform();
        
        SWTGraphics graphics = new SWTGraphics(image, display);
		drawFeature( graphics, feature, 
                worldToScreenTransform, false, getSymbolizers(rule), null);
		graphics.dispose();
    }
    
    public void drawFeature( ViewportGraphics graphics, Feature feature,
            AffineTransform worldToScreenTransform, boolean drawVertices, MathTransform mt ) {
            if (feature == null)
                return;
            drawFeature(graphics, feature, worldToScreenTransform, drawVertices,getSymbolizers(feature), mt);
        }
    
    public void drawFeature( ViewportGraphics graphics, Feature feature,
            AffineTransform worldToScreenTransform ) {
            if (feature == null)
                return;
            drawFeature(graphics, feature, worldToScreenTransform, false,getSymbolizers(feature), null);
        }

    public void drawFeature( ViewportGraphics graphics, Feature feature,
            AffineTransform worldToScreenTransform, Style style ) {
            if (feature == null)
                return;
            drawFeature(graphics, feature, worldToScreenTransform, false,getSymbolizers(style), null);
        }
    
    public void drawFeature( ViewportGraphics graphics, Feature feature, Style style,
            AffineTransform worldToScreenTransform) {
            if (feature == null)
                return;
            
            drawFeature(graphics, feature, worldToScreenTransform, false, getSymbolizers(style), null);
    }

    Symbolizer[] getSymbolizers(Style style){
        List<Symbolizer> symbs=new ArrayList<Symbolizer>();
        FeatureTypeStyle[] styles=style.getFeatureTypeStyles();
        for( int i = 0; i < styles.length; i++ ) {
            FeatureTypeStyle fstyle = styles[i];
            Rule[] rules=fstyle.getRules();
            for( int j = 0; j < rules.length; j++ ) {
                Rule rule = rules[j];
                symbs.addAll(Arrays.asList(rule.getSymbolizers()));
            }
        }
        return symbs.toArray(new Symbolizer[symbs.size()]);
    }
    
    Symbolizer[] getSymbolizers(Rule rule) {
        List<Symbolizer> symbs=new ArrayList<Symbolizer>();
        symbs.addAll(Arrays.asList(rule.getSymbolizers()));
        return symbs.toArray(new Symbolizer[symbs.size()]);        
    }

    public void drawFeature( ViewportGraphics graphics, Feature feature,
            AffineTransform worldToScreenTransform, boolean drawVertices, Symbolizer[] symbs, MathTransform mt  ) {

        LiteShape shape = new LiteShape(null, worldToScreenTransform, false);

        for( int m = 0; m < symbs.length; m++ ) {
            drawFeature(graphics, feature, worldToScreenTransform, drawVertices,
                    symbs[m], mt, shape);            
        }
    }
    
    public void drawFeature(ViewportGraphics graphics, Feature feature,
            AffineTransform worldToScreenTransform, boolean drawVertices,
            Symbolizer symbolizer, MathTransform mathTransform, LiteShape shape) {
        if (symbolizer instanceof RasterSymbolizer) {
            // TODO
        } else {
            Geometry g = findGeometry(feature, symbolizer);
            if (g == null)
                return;
            if ( mathTransform!=null ){
                try {
                    g=JTS.transform(g, mathTransform);
                } catch (Exception e) {
                    // do nothing
                } 
            }
            shape.setGeometry(g);

            paint(graphics, feature, shape, symbolizer);
            if (drawVertices) {
                double averageDistance = 0;
                Coordinate[] coords = g.getCoordinates();
                java.awt.Point oldP = worldToPixel(coords[0], worldToScreenTransform);
                for( int i = 1; i < coords.length; i++ ) {
                    Coordinate coord = coords[i];
                    java.awt.Point p = worldToPixel(coord, worldToScreenTransform);
                    averageDistance += p.distance(oldP) / i;
                    oldP = p;
                }
                int pixels = 1;
                if (averageDistance > 20)
                    pixels = 3;
                if (averageDistance > 60)
                    pixels = 5;
                if (pixels > 1) {
                    graphics.setColor(Color.RED);
                    for( int i = 0; i < coords.length; i++ ) {
                        Coordinate coord = coords[i];
                        java.awt.Point p = worldToPixel(coord, worldToScreenTransform);
                        graphics.fillRect(p.x - (pixels - 1) / 2, p.y - (pixels - 1) / 2,
                                pixels, pixels);
                    }
                }
            }
        }
    }

    public java.awt.Point worldToPixel( Coordinate coord, AffineTransform worldToScreenTransform ) {
        Point2D w = new Point2D.Double(coord.x, coord.y);
        AffineTransform at = worldToScreenTransform;
        Point2D p = at.transform(w, new Point2D.Double());
        return new java.awt.Point((int) p.getX(), (int) p.getY());
    }

    /** Unsure if this is the paint for the border, or the fill? */
    private void paint( ViewportGraphics g, Feature feature, LiteShape shape, Symbolizer symb ) {
         if( symb instanceof PolygonSymbolizer){
            PolygonSymbolizer polySymb = (PolygonSymbolizer) symb;
            Color stroke=SLDs.polyColor(polySymb);
            double opacity=SLDs.polyFillOpacity(polySymb);
            Color fill=SLDs.polyFill(polySymb);
            if( Double.isNaN(opacity) ) opacity = 1.0;
            if( fill != null ){
                fill=new Color(fill.getRed(), fill.getGreen(), fill.getBlue(), (int)(255*opacity));
                g.setColor( fill );
                g.fill( shape );
            }
            if( stroke != null ){
                g.setColor( stroke );
                g.draw( shape );
            }
        }        
        if( symb instanceof LineSymbolizer){
            LineSymbolizer lineSymbolizer = (LineSymbolizer) symb;
            Color c = SLDs.color( lineSymbolizer );
            int w = SLDs.width( lineSymbolizer );
            if( c != null && w > 0 ){
                g.setColor( c );
                g.draw( shape );
            }                      
        }
        if( symb instanceof PointSymbolizer){
            PointSymbolizer pointSymbolizer = (PointSymbolizer) symb;
            
            Color c = SLDs.pointColor( pointSymbolizer );
            Color fill=SLDs.pointFill( pointSymbolizer );
            int size = Math.min(15,SLDs.pointSize( pointSymbolizer ));
            int width = SLDs.width(SLDs.stroke(pointSymbolizer));
            //Rectangle rect=new Rectangle(4,4,6,6 );
            float[] point=new float[6];
            shape.getPathIterator(null).currentSegment(point);
            Rectangle rect = new Rectangle((int)point[0]-size/2, (int)point[1]-size/2, 
            		size, size);
//FOR AMR            Style2D style = styleFactory.createStyle(feature, symbolizers[m], scaleRange);
            
            if( c == null && fill == null ){
                g.setColor( Color.GRAY );
                g.fill( rect );
            }
            

            if( fill != null ){
                g.setColor( fill );
                g.fill( rect );                
            }  else{
                g.setColor( Color.GRAY );
                g.fill( rect );                
            }
            if( c != null ){
            	g.setStroke(ViewportGraphics.LINE_SOLID, width);
                g.setColor( c );
                g.draw( rect );                
            }else{
                g.setColor( Color.DARK_GRAY );
                g.draw( rect );
            }
            //pointSymbolizer.getGraphic().
            //TODO: Fix this code to call the literendered correctly and render the point glyph
//            BufferedImage bi = new BufferedImage(1,2,3);
//            Graphics2D g2d = bi.createGraphics();
//            LiteRenderer2 lite = new LiteRenderer2();
//            Style style = new StyleBuilder().createStyle(pointSymbolizer);
//            lite.paint(g2d, new Rectangle(16,16), feature.getBounds());
        }
    }
    public static Symbolizer[] getSymbolizers(Feature feature) {
    	return getSymbolizers(feature.getDefaultGeometry().getClass(), Color.RED);
    }

    public static Symbolizer[] getSymbolizers(Class<? extends Geometry> type, Color baseColor) {
        StyleBuilder builder=new StyleBuilder();
        Symbolizer[] syms=new Symbolizer[1];
		if( LineString.class.isAssignableFrom(type) || 
        		MultiLineString.class.isAssignableFrom(type) )
        	syms[0]=builder.createLineSymbolizer(baseColor,2);
        if( Point.class.isAssignableFrom(type) || 
        		MultiPoint.class.isAssignableFrom(type)){
        	PointSymbolizer point = builder.createPointSymbolizer(builder.createGraphic());
        	point.getGraphic().getMarks()[0].setSize(FilterFactory.createFilterFactory().createLiteralExpression(10));
        	point.getGraphic().getMarks()[0].setFill(builder.createFill(baseColor));
        	syms[0]=point;
        }
        if( Polygon.class.isAssignableFrom(type) || 
        		MultiPolygon.class.isAssignableFrom(type)){
        	syms[0]=builder.createPolygonSymbolizer(builder.createStroke(baseColor,2), builder.createFill(baseColor, .6));
        }
        return syms;
    }

    /**
     * Finds the geometric attribute requested by the symbolizer
     * 
     * @param f The victim
     * @param s The symbolizer
     * @param style the resolved style for the specified victim
     * @return The geometry requested in the symbolizer, or the default geometry if none is
     *         specified
     */
    private com.vividsolutions.jts.geom.Geometry findGeometry( Feature f, Symbolizer s) {
        String geomName = getGeometryPropertyName(s);
        // get the geometry
        com.vividsolutions.jts.geom.Geometry geom;
        if (geomName == null) {
            geom = f.getDefaultGeometry();
        } else {
            geom = (com.vividsolutions.jts.geom.Geometry) f.getAttribute(geomName);
        }
        // if the symbolizer is a point or text symbolizer generate a suitable
        // location to place the
        // point in order to avoid recomputing that location at each rendering
        // step
        if ((s instanceof PointSymbolizer || s instanceof TextSymbolizer)
                && !(geom instanceof Point)) {
            if (geom instanceof LineString && !(geom instanceof LinearRing)) {
                // use the mid point to represent the point/text symbolizer
                // anchor
                Coordinate[] coordinates = geom.getCoordinates();
                Coordinate start = coordinates[0];
                Coordinate end = coordinates[1];
                Coordinate mid = new Coordinate((start.x + end.x) / 2, (start.y + end.y) / 2);
                geom = geom.getFactory().createPoint(mid);
            } else {
                // otherwise use the centroid of the polygon
                geom = geom.getCentroid();
            }
        }
        return geom;
    }
    private String getGeometryPropertyName( Symbolizer s ) {
        String geomName = null;
        // TODO: fix the styles, the getGeometryPropertyName should probably be
        // moved into an interface...
        if (s instanceof PolygonSymbolizer) {
            geomName = ((PolygonSymbolizer) s).getGeometryPropertyName();
        } else if (s instanceof PointSymbolizer) {
            geomName = ((PointSymbolizer) s).getGeometryPropertyName();
        } else if (s instanceof LineSymbolizer) {
            geomName = ((LineSymbolizer) s).getGeometryPropertyName();
        } else if (s instanceof TextSymbolizer) {
            geomName = ((TextSymbolizer) s).getGeometryPropertyName();
        }
        return geomName;
    }
    /**
     * TODO summary sentence for worldToScreenTransform ...
     * 
     * @param bounds
     * @param rectangle
     * @return
     */
    public static AffineTransform worldToScreenTransform( Envelope mapExtent, Rectangle screenSize ) {
        double scaleX = screenSize.getWidth() / mapExtent.getWidth();
        double scaleY = screenSize.getHeight() / mapExtent.getHeight();

        double tx = -mapExtent.getMinX() * scaleX;
        double ty = (mapExtent.getMinY() * scaleY) + screenSize.getHeight();

        AffineTransform at = new AffineTransform(scaleX, 0.0d, 0.0d, -scaleY,
                tx, ty);

        return at;
    }
    /**
     * Create a FeatureType schema using a type short hand.
     * <p>
     * Code Example:<pre><code>
     * new Drawing().schema("namespace.typename", "id:0,*geom:LineString,name:String,*centroid:Point");
     * </code></pre>
     * <ul>
     * <li>FeatureType with identifier "namespace.typename"
     * <li>Default Geometry "geom" of type LineStirng indicated with a "*"
     * <li>Three attributes: id of type Integer, name of type String and centroid of type Point
     * </ul>
     * </p>
     * @param name namespace.name
     * @param spec
     * @return Generated FeatureType
     */
    public FeatureType schema( String name, String spec ){
        try {
            return DataUtilities.createType( name, spec );
        } catch (SchemaException e) {
            throw new IllegalArgumentException( e );
        }
    }
    
    static FeatureType pointSchema;
    static FeatureType lineSchema ;
    static FeatureType polygonSchema ;
    static FeatureType multipointSchema ;
    static FeatureType multilineSchema ;
    static FeatureType multipolygonSchema ;
    static {
        try {
            pointSchema = DataUtilities.createType( "generated:point", "*point:Point" ); //$NON-NLS-1$ //$NON-NLS-2$
            lineSchema = DataUtilities.createType( "generated:linestring", "*linestring:LineString" ); //$NON-NLS-1$ //$NON-NLS-2$
            polygonSchema = DataUtilities.createType( "generated:polygon", "*polygon:Polygon" ); //$NON-NLS-1$ //$NON-NLS-2$
            multipointSchema = DataUtilities.createType( "generated:multipoint", "*multipoint:MultiPoint" ); //$NON-NLS-1$ //$NON-NLS-2$
            multilineSchema = DataUtilities.createType( "generated:multilinestring", "*multilinestring:MultiLineString" ); //$NON-NLS-1$ //$NON-NLS-2$
            multipolygonSchema = DataUtilities.createType( "generated:multipolygon", "*multipolygon:MultiPolygon" ); //$NON-NLS-1$ //$NON-NLS-2$
        }
        catch ( SchemaException unExpected ){
            System.err.println( unExpected );
        }
    }    
    /**
     * Simple feature with one attribute called "point".
     * @param point 
     * @return Feature with a default geometry and no attribtues 
     */
    public Feature feature( Point point ) {
        if( point == null ) throw new NullPointerException("Point required"); //$NON-NLS-1$
        try {
            return pointSchema.create( new Object[]{ point } );
        } catch (IllegalAttributeException e) {
            // this should not happen because we *know* the parameter matches schame
            throw new RuntimeException("Could not generate feature for point "+point );  //$NON-NLS-1$
        }
    }
    /**
     * Simple Feature with a default geometry and no attribtues.
     * @param line 
     * @return Feature with a default geometry and no attribtues 
     */
    public Feature feature( LineString line ) {
        if( line == null ) throw new NullPointerException("line required"); //$NON-NLS-1$
        try {
            return lineSchema.create( new Object[]{ line } );
        } catch (IllegalAttributeException e) {
            // this should not happen because we *know* the parameter matches schame
            throw new RuntimeException("Could not generate feature for point "+line );  //$NON-NLS-1$
        }
    }
    
    /**
     * Simple Feature with a default geometry and no attribtues.
     * @param polygon 
     * @return Feature with a default geometry and no attribtues 
     */
    public Feature feature( Polygon polygon ) {
        if( polygon == null ) throw new NullPointerException("polygon required"); //$NON-NLS-1$
        try {
            return polygonSchema.create( new Object[]{ polygon } );
        } catch (IllegalAttributeException e) {
            // this should not happen because we *know* the parameter matches schame
            throw new RuntimeException("Could not generate feature for point "+polygon );  //$NON-NLS-1$
        }
    }
    
    /**
     * Simple Feature with a default geometry and no attribtues.
     * @param multipoint 
     * @return Feature with a default geometry and no attribtues 
     */
    public Feature feature( MultiPoint multipoint ) {
        if( multipoint == null ) throw new NullPointerException("multipoint required"); //$NON-NLS-1$
        try {
            return multipointSchema.create( new Object[]{ multipoint } );
        } catch (IllegalAttributeException e) {
            // this should not happen because we *know* the parameter matches schame
            throw new RuntimeException("Could not generate feature for point "+multipoint );  //$NON-NLS-1$
        }
    }
    /**
     * Simple Feature with a default geometry and no attribtues.
     * @param multilinestring
     * @return Feature with a default geometry and no attribtues 
     */
    public Feature feature( MultiLineString multilinestring ) {
        if( multilinestring == null ) throw new NullPointerException("multilinestring required"); //$NON-NLS-1$
        try {
            return multilineSchema.create( new Object[]{ multilinestring } );
        } catch (IllegalAttributeException e) {
            // this should not happen because we *know* the parameter matches schame
            throw new RuntimeException("Could not generate feature for point "+multilinestring );  //$NON-NLS-1$
        }
    }
    /**
     * Simple Feature with a default geometry and no attribtues.
     * @param multipolygon 
     * @return Feature with a default geometry and no attribtues 
     */
    public Feature feature( MultiPolygon multipolygon ) {
        if( multipolygon == null ) throw new NullPointerException("multipolygon required"); //$NON-NLS-1$
        try {
            return multipolygonSchema.create( new Object[]{ multipolygon } );
        } catch (IllegalAttributeException e) {
            // this should not happen because we *know* the parameter matches schame
            throw new RuntimeException("Could not generate feature for point "+multipolygon );  //$NON-NLS-1$
        }
    }
    
    /**
     * Generate Point from two dimensional ordinates
     * 
     * @param x
     * @param y
     * @return Point
     */
    public Point point( int x, int y ){
        return gf.createPoint( new Coordinate( x, y ) );
    }
    /**
     * Generate LineStrings from two dimensional ordinates
     * 
     * @param xy
     * @return LineStirng
     */
    public LineString line(int[] xy) {
        Coordinate[] coords = new Coordinate[xy.length / 2];

        for (int i = 0; i < xy.length; i += 2) {
            coords[i / 2] = new Coordinate(xy[i], xy[i + 1]);
        }

        return gf.createLineString(coords);
    }

    /**
     * Generate a MultiLineString from two dimensional ordinates
     * 
     * @param xy
     * @return MultiLineStirng
     */
    public MultiLineString lines(int[][] xy) {
        LineString[] lines = new LineString[xy.length];

        for (int i = 0; i < xy.length; i++) {
            lines[i] = line(xy[i]);
        }

        return gf.createMultiLineString(lines);
    }
    /**
     * Convience constructor for GeometryFactory.createPolygon.
     * <p>
     * The provided xy ordinates are turned into a linear rings.
     * </p>
     * @param xy Two dimensional ordiantes.
     * @return Polygon
     */
    public Polygon polygon( int[] xy ){
        LinearRing shell = ring( xy );        
        return gf.createPolygon( shell, null );        
    }

    /**
     * Convience constructor for GeometryFactory.createPolygon.
     * <p>
     * The provided xy and holes are turned into linear rings.
     * </p>
     * @param xy Two dimensional ordiantes.
     * @param holes Holes in polygon or null.
     * 
     * @return Polygon 
     */
    public Polygon polygon( int[] xy, int []holes[] ){
        if( holes == null || holes.length == 0){
           return polygon( xy );
        }
        LinearRing shell = ring( xy );        
        
        LinearRing[] rings = new LinearRing[holes.length];

        for (int i = 0; i < xy.length; i++) {
            rings[i] = ring(holes[i]);
        }        
        return gf.createPolygon( shell, rings );        
    }
        
    /**
     * Convience constructor for GeometryFactory.createLinearRing.
     * 
     * @param xy Two dimensional ordiantes.
     * @return LinearRing for use with polygon
     */
    public LinearRing ring( int[] xy ){
        int length = xy.length / 2;
        if( xy[0] != xy[xy.length-2] || xy[1] != xy[xy.length-1]){
            length++;
        }
        Coordinate[] coords = new Coordinate[length];

        for (int i = 0; i < xy.length; i += 2) {
            coords[i / 2] = new Coordinate(xy[i], xy[i + 1]);
        }
        if( xy[0] != xy[xy.length-2] || xy[1] != xy[xy.length-1]){
            coords[length-1] = coords[0];
        }
        return gf.createLinearRing(coords);        
    }   
}
