/* uDig - User Friendly Desktop Internet GIS client
 * http://udig.refractions.net
 * (C) 2004, Refractions Research Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */
package net.refractions.udig.ui;

import java.util.Collections;
import java.util.List;

import net.refractions.udig.internal.ui.Policy;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.dialogs.IDialogSettings;
import org.eclipse.jface.viewers.IBaseLabelProvider;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.ListViewer;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.IMemento;
import org.eclipse.ui.IPartListener2;
import org.eclipse.ui.IViewSite;
import org.eclipse.ui.IWorkbenchPartReference;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.part.ISetSelectionTarget;
import org.eclipse.ui.part.ViewPart;

/**
 * A ViewPart with support for a background "search" process.
 * <p>
 * The part will automatically show the "progress" when the monitored
 * Job(s) are running.
 * </p>
 * @author jgarnett
 * @since 1.0.0
 */
public class SearchPart extends ViewPart implements ISetSelectionTarget {
    
    /** Viewer for search results List/Tree/Table ... */
    protected StructuredViewer viewer; // often a tree or table viewer
        
    /**
     * Used to cancel current searchMonitor
     * <p>
     * Usually available in the toolbar.
     * </p>
     */
    protected IAction cancel;
    
    /**
     * Used to control search jobs.
     * <p>
     * Only one "batch" of search jobs is outstanding at anyone
     * time.
     * </p>
     */
    protected IProgressMonitor searchMonitor;

    protected SashForm splitter;

    /** Save state here */
    protected IMemento save;
    protected Composite parent;
    
    IDialogSettings settings;
    
    /** We need dialog settings for persistence */
    protected SearchPart( IDialogSettings dialogSettings ) {
        settings = dialogSettings;
    }
    @Override
    public void init( IViewSite site, IMemento memento ) throws PartInitException {
        super.init(site, memento);
        save = memento;
    }        
    
    @Override
    public void saveState( IMemento memento ) {
        if ( splitter == null) { // part has not been created
            if (save != null) { //Keep the old state;
                memento.putMemento( save );
            }
            return;
        }        
    }
    private void addResizeListener(Composite parent) {
        parent.addControlListener(new ControlListener() {
            public void controlMoved(ControlEvent e) {
            }
            public void controlResized(ControlEvent e) {
                computeOrientation();
            }
        });
    }
     
    protected enum Orientation { VERTICAL, HORIZONTAL, SINGLE, AUTOMATIC };
    protected Orientation orientation = Orientation.VERTICAL;
    
    void computeOrientation() {
        saveSplitterRatio();
        if( save != null ) save.putInteger("orientation", orientation.ordinal() );
        if ( orientation != Orientation.AUTOMATIC ) {
            setOrientation( orientation );
        }
        else {
            if (orientation == Orientation.SINGLE ) return;
            
            Point size= parent.getSize();
            if (size.x != 0 && size.y != 0) {
                if (size.x > size.y) 
                    setOrientation( Orientation.HORIZONTAL );
                else 
                    setOrientation( Orientation.VERTICAL );
            }
        }
    }
    private void saveSplitterRatio() {
        if (splitter != null && ! splitter.isDisposed()) {
            int[] weigths = splitter.getWeights();
            int ratio = (weigths[0] * 1000) / (weigths[0] + weigths[1]);
            settings.put( "ratio"+orientation, ratio );            
        }
    }
    private void restoreSplitterRatio() {
        try {
            Integer ratio= settings.getInt("ratio"+orientation);
            if (ratio == null)
                return;
            splitter.setWeights(new int[] {ratio, 1000 - ratio});
        }
        catch( NumberFormatException nan ) {
            // ignore bad setting
        }
    }
    
    Orientation currentOrientation;
    private boolean showDetails;
    /**
     * called from ToggleOrientationAction (or compute).
     * 
     * @param orientation Orientation.HORIZONTAL or Orientation.VERTICAL
     */
    protected void setOrientation(Orientation orientation) {
        if (currentOrientation == orientation) {
            return; // no change needed
        }
        if (viewer != null && !viewer.getControl().isDisposed() &&
            splitter != null && !splitter.isDisposed() ) {
            
            if (orientation == Orientation.SINGLE) {
                setShowDetails(false);
            } else {
                if (currentOrientation == Orientation.SINGLE) {
                    setShowDetails(true);
                }
                boolean horizontal = orientation == Orientation.HORIZONTAL;
                splitter.setOrientation( horizontal ? SWT.HORIZONTAL : SWT.VERTICAL);
            }
            splitter.layout();
        }
        updateCheckedState();

        currentOrientation = orientation;
        
        restoreSplitterRatio();
    }
    private ToggleOrientationAction[] toggleOrientationActions;
    private Composite details;
    private IPartListener2 partListener;
    private void updateCheckedState() {
        for (ToggleOrientationAction toggle : toggleOrientationActions) {
            toggle.setChecked( orientation == toggle.getOrientation());
        }
    }
    public void setShowDetails(boolean show) {
        showDetails = show;
        showOrHideDetails();
    }
    private void showOrHideDetails() {
        if (showDetails) {
            splitter.setMaximizedControl(null);
        } else {
            splitter.setMaximizedControl( viewer.getControl() );
        }
    }
    
    /**
     * Creates the SWT controls for this workbench part.
     * <p>
     * The details (from IWorkbenchPart.createPartControl( Composite ))
     * </p>
     * <p>
     * Multi-step process:
     * <ol>
     * <li>Create one or more controls within the parent.</li>
     * <li>Set the parent layout as needed.</li>
     * <li>Register any global actions with the <code>IActionService</code>.</li>
     * <li>Register any popup menus with the <code>IActionService</code>.</li>
     * <li>Register a selection provider with the <code>ISelectionService</code> (optional).
     * </li>
     * </ol>
     * </p>
     * 
     * @see org.eclipse.ui.part.WorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
     * @param parent
     */
    public void createPartControl( Composite aParent ) {
        parent= aParent;
        addResizeListener(parent);
        
        // split
        splitter = new SashForm( parent, SWT.HORIZONTAL);
        // TODO: key listener
        
        viewer = createViewer( splitter );
        viewer.setContentProvider( createContentProvider() );
        viewer.setLabelProvider( createLableProvider() );        
        viewer.addSelectionChangedListener( createSelectionListener() );
        getSite().setSelectionProvider(viewer);        
        details = createDetails( splitter );
        
        initDragAndDrop();
        
        makeActions();
        fillViewMenu();
        fillActionBars();
        
        initOrientation();
        
        if (save != null) {
            restoreState( save );
        }
        restoreSplitterRatio();
        addPartListener();
    }
    
    /**
     * You can now restore whatever state you need.
     */
    private void restoreState(IMemento memento) {
        //
    }
    protected void saveViewSettings() {        
        saveSplitterRatio();
        settings.put( "orientation", orientation.ordinal() );        
    }
    private void addPartListener() {
        final String ID = getViewSite().getId();
        partListener= new IPartListener2() {
                    public void partActivated(IWorkbenchPartReference partRef) { }
                    public void partBroughtToTop(IWorkbenchPartReference partRef) { }
                    public void partClosed(IWorkbenchPartReference partRef) {
                        if (ID.equals(partRef.getId()))
                            saveViewSettings();
                    }
                    public void partDeactivated(IWorkbenchPartReference partRef) {
                        if (ID.equals(partRef.getId()))
                            saveViewSettings();
                    }
                    public void partOpened(IWorkbenchPartReference partRef) { }
                    public void partHidden(IWorkbenchPartReference partRef) { }
                    public void partVisible(IWorkbenchPartReference partRef) { }
                    public void partInputChanged(IWorkbenchPartReference partRef) { }
                };
        getViewSite().getPage().addPartListener(partListener);
    }
    /**
     * Should do the dialog settings thing here ...
     */
    private void initOrientation() {
        orientation = Orientation.AUTOMATIC;
        if( save != null ) {
            Integer integer = save.getInteger( "orientation" );
            if( integer != null ) {
                orientation = Orientation.values()[ integer ];    
            }            
        }

        // force the update
        currentOrientation = null;
        setOrientation( orientation );
    }
    /**
     * Subclass should override to provide custom details display.
     * <p>
     * Many subclasses use PageBook with page selection based on
     * the selection in the viewer.
     * </p>
     * @param splitter
     * @return Composite to use for details display
     */
    protected Composite createDetails( SashForm splitter ) {
        return new Composite( splitter, SWT.NONE );        
    }
    /**
     * Subclass can override to return its kind of details,
     * example a PageBook.
     *
     * @return
     */
    protected Composite getDetails() {
        return details;
    }    
    
    /**
     * Create viewer (default is a ListViewer please override if you want a Tree or Table Viewer.
     * <p>
     * Note the following will be called after creation:
     * <ul>
     * <li>setContentProvider( createContentProvider() );
     * <li>viewer.setLabelProvider( createLableProvider() );
     * </ul>
     * This allows people who subclass you to do there own thing.
     * </p>    
     * @param page
     * @return
     */
    protected StructuredViewer createViewer( Composite parent ) {
        ListViewer viewer = new ListViewer( parent );
        return viewer;
    }
    /**
     * Default implementation calls showDetail( IStructuredSelection ).
     * <p>
     * Override if you want to do something else.
     */
    protected ISelectionChangedListener createSelectionListener() {
        return new ISelectionChangedListener() {
            public void selectionChanged( SelectionChangedEvent event ) {
                ISelection sel= event.getSelection();
                if( sel instanceof IStructuredSelection ) {
                    IStructuredSelection selection = (IStructuredSelection) sel;
                    showDetail( selection.getFirstElement() );
                }
            }            
        };
    }
    /** Allows subclass to focus the detail on this selection  */ 
    protected void showDetail( Object selection ) {
        //
    }

    /**
     * Default implementation will work for lists, please overide if
     * you are into the whole tree thing.
     * 
     * @return
     */
    protected IStructuredContentProvider createContentProvider() {
        return new IStructuredContentProvider() {
            public Object[] getElements( Object inputElement ) {
                if( inputElement instanceof List ) {
                    return ((List)inputElement).toArray();
                }
                return null;
            }
            public void dispose() {
            }
            public void inputChanged( Viewer viewer, Object oldInput, Object newInput ) {
                assert( newInput instanceof List );
                // lists don't have events for us to watch
            }            
        };
    }

    /**
     *
     * @return
     */
    protected IBaseLabelProvider createLableProvider() {
        return new LabelProvider();
    }

    /** Will cancel any outstanding search */
    synchronized void stopSearch(){        
        if( searchMonitor != null ){
            IProgressMonitor cancelMonitor = searchMonitor;
            
            searchMonitor = null;
            cancelMonitor.setCanceled( true );            
        }        
    }
        
    /**
     * Subclass should override to support DnD.
     */
    protected void initDragAndDrop() {
        //
    }    

    protected void makeActions() {
        cancel = new Action() {
            public void run() {
                stopSearch();
                //book.showPage( promptPage );                
            }
        };
        Policy.initAction( cancel, "cancel" );
        cancel.setEnabled( false );
        
        toggleOrientationActions = new ToggleOrientationAction[] {
                new ToggleOrientationAction(this, Orientation.VERTICAL),
                new ToggleOrientationAction(this, Orientation.HORIZONTAL),
                new ToggleOrientationAction(this, Orientation.AUTOMATIC),
                new ToggleOrientationAction(this, Orientation.SINGLE)
            };
        
    }
    /**
    *
    */
   protected void fillViewMenu() {
       IActionBars actionBars = getViewSite().getActionBars();
       IMenuManager viewMenu = actionBars.getMenuManager();
       viewMenu.add(new Separator());
       viewMenu.add( cancel );
       viewMenu.add(new Separator());
       for (ToggleOrientationAction toggle : toggleOrientationActions) {
           viewMenu.add( toggle );
       }
   }    
    
    /**
     * Create toolbar with cancel.
     */
    protected void fillActionBars() {
        IActionBars actionBars = getViewSite().getActionBars();
        IToolBarManager toolBar = actionBars.getToolBarManager();

        toolBar.add( cancel );                
    }
    
    public void quick( String pattern ) {
        // search through existing contents based on label?
    }

    Object filter = null;
    Job looking = new Job( getPartName()+"..."){
        protected IStatus run( IProgressMonitor monitor ) {
            try {                    
                final List results = searchImplementation( filter, monitor );
                cancel.setEnabled( false );                    
                Display.getDefault().asyncExec(new Runnable(){
                    public void run() {
                        if( results == null || results.isEmpty() ) {
                            viewer.setInput( null );
                        }
                        else {
                            viewer.setInput( results );                                    
                        }
                    }
                }); 
            }                
            catch (Throwable t ) {
                //book.showPage( promptPage );                    
            }
            return Status.OK_STATUS;                
        }
    }; 
    
    /**
     * Search the catalog for text and update view contents
     * 
     * @param pattern
     */
    public void search( final Object newFilter ) {        
        if ( newFilter== null) {
            stopSearch();            
            //book.showPage( promptPage );
            return;
        }        
        stopSearch();
        
        viewer.setInput( null );
        filter = newFilter;        
        searchMonitor=Platform.getJobManager().createProgressGroup();
        searchMonitor.setTaskName( getPartName());           
        looking.setPriority( Job.BUILD );
        looking.setProgressGroup( searchMonitor, IProgressMonitor.UNKNOWN );
        cancel.setEnabled( true );
        looking.schedule(); // do it                
    }
  
    protected List searchImplementation( Object filter, IProgressMonitor monitor ) {
        return Collections.EMPTY_LIST;
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see org.eclipse.ui.part.ISetSelectionTarget#selectReveal(org.eclipse.jface.viewers.ISelection)
     */
    public void selectReveal( ISelection selection ) {
        //viewer.setSelection(selection);
    }

    @Override
    public void setFocus() {
        if( viewer != null && viewer.getControl().isVisible() ) {
            viewer.getControl().setFocus();
        }
    }
    
    /**
     * Toggles the orientationof the layout
     */
    static class ToggleOrientationAction extends Action {
        private SearchPart view;    
        private Orientation orientation;        
        public ToggleOrientationAction(SearchPart searchPart, Orientation orientation) {
            super("", AS_RADIO_BUTTON); //$NON-NLS-1$            
            Policy.initAction( this, "orientation."+orientation.toString().toLowerCase() );            
            view = searchPart;
            this.orientation = orientation;
            //PlatformUI.getWorkbench().getHelpSystem().setHelp(this, lookup context id? );                    
        }        
        public Orientation getOrientation() {
            return orientation;
        }           
        /*
         * @see Action#actionPerformed
         */     
        public void run() {
            if (isChecked()) {
                view.orientation = orientation; 
                view.computeOrientation();
            }
        }
        
    }
}
